(function($) {
    let optionsInitialised = false;
    let editMode = false;
    let currentCursor = null;
    let modalMessageVisible = false;
    let cindex = 0, uindex = 0;
    let colourCycle = [ [ 136, 221, 204 ], [ 213, 178, 64 ], [ 187, 187, 221 ], [ 255, 136, 119 ], [ 136, 187, 221 ], [ 255, 187, 102 ], [ 187, 221, 102 ], [ 255, 204, 238 ], [ 166, 206, 227 ], [ 187, 136, 187 ], [ 204, 238, 204 ], [ 166, 206, 227 ] ];
    let publicMethods = {
        init: function(args) {
            if (!optionsInitialised) {
                optionsInitialised = true;
                $(document).on("keydown", function(e) {
                    if (editMode && $(".fancybox-desktop:visible").length < 1) {
                        keydownEdit(e);
                    } else {
                        if (e.key === "Escape") {
                            $(".hoverResidue").removeClass("hoverResidue");
                            $(".selectResidue").removeClass("selectResidue");
                            document.body.dispatchEvent(new CustomEvent("selectResidues", {
                                detail: {
                                    src: "alignment"
                                }
                            }));
                        }
                    }
                });
            }
            return this.each(function() {
                let templateData = $.extend({}, {
                    type: "target-template",
                    chaingroups: [],
                    keyseq_sequence: args.keyseq_sequence,
                    keyseq_chain_name: args.keyseq_chain_name ? args.keyseq_chain_name : [ "A" ],
                    keyseq_annotation: args.keyseq_annotation
                }, args.templateData);
                if (args.nowrap) templateData.nowrap = true;
                if (args.charts) {
                    templateData.qmLength = 0;
                    templateData.charts = args.charts;
                    templateData.pageResidueCount = args.pageResidueCount;
                }
                if (args.resetColourCycle) cindex = 0;
                args["firstDraw"] = true;
                $(this).data("templateData", templateData);
                drawAlignment($(this), null, args);
                publicMethods.toggleIdenticalSeqs.call(this, {
                    chaingroups: templateData.chaingroups
                }, false);
                publicMethods.resizeToParent.call($(this), {
                    redrawAnnotation: false
                });
                drawAnnotation($(this));
                addSequenceListener($(this));
            });
        },
        toggleIdenticalSeqs: function(chaingroups, redrawAnnotation) {
            return $(this).each(function() {
                let $alignTbl = $(this);
                let $alignNameTbl = getAlignNameTbl($alignTbl);
                for (let c = 0; c < chaingroups.chaingroups.length; c++) {
                    let group = chaingroups.chaingroups[c].filter(function() {
                        return true;
                    });
                    if (group.length < 2) continue;
                    let repRow = group[0];
                    let tmp = $alignNameTbl.find('tr[chain="' + repRow + '"]').first().text();
                    let compactText = tmp.substring(0, tmp.lastIndexOf(".") + 1) + "(" + group.join("") + ")";
                    let originalText = tmp.substring(0, tmp.lastIndexOf(".")) + "." + repRow.charAt(0);
                    let sel = "";
                    for (let i = 1; i < group.length; i++) {
                        sel += "[chain=" + group[i] + "]";
                        if (i < group.length - 1) sel += ",";
                    }
                    $alignTbl.find("tr" + sel).toggle();
                    $alignNameTbl.find("tr" + sel).toggle();
                    if ($alignNameTbl.find("tr[chain=" + group[1] + "]").css("display") == "none") $alignNameTbl.find("tr[chain=" + repRow + "]").find("td").html(compactText); else $alignNameTbl.find("tr[chain=" + repRow + "]").find("td").html(originalText);
                }
                if (typeof redrawAnnotation == "undefined" || redrawAnnotation) {
                    drawAnnotation($alignTbl);
                }
            });
        },
        drawAlignment: function(args) {
            return this.each(function() {
                drawAlignment($(this), args ? args.seqs : null, args ? args : null);
                addSequenceListener($(this));
            });
        },
        getFormattedAlignment: function(args) {
            return getFormattedAlignment($(this).attr("id").substring(8), args.format);
        },
        exportAlignment: function(args) {
            return this.each(function() {
                $("#formattedAlignment").show();
                $("#alignmentFormatSelect").val(args["format"]);
                updateExportText();
                $("<a href='#formattedAlignment' title='Copy & Paste Alignment'/>'").fancybox({
                    helpers: {
                        title: {
                            type: "inside"
                        }
                    }
                }).trigger("click");
                $("#formattedAlignmentTextarea").focus();
                $("#formattedAlignmentTextarea").select();
            });
        },
        html2image: function(args) {
            if ($.fancybox) $.fancybox.close();
            $(".popover").hide();
            let $alTbl = $("#" + (args && args["alignId"] ? args["alignId"] : "alignTbl" + $("#currentAlignmentIndex").val()));
            let svg = html2svg($alTbl);
            let blob = new Blob([ svg ], {
                type: "image/svg+xml;charset=utf-8"
            });
            let DOMURL = window.URL || window.webkitURL || window;
            let url = DOMURL.createObjectURL(blob);
            let img = new Image();
            let c = document.createElement("canvas");
            let ctx = c.getContext("2d");
            img.onload = function() {
                c.width = this.naturalWidth;
                c.height = this.naturalHeight;
                ctx.drawImage(this, 0, 0);
                c.toBlob(function(blob) {
                    saveAs(blob, "alignment.png");
                }, "image/png");
            };
            img.src = url;
        },
        colourAllAlignments: function(args) {
            resetChainMappingColours();
            if (args.features) {
                $(".alignTbl").each(function() {
                    let thisStrucIds = [];
                    $(this).find("tr[struc_id]").each(function(i, row) {
                        let struc_id = row.getAttribute("struc_id");
                        if (thisStrucIds.indexOf(struc_id) == -1) thisStrucIds.push(struc_id);
                    });
                    let featuresForThis = {};
                    for (let struc_id in args.features) if (thisStrucIds.indexOf(struc_id) > -1) featuresForThis[struc_id] = args.features[struc_id]; else if (thisStrucIds.indexOf(struc_id + ".dna") > -1) featuresForThis[struc_id + ".dna"] = args.features[struc_id];
                    if (Object.keys(featuresForThis).length == 0) $(this).removeData("features"); else $(this).data("features", featuresForThis);
                    colourAlignment("features", $(this));
                });
            } else {
                if (args.col) {
                    if (args.col != chainMapping.defaultColourScheme && args.col != jQuery.cookie("colourScheme")) jQuery.cookie("colourScheme", args.col, {
                        expires: 365,
                        path: "/"
                    });
                } else {
                    args.col = chainMapping.defaultColourScheme;
                    if (!args.col) args.col = jQuery.cookie("colourScheme");
                }
                $(".alignTbl").each(function() {
                    $(this).removeData("features");
                    colourAlignment(args.col, $(this));
                });
            }
            getColoursForViewer();
        },
        changeExportFormat: function(args) {
            return this.each(function() {
                updateExportText();
            });
        },
        secondaryStructure: function(args) {
            if (args) jQuery.cookie("secStruct", args.sstype, {
                expires: 365,
                path: "/"
            });
            let ss = null;
            if (args) ss = args.sstype; else ss = $.cookie("secStruct");
            if (ss == null) ss = "dssp";
            let ret = this.each(function() {
                if ($(this).data("templateData")) {
                    drawAnnotation($(this));
                    if ($.cookie("colourScheme") == "secstruc") {
                        colourAlignment("secstruc", $(this));
                    }
                }
            });
            let pv_structure = null;
            let tpls = null;
            let keyseq_annotation = null;
            if (typeof globalTemplateData != "undefined") {
                if (typeof globalTemplateData != "undefined") {
                    tpls = globalTemplateData;
                }
            } else if ($(this).data("templateData")) {
                if ($(this).data("templateData").type == "model-template") {
                    keyseq_annotation = true;
                } else {
                    tpls = [];
                    this.each(function() {
                        if ($(this).data("templateData")) {
                            tpls = tpls.concat($(this).data("templateData")["templates"]);
                        }
                    });
                }
            }
            if ($("#pv:visible", currentWin.document).length > 0 && typeof pv_viewer !== "undefined") {
                if (keyseq_annotation) {
                    this.each(function() {
                        let tplData = $(this).data("templateData");
                        if (!tplData) return;
                        let struc_id = tplData.struc_id;
                        if (!pv_viewer.get(struc_id)) return;
                        let pv_structure = pv_viewer.get(struc_id).structure();
                        pv_viewer.rm(struc_id);
                        let clearSecStruc = ss == "none";
                        if (!clearSecStruc) {
                            for (let c = 0; c < tplData["keyseq_chain_name"].length; c++) {
                                let kschain = tplData["keyseq_chain_name"][c];
                                let trg_offset = tplData["templates"][0]["trg_offset"];
                                if (!trg_offset) trg_offset = 1;
                                let anno_string = tplData["keyseq_annotation"][kschain][ss + "_string"];
                                let chain = pv_structure.select({
                                    chain: kschain
                                });
                                if (anno_string) chain.eachResidue(function(r) {
                                    r.full().setSS(anno_string[r.num() - trg_offset]);
                                }); else clearSecStruc = true;
                            }
                        }
                        if (clearSecStruc) {
                            pv_structure.select("protein").eachResidue(function(r) {
                                r.full().setSS("");
                            });
                        }
                        pv_viewer.cartoon(struc_id, pv_structure, {
                            color: pv.color.uniform("white")
                        });
                    });
                } else {
                    for (let i = 0; i < tpls.length; i++) {
                        if (tpls[i].selected) {
                            let struc_id = tpls[i].struc_id;
                            pv_structure = pv_viewer.get(struc_id).structure();
                            pv_viewer.rm(struc_id);
                            if (ss == "none") {
                                pv_structure.select("protein").eachResidue(function(r) {
                                    r.full().setSS("");
                                });
                            } else {
                                if (tpls[i]["targets"]) {
                                    for (let t = 0; t < tpls[i]["targets"].length; t++) {
                                        let trg = tpls[i]["targets"][t];
                                        for (let c = 0; c < trg["chain"].length; c++) {
                                            annotation = trg["annotation"][trg["chain"][c]][ss + "_string"];
                                            if (!annotation) annotation = trg["annotation"][trg["chain"][c]][ss];
                                            pv_structure.select({
                                                chain: trg["chain"][c]
                                            }).eachResidue(function(r) {
                                                r.full().setSS(annotation[r.num() - 1]);
                                            });
                                        }
                                    }
                                } else {
                                    let annotation = tpls[i]["annotation"][ss + "_string"];
                                    if (!annotation) annotation = tpls[i]["annotation"][ss];
                                    let chn = pv_structure.select({
                                        chain: tpls[i]["chain"]
                                    });
                                    chn.eachResidue(function(r) {
                                        r.full().setSS(annotation ? annotation[r.num() - 1] : "");
                                    });
                                }
                            }
                            pv_viewer.cartoon(struc_id, pv_structure, {
                                color: pv.color.uniform("white")
                            });
                        }
                    }
                }
            } else if ($("#ngl:visible", currentWin.document).length > 0 && typeof ngl_stage !== "undefined") {
                if (keyseq_annotation) {
                    this.each(function() {
                        let tplData = $(this).data("templateData");
                        if (!tplData) return;
                        let struc_id = tplData.struc_id;
                        let clearSecStruc = ss == "none";
                        for (let c = 0; c < tplData["keyseq_chain_name"].length; c++) {
                            let kschain = tplData["keyseq_chain_name"][c];
                            let trg_offset = tplData["templates"][0]["trg_offset"];
                            if (!trg_offset) trg_offset = 1;
                            let anno_string = tplData["keyseq_annotation"][kschain][ss + "_string"];
                            if (!anno_string) clearSecStruc = true;
                            for (let comp in ngl_stage.compList) {
                                if (ngl_stage.compList[comp].name == struc_id) {
                                    ngl_stage.compList[comp].structure.eachResidue(function(a) {
                                        if (a.chainname == kschain) if (clearSecStruc) a.sstruc = ""; else a.sstruc = anno_string[a.resno - trg_offset].toLowerCase();
                                    });
                                    break;
                                }
                            }
                        }
                    });
                } else {
                    for (let i = 0; i < tpls.length; i++) {
                        if (tpls[i].selected) {
                            ngl_stage.eachRepresentation(function(rep) {
                                if (rep.name && !rep.name.match(/(ligands|dna|over)/)) {
                                    let strucId = rep.name;
                                    for (let comp in ngl_stage.compList) {
                                        if (ngl_stage.compList[comp].name == strucId) {
                                            ngl_stage.compList[comp].removeRepresentation(rep);
                                        }
                                    }
                                }
                            });
                            if (ss == "none") {
                                for (let comp in ngl_stage.compList) if (ngl_stage.compList[comp].name == tpls[i].struc_id) ngl_stage.compList[comp].structure.eachResidue(function(a) {
                                    a.sstruc = "";
                                });
                            } else {
                                if (tpls[i]["targets"]) {
                                    for (let t = 0; t < tpls[i]["targets"].length; t++) {
                                        let trg = tpls[i]["targets"][t];
                                        for (let c = 0; c < trg["chain"].length; c++) {
                                            annotation = trg["annotation"][trg["chain"][c]][ss + "_string"];
                                            if (!annotation) annotation = trg["annotation"][trg["chain"][c]][ss];
                                            for (let comp in ngl_stage.compList) {
                                                if (ngl_stage.compList[comp].name == tpls[i].struc_id) {
                                                    ngl_stage.compList[comp].structure.eachResidue(function(a) {
                                                        if (a.chainname == trg["chain"][c]) {
                                                            a.sstruc = annotation[a.resno - 1].toLowerCase();
                                                        }
                                                    });
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                } else {
                                    if (tpls[i]["annotation"][ss + "_string"]) annotation = tpls[i]["annotation"][ss + "_string"]; else annotation = tpls[i]["annotation"][ss];
                                    for (let comp in ngl_stage.compList) {
                                        if (ngl_stage.compList[comp].name == tpls[i].struc_id) {
                                            ngl_stage.compList[comp].structure.eachResidue(function(a) {
                                                if (a.chainname == tpls[i]["chain"]) {
                                                    a.sstruc = annotation ? annotation[a.resno - 1].toLowerCase() : "";
                                                }
                                            });
                                            break;
                                        }
                                    }
                                }
                            }
                            struc_id = tpls[i].struc_id;
                            for (let comp in ngl_stage.compList) {
                                if (ngl_stage.compList[comp].name == struc_id) {
                                    ngl_stage.compList[comp].addRepresentation("cartoon", {
                                        name: struc_id,
                                        sele: "not :_ and not nucleic and not :-",
                                        smoothSheet: true,
                                        aspectRatio: 6,
                                        roughness: 1
                                    });
                                }
                            }
                        }
                    }
                }
            }
            changeRep($.cookie("defaultRepr"));
            getColoursForViewer();
            return ret;
        },
        targetMismatch: function(args) {
            if (args.match) jQuery.cookie("targetMismatch", args.match, {
                expires: 365,
                path: "/"
            }); else jQuery.removeCookie("targetMismatch", {
                path: "/"
            });
            return this.each(function() {
                $(this).find("span.targetMismatch").removeClass("targetMismatch");
                if (args.match == "fade") {
                    $("#enhanceMismatch_cb").prop("checked", false);
                    $(this).find("span[nm]").addClass("targetMismatch");
                } else if (args.match == "enhance") {
                    $("#fadeMismatch_cb").prop("checked", false);
                    let $span = $(this).find("span");
                    $span.not("[nm]").addClass("targetMismatch");
                    $span.not("[s]").addClass("targetMismatch");
                }
            });
        },
        viewerInitialised: function(args) {
            let col = null;
            if (chainMapping.defaultColourScheme) col = chainMapping.defaultColourScheme; else {
                if ($(this).data("features")) col = "features"; else col = $.cookie("colourScheme");
            }
            if ([ "qmean_local", "confClass", "bfactor", "bfactor_range", "secstruc", "lddt" ].indexOf(col) > -1) $(".alignTbl:first").alignment("colourAllAlignments", {
                col: col
            }); else getColoursForViewer(args);
        },
        updateColoursSingleResidue: function(args) {
            getColoursForViewer({
                singleResidue: args
            });
        },
        resizeToParent: function(args, redrawAnnotation) {
            if (args) redrawAnnotation = args["redrawAnnotation"]; else redrawAnnotation = true;
            return this.each(function() {
                let $alTbl = $(this);
                if (!$alTbl.is(":visible") || !$alTbl.data("templateData")) return;
                let $firstTd = $(this).find("tr:visible td").first();
                let firstSpan = $firstTd.find("span").first()[0];
                if (!firstSpan) return;
                let indexWidth = $alTbl.width() - $firstTd.width();
                let $container = $alTbl.parent().closest("table").parent();
                let availableWidth = $container.width() - ($alTbl.offset().left - $container.offset().left) - indexWidth;
                if ($container.css("maxHeight")) availableWidth -= 8;
                if ($alTbl.data("templateData").nowrap) availableWidth = 1e7;
                let txt = $firstTd.text();
                let oldSize = txt.length;
                let newSize = Math.floor(availableWidth / firstSpan.getBoundingClientRect().width);
                if ($alTbl.data("options")["textOnly"]) newSize = Math.floor(availableWidth / ($firstTd.width() / $firstTd.text().length));
                newSize -= newSize % 5;
                if (newSize < 1) newSize = 5;
                let hideIdenticalSeqs = $alTbl.find("tr:hidden").length > 0;
                let chGroups = $alTbl.data("templateData").chaingroups;
                if (newSize != oldSize) {
                    if ($alTbl.attr("cols_cookie")) $.cookie($alTbl.attr("cols_cookie"), newSize, {
                        expires: 365,
                        path: "/"
                    });
                    drawAlignment($alTbl, null, {
                        cols: newSize
                    });
                    if (redrawAnnotation) drawAnnotation($alTbl);
                    if (hideIdenticalSeqs) {
                        publicMethods.toggleIdenticalSeqs.call($alTbl, {
                            chaingroups: chGroups
                        });
                    }
                } else {
                    if (redrawAnnotation) drawAnnotation($alTbl);
                    repositionAnnotation($alTbl);
                }
            });
        },
        toggleEditMode: function(args) {
            if (args && args.edit != null) editMode = args.edit; else editMode = !editMode;
            if (editMode) {
                $("a[id^=gapEdit_]").css("opacity", 1);
                $(".hoverResidue").removeClass("hoverResidue");
                if (currentCursor == null) currentCursor = {
                    seqno: 1,
                    seqidx: 0
                };
                $(this).find("tr[seqno=" + currentCursor.seqno + "]").find("span").eq(currentCursor.seqidx).addClass("cursor");
            } else {
                $("a[id^=gapEdit_]").css("opacity", .6);
                $(".cursor").removeClass("cursor");
            }
        },
        removeGappedColumns: function(args) {
            let templates = $(this).data("templateData").templates;
            let tmp = removeGappedColumns(templates[0].trg_seq, templates[0].tpl_seq);
            templates[0].trg_seq = tmp[0];
            templates[0].tpl_seq = tmp[1];
            drawAlignment($(this));
        }
    };
    function getAlignNameTbl(alignTbl) {
        return $("#alignNameTbl" + alignTbl.attr("id").substring(8));
    }
    function addSequenceListener($alignTbl) {
        $(document).on("mouseup", function() {
            if ($.data($alignTbl, "downchain")) {
                let $selection = $("#offscreen_selection span");
                if (window.getSelection && $selection.length > 0) {
                    let residues = [];
                    let struc_id = null;
                    for (let sp of document.querySelectorAll("tr[struc_id] span.selectResidue")) {
                        let row = sp.closest("tr");
                        struc_id = row.getAttribute("struc_id");
                        if (!row) continue;
                        let chn = row.getAttribute("chain");
                        residues.push({
                            chain: chn ? chn : row.getAttribute("keyseqchain"),
                            resno: parseInt(sp.getAttribute("s"))
                        });
                    }
                    document.body.dispatchEvent(new CustomEvent("selectResidues", {
                        detail: {
                            src: "alignment",
                            struc_id: struc_id,
                            residues: residues
                        }
                    }));
                    let windowSelection = window.getSelection();
                    let range = document.createRange();
                    range.setStart($selection.get(0), 0);
                    range.setEndAfter($selection.get(-1), 0);
                    windowSelection.addRange(range);
                } else if (document.selection) document.selection.empty();
                $.data($alignTbl, "downchain", null);
            }
            $alignTbl.data("mouseDragEdit", null);
        });
        $(document).on("mousemove", function() {
            if ($.data($alignTbl, "downchain")) {
                if (window.getSelection) {
                    windowSelection = window.getSelection();
                    windowSelection.removeAllRanges();
                } else if (document.selection) document.selection.empty();
            }
        });
        $alignTbl.off();
        if ($alignTbl.data("templateData") && $alignTbl.data("templateData").charts) $alignTbl.on("mouseover mousemove mousedown mouseup dblclick", "td", function(e) {
            let $spans = $(this).find("span");
            if ($spans.length == 0) return;
            let spanWidth = $spans[0].getBoundingClientRect().width;
            let x = e.pageX || e.pageY ? e.pageX - $(this).offset().left : e.clientX + document.body.scrollLeft + document.documentElement.scrollLeft - $(this).offset().left;
            let $span = $($spans[Math.floor(x / spanWidth)]);
            mouseOverAlignment($alignTbl, $span, e);
        }); else $alignTbl.on("mouseover mousedown mouseup dblclick", "span", function(e) {
            mouseOverAlignment($alignTbl, $(this), e);
        });
        $alignTbl.on("mouseleave", e => {
            document.querySelectorAll(".hoverResidue").forEach(el => el.classList.remove("hoverResidue"));
            document.body.dispatchEvent(new CustomEvent("highlightResidue", {
                detail: {
                    src: "alignment"
                }
            }));
        });
    }
    let lastSpanOver = null;
    function mouseOverAlignment($alignTbl, $span, e) {
        let text = $span.text();
        if (editMode) delegateEditMode($span, e);
        if (text === "-") return;
        let seqres = $span.attr("s");
        if (!seqres) return;
        let $row = $span.closest("tr");
        let chain = $row.attr("chain");
        if (!chain) {
            if ($row.attr("keyseqchain")) chain = $row.attr("keyseqchain"); else chain = null;
        }
        let struc_id = $row.attr("struc_id");
        let seqno = $row.attr("seqno");
        let spanId = seqno + struc_id + "_" + $row.index() + "_" + $span.index();
        if (e.type === "mousemove") {
            if (lastSpanOver === spanId) return;
            e.type = "mouseover";
        } else if (e.type === "mouseup") {
            if (struc_id && chain) {
                if ($.data($alignTbl, "downspan")[0] === $span[0]) {
                    let isSelected = $span.hasClass("selectResidue");
                    if (isSelected) $span.removeClass("selectResidue"); else $span.addClass("selectResidue");
                }
                let residues = [];
                for (let sp of document.querySelectorAll('tr[struc_id="' + struc_id + '"] span.selectResidue')) {
                    let row = sp.closest("tr");
                    if (!row) continue;
                    let chn = row.getAttribute("chain");
                    residues.push({
                        chain: chn ? chn : row.getAttribute("keyseqchain"),
                        resno: parseInt(sp.getAttribute("s"))
                    });
                }
                document.body.dispatchEvent(new CustomEvent("selectResidues", {
                    detail: {
                        src: "alignment",
                        struc_id: struc_id,
                        residues: residues
                    }
                }));
            }
            return;
        }
        lastSpanOver = spanId;
        let templateData = $alignTbl.data("templateData");
        if (e.type === "mouseover") {
            if (residueProperties[text.toUpperCase()]) {
                let title = residueProperties[text.toUpperCase()].triplet + " " + seqres + (chain ? " " + chain : "");
                if ($span.attr("rx")) title += "\nResidue missing from template structure!\n";
                if (chainMapping.structures[struc_id]) {
                    for (let i = 0; i < chainMapping.annotations.length; i++) if (chainMapping.annotations[i].active) {
                        if (chainMapping.structures[struc_id][chain][seqres][chainMapping.annotations[i].key] != null) {
                            let val = chainMapping.structures[struc_id][chain][seqres][chainMapping.annotations[i].key];
                            title += "; " + chainMapping.annotations[i].label + ": " + (typeof val === "number" ? val.toFixed(chainMapping.annotations[i].key == "ramachandran" ? 4 : 2) : val);
                        }
                    }
                }
                $span.attr("title", title);
                if ($span[0] !== e.target) e.target.setAttribute("title", title);
            }
            if (chain) document.body.dispatchEvent(new CustomEvent("highlightResidue", {
                detail: {
                    src: "alignment",
                    struc_id: struc_id,
                    chain: chain,
                    resno: parseInt(seqres),
                    mouseDown: $.data($alignTbl, "downchain") != null
                },
                bubbles: true
            }));
        }
        if (editMode) return;
        if (!templateData) return;
        if (e.type === "mousedown") {
            $.data($alignTbl, "downspan", $span);
            $.data($alignTbl, "downchain", seqno + struc_id);
            $(".hoverResidue").removeClass("hoverResidue");
            if (!e.ctrlKey && !e.metaKey && !$span.hasClass("selectResidue")) $(".selectResidue").removeClass("selectResidue");
        } else if (e.type === "mouseover") {
            if ($.data($alignTbl, "downchain")) {
                if (window.getSelection) {
                    windowSelection = window.getSelection();
                    windowSelection.removeAllRanges();
                    range = document.createRange();
                } else if (document.selection) document.selection.empty();
            }
            if (struc_id && chain) {
                document.querySelectorAll(".hoverResidue").forEach(el => el.classList.remove("hoverResidue"));
                if ($.data($alignTbl, "downchain") == null) $span.addClass("hoverResidue");
            }
            if (!struc_id || !chain || $.data($alignTbl, "downchain") !== seqno + struc_id) return;
            if ($.data($alignTbl, "lastMouseDrag") != null && $.data($alignTbl, "lastMouseDrag") === seqno + struc_id + $span.index()) return;
            $.data($alignTbl, "lastMouseDrag", seqno + struc_id + $span.index());
            let $selectThese = $alignTbl.find(`tr[seqno=${seqno}][struc_id="${struc_id}"]`);
            let $startRes = $.data($alignTbl, "downspan");
            let $endRes = $span;
            let startRowIndex = $selectThese.index($startRes.closest("tr"));
            let endRowIndex = $selectThese.index($endRes.closest("tr"));
            if (startRowIndex > endRowIndex || startRowIndex === endRowIndex && $startRes.index() > $endRes.index()) {
                let t2 = startRowIndex;
                $endRes = $startRes;
                $startRes = $span;
                startRowIndex = endRowIndex;
                endRowIndex = t2;
            }
            if ($startRes.index() != $endRes.index() || startRowIndex != endRowIndex) {
                document.querySelectorAll(".hoverResidue").forEach(el => el.classList.remove("hoverResidue"));
                if (!e.ctrlKey && !e.metaKey) document.querySelectorAll(".selectResidue").forEach(el => el.classList.remove("selectResidue"));
                $startRes.addClass("selectResidue");
                $endRes.addClass("selectResidue");
                $selectThese.each(function(i) {
                    if (i < startRowIndex || i > endRowIndex) return;
                    if (i == startRowIndex) $startRes.nextUntil($endRes).addClass("selectResidue"); else if (i == endRowIndex) $endRes.prevUntil($endRes).addClass("selectResidue"); else $(this).find("span").addClass("selectResidue");
                });
            }
            let $selection = $alignTbl.find(".selectResidue");
            if ($selection.length > 0) {
                $("#offscreen_selection").html("");
                $selection.clone().appendTo($("#offscreen_selection"));
                if (window.getSelection) {
                    let windowSelection = window.getSelection();
                    let range = document.createRange();
                    range.setStart($("#offscreen_selection span").get(0), 0);
                    range.setEndAfter($("#offscreen_selection span").get(-1), 0);
                    windowSelection.addRange(range);
                } else if (document.selection) document.selection.empty();
            }
        }
    }
    function delegateEditMode($span, e) {
        e.stopPropagation();
        e.preventDefault();
        if (modalMessageVisible) return;
        let $alignTbl = $span.closest(".alignTbl");
        let $row = $span.closest("tr");
        let seqno = $span.closest("tr").attr("seqno");
        let chain = $row.attr("chain");
        if (!chain) {
            chain = $row.attr("keyseqtype");
            keyseqchain = $row.attr("keyseqchain");
        }
        if (e.type == "mouseover") {
            return;
        }
        if (e.type == "dblclick") {
            if ($span.hasClass("shadowCursor")) {
                if (!$alignTbl.data("mouseDragEdit")) {
                    if ($span.text() == "-") {
                        deleteGap();
                    } else {
                        insertGap();
                    }
                }
            }
        } else if (e.type == "mousedown") {
            if ($span.hasClass("shadowCursor")) {} else {
                let $start = $(".shadowCursor").first();
                if (e.shiftKey && $start) {
                    if ($start.closest("tr").attr("seqno") == seqno) {
                        $(".shadowCursor").removeClass("shadowCursor");
                        let $spans = $alignTbl.find("tr[seqno=" + seqno + "]").find("span");
                        let start = $spans.index($start);
                        let end = $spans.index($span);
                        if (start < end) {
                            $alignTbl.data("startEditRegion", $start);
                            $alignTbl.data("endEditRegion", $span);
                        } else {
                            let tmp = start;
                            start = end;
                            end = tmp;
                            $alignTbl.data("startEditRegion", $span);
                            $alignTbl.data("endEditRegion", $start);
                        }
                        $spans.slice(start, end).each(function() {
                            $(this).addClass("shadowCursor");
                        });
                        $alignTbl.data("mouseDragEdit", false);
                    } else return;
                } else {
                    $(".shadowCursor").removeClass("shadowCursor");
                    $span.addClass("shadowCursor");
                    $alignTbl.data("mouseDragEdit", true);
                    $alignTbl.data("startEditRegion", $span);
                    $alignTbl.data("endEditRegion", $span);
                }
            }
            $(".cursor").removeClass("cursor");
            $span.addClass("cursor");
        } else if (e.type == "mousemove") {
            let sel = window.getSelection ? window.getSelection() : document.selection;
            if (sel) {
                if (sel.removeAllRanges) {
                    sel.removeAllRanges();
                } else if (sel.empty) {
                    sel.empty();
                }
            }
            if (!$alignTbl.data("mouseDragEdit")) return;
            let $startEditRegion = $alignTbl.data("startEditRegion");
            if ($startEditRegion.length == 0) return;
            let shift = 0;
            if (e.pageX > $startEditRegion.offset().left) shift = (e.pageX - $startEditRegion.offset().left) / $startEditRegion.outerWidth() | 0; else shift = (e.pageX - $startEditRegion.offset().left - $startEditRegion.outerWidth()) / $startEditRegion.outerWidth() | 0;
            $(".shadowCursor").removeClass("shadowCursor");
            if (shift > 0) $startEditRegion.nextAll(":lt(" + shift + ")").addClass("shadowCursor"); else if (shift < 0) {
                $startEditRegion.prevAll().slice(0, Math.abs(shift)).each(function() {
                    $(this).addClass("shadowCursor");
                });
            }
            $startEditRegion.addClass("shadowCursor");
        } else if (e.type == "mouseup") {
            let $startEditRegion = $alignTbl.data("startEditRegion");
            if (!$startEditRegion) return;
            let $spans = $alignTbl.find("tr[seqno=" + seqno + "]").find("span");
            currentCursor = {
                seqno: seqno,
                seqidx: $spans.index($span)
            };
            $alignTbl.data("startEditRegion", $("span.shadowCursor").first());
            $alignTbl.data("endEditRegion", $("span.shadowCursor").last());
            $alignTbl.data("mouseDragEdit", false);
        }
    }
    function keydownEdit(e) {
        if (modalMessageVisible) {
            e.preventDefault();
            return;
        }
        let $alignTbl = $("span.cursor").closest(".alignTbl");
        if (e.keyCode == 32) {
            e.preventDefault();
            insertGap();
        } else if (e.keyCode == 8 || e.keyCode == 46) {
            e.preventDefault();
            deleteGap(e.keyCode == 8);
        } else if (e.keyCode == 27) {
            $alignTbl.data("startEditRegion", null);
            $alignTbl.data("endEditRegion", null);
            $(".shadowCursor").removeClass("shadowCursor");
            return;
        } else if ((e.ctrlKey || e.metaKey) && e.keyCode == 90) {
            undoEdit($alignTbl);
            return;
        }
        moveCursor(null, e);
    }
    function addHistory($alignTbl, historyItems) {
        if (historyItems == null || historyItems.length < 1) return false;
        let history = $("body").data("editHistory");
        if (history == null) history = [];
        history.push(historyItems);
        $("body").data("editHistory", history);
        let otherAlignments = $(".alignTbl").not($alignTbl);
        if (otherAlignments.length > 0) {
            updateAllTemplateData(otherAlignments, historyItems, false);
            $(".cursor").removeClass("cursor");
        }
        return true;
    }
    function undoEdit($alignTbl) {
        let history = $("body").data("editHistory");
        if (history == null || history.length < 1) return;
        let groupedItems = history.pop();
        updateAllTemplateData($alignTbl, groupedItems, true, true);
        let otherAlignments = $(".alignTbl").not($alignTbl);
        if (otherAlignments.length > 0) {
            updateAllTemplateData(otherAlignments, groupedItems, true, false);
            $(".cursor").removeClass("cursor");
            $alignTbl.find("tr[seqno=" + currentCursor.seqno + "]").find("span").eq(currentCursor.seqidx).addClass("cursor");
        }
    }
    function updateAllTemplateData($alignTbls, historyItems, undo, master) {
        $alignTbls.each(function() {
            if ($(this).data("templateData") != null) {
                let templates = $(this).data("templateData").templates;
                let groupedItems = historyItems;
                for (let i = 0; i < groupedItems.length; i++) {
                    let historyItems2 = groupedItems[i];
                    for (let h = 0; h < historyItems2.length; h++) for (let j = 0; j < templates.length; j++) {
                        if (templates[j].id == historyItems2[h].template.id) {
                            if (undo) {
                                templates[j]["trg_seq"] = historyItems2[h].trg_seq;
                                templates[j]["tpl_seq"] = historyItems2[h].tpl_seq;
                            } else {
                                templates[j]["trg_seq"] = historyItems2[h].template.trg_seq;
                                templates[j]["tpl_seq"] = historyItems2[h].template.tpl_seq;
                            }
                            if (master) {
                                currentCursor = historyItems2[h].currentCursor;
                                if (i == groupedItems.length - 1) updatePairwiseStats(templates[j]);
                            }
                            break;
                        }
                    }
                }
                if ($(this).is(":visible")) drawAlignment($(this), null, {
                    firstDraw: false
                });
            }
        });
    }
    function updatePairwiseStats(tpl) {
        if ($("#seqid_" + tpl.id).length < 1) return;
        let gapCount = 0, pid = 0;
        for (let i = 0; i < tpl.trg_seq.length; i++) {
            if (tpl.trg_seq.charAt(i) == "-" || tpl.tpl_seq.charAt(i) == "-") gapCount++; else if (tpl.trg_seq.charAt(i) == tpl.tpl_seq.charAt(i)) pid++;
        }
        $("#seqid_" + tpl.id).text((pid / (tpl.trg_seq.length - gapCount) * 100).toFixed(2));
    }
    function moveCursor(direction, e) {
        if (direction == null) {
            if (e.keyCode == 38) direction = "up"; else if (e.keyCode == 40) direction = "down"; else if (e.keyCode == 37) direction = "left"; else if (e.keyCode == 39) direction = "right";
        }
        if (direction == null) return;
        if (e) e.preventDefault();
        let $cursor = $("span.cursor").first();
        let $row = $cursor.closest("tr");
        let seqno = $row.attr("seqno");
        let rowidx = $cursor.index();
        let seqidx = rowidx;
        let $prevAllRow = $row.prevAll("[seqno=" + seqno + "]");
        let $prevAllSpan;
        if ($prevAllRow.length > 0) {
            $prevAllSpan = $prevAllRow.find("span");
            seqidx += $prevAllSpan.length;
        }
        let $next, newseqno = seqno, newseqidx = seqidx;
        if (direction == "up" && $row.index() > 0) {
            let $tmp = $row.prev();
            newseqno = $tmp.attr("seqno");
            $next = $tmp.find("span:eq(" + rowidx + ")");
        } else if (direction == "down" && $row.next().length > 0) {
            let $tmp = $row.next();
            newseqno = $tmp.attr("seqno");
            $next = $tmp.find("span:eq(rowidx)");
            $next = $row.next().find("span");
            if ($next.length > rowidx) $next = $next.eq(rowidx); else $next = $next.last();
        } else if (direction == "left") {
            if ($cursor.prev().length == 0) {
                if ($prevAllRow.length > 0) {
                    $next = $prevAllSpan.last();
                    newseqidx--;
                }
            } else {
                $next = $cursor.prev();
                newseqidx--;
            }
        } else if (direction == "right") {
            if ($cursor.next().length > 0) {
                $next = $cursor.next();
                newseqidx++;
            } else {
                let $nextRow = $row.nextAll("[seqno=" + seqno + "]:eq(0)");
                if ($nextRow.length > 0) {
                    $next = $nextRow.find("span").first();
                    newseqidx++;
                }
            }
        }
        if ($next != null) {
            $cursor.removeClass("cursor");
            $next.addClass("cursor");
            currentCursor = {
                seqno: newseqno,
                seqidx: newseqidx
            };
        }
    }
    function insertGap() {
        let historyItems = [];
        let $cursor = $("span.cursor").first();
        let $row = $cursor.closest("tr");
        let seqno = $row.attr("seqno");
        let $alignTbl = $row.closest(".alignTbl");
        if (!$cursor.hasClass("shadowCursor")) {
            if ($alignTbl.find("span.shadowCursor").length == 0) $cursor.addClass("shadowCursor"); else return historyItems;
        }
        if ($alignTbl.data("templateData") == null) return historyItems;
        let $lastSpan = $alignTbl.find("span.shadowCursor").last();
        let endSpanIdx = $lastSpan.index();
        let singleChar = $lastSpan[0] == $cursor[0];
        let $allrows = $alignTbl.find("tr");
        let endRow = $lastSpan.closest("tr");
        let $editSpan = $();
        let wrapCount = 0;
        $allrows.filter("[seqno=" + seqno + "]").each(function() {
            wrapCount++;
            let $thisSpan = $(this).find("span");
            if ($(this)[0] == endRow[0]) {
                $editSpan = $editSpan.add($thisSpan.slice(0, endSpanIdx + 1));
                return false;
            }
            $editSpan = $editSpan.add($thisSpan);
        });
        let startSpanIdx = $editSpan.index($cursor);
        let strEdit = $editSpan.text();
        let templates = $alignTbl.data("templateData").templates;
        let cursorOnKeyseq = $row.attr("keyseqchain") || $row.attr("keyseqtype");
        let s1, s2 = "";
        let $tmpRows = cursorOnKeyseq ? null : $allrows.filter("[seqno=1]").slice(0, wrapCount);
        let $editRegion = $cursor.nextAll(".shadowCursor").addBack();
        for (let i = 0; i < templates.length; i++) {
            if (templates[i].selected == null || templates[i].selected < 1) continue;
            if (cursorOnKeyseq) {
                s1 = templates[i].trg_seq;
                s2 = templates[i].tpl_seq;
                $tmpRows = $allrows.filter("[seqno=" + (templates[i].selected + 1) + "]").slice(0, wrapCount);
            } else if (templates[i].selected == seqno - 1) {
                s1 = templates[i].tpl_seq;
                s2 = templates[i].trg_seq;
            } else {
                continue;
            }
            strEdit2 = $tmpRows.find("span").slice(0, strEdit.length).text();
            let tmp = "";
            let edited = false;
            let breakout = false;
            for (let j = 0; j < strEdit.length; j++) {
                if (j < startSpanIdx) {
                    tmp += s1.charAt(tmp.length);
                } else {
                    {
                        if (!edited) {
                            tmp += "-";
                            edited = true;
                        } else {
                            if (strEdit.charAt(j) == "-") {
                                tmp += s1.charAt(tmp.length - 1);
                                breakout = true;
                                break;
                            } else tmp += s1.charAt(tmp.length - 1);
                        }
                    }
                }
            }
            if (breakout) s1 = tmp + s1.substring(tmp.length); else if (edited) {
                s1 = tmp + s1.substring(tmp.length - 1);
                if (!singleChar) s2 = s2.substring(0, tmp.length - 1) + "-" + s2.substring(tmp.length - 1);
            }
            if (s1.length > s2.length) s2 += Array(s1.length - s2.length + 1).join("-"); else if (s2.length > s1.length) s1 += Array(s2.length - s1.length + 1).join("-");
            let historyItem = {
                template: templates[i],
                trg_seq: templates[i].trg_seq,
                tpl_seq: templates[i].tpl_seq,
                currentCursor: {
                    seqno: currentCursor.seqno,
                    seqidx: currentCursor.seqidx
                }
            };
            if (cursorOnKeyseq) {
                templates[i].trg_seq = s1;
                templates[i].tpl_seq = s2;
            } else {
                templates[i].trg_seq = s2;
                templates[i].tpl_seq = s1;
            }
            updatePairwiseStats(templates[i]);
            historyItems[historyItems.length] = historyItem;
        }
        if ($editRegion.length == 1) $(".shadowCursor").removeClass("shadowCursor");
        if (historyItems.length > 0) {
            moveCursor("right");
            $alignTbl.data("templateData");
            if (singleChar) {
                $alignTbl.data("startEditRegion", null);
                $alignTbl.data("endEditRegion", null);
            }
            if (addHistory($alignTbl, [ historyItems ])) drawAlignment($alignTbl, null, {
                firstDraw: false
            });
        } else {
            showWarning($cursor, "The pairwise alignment already contains a gap at this position");
        }
    }
    function removeGappedColumns(s1, s2) {
        let tmp1 = "", tmp2 = "";
        for (let j = 0; j < s1.length; j++) {
            if (!(s1.charAt(j) == "-" && s2.charAt(j) == "-")) {
                tmp1 += s1.charAt(j);
                tmp2 += s2.charAt(j);
            }
        }
        return [ tmp1, tmp2 ];
    }
    function showWarning($cursor, message) {
        modalMessageVisible = true;
        $cursor.removeAttr("title").popover({
            html: true,
            content: message,
            container: "body"
        }).popover("show");
        setTimeout(function() {
            $cursor.popover("destroy");
            modalMessageVisible = false;
        }, 2e3);
    }
    function deleteGap(deleteLeft) {
        let historyItems = [];
        let $cursor = $("span.cursor").first();
        let $row = $cursor.closest("tr");
        let seqno = $row.attr("seqno");
        let $alignTbl = $row.closest(".alignTbl");
        let singleChar = $alignTbl.find("span.shadowCursor").length < 2;
        if (deleteLeft) {
            if ($cursor.prev().text() != "-") return;
            if (singleChar) {
                $cursor.removeClass("shadowCursor");
            }
            $cursor = $cursor.prev();
            $cursor.addClass("shadowCursor");
        } else if ($cursor.text() != "-") return;
        if ($alignTbl.data("templateData") == null) {
            return historyItems;
        }
        if (!$cursor.hasClass("shadowCursor")) {
            if ($alignTbl.find("span.shadowCursor").length == 0) $cursor.addClass("shadowCursor"); else return historyItems;
        }
        let $lastSpan = $alignTbl.find("span.shadowCursor").last();
        let endSpanIdx = $lastSpan.index();
        let $allrows = $alignTbl.find("tr");
        let endRow = $lastSpan.closest("tr");
        let $editSpan = $();
        let wrapCount = 0;
        $allrows.filter("[seqno=" + seqno + "]").each(function() {
            wrapCount++;
            let $thisSpan = $(this).find("span");
            if ($(this)[0] == endRow[0]) {
                $editSpan = $editSpan.add($thisSpan.slice(0, endSpanIdx + 1));
                return false;
            }
            $editSpan = $editSpan.add($thisSpan);
        });
        let startSpanIdx = $editSpan.index($cursor);
        let strEdit = $editSpan.text();
        let templates = $alignTbl.data("templateData").templates;
        let cursorOnKeyseq = $row.attr("keyseqchain") || $row.attr("keyseqtype");
        let s1, s2, strEdit2 = "";
        let $tmpRows = cursorOnKeyseq ? null : $allrows.filter("[seqno=1]").slice(0, wrapCount);
        for (let i = 0; i < templates.length; i++) {
            if (templates[i].selected == null || templates[i].selected < 1) continue;
            if (cursorOnKeyseq) {
                s1 = templates[i].trg_seq;
                s2 = templates[i].tpl_seq;
                $tmpRows = $allrows.filter("[seqno=" + (templates[i].selected + 1) + "]").slice(0, wrapCount);
            } else if (templates[i].selected == seqno - 1) {
                s1 = templates[i].tpl_seq;
                s2 = templates[i].trg_seq;
            } else {
                continue;
            }
            strEdit2 = $tmpRows.find("span").slice(0, strEdit.length).text();
            let tmp = "";
            let firstgap = true;
            let nongapfound = false;
            for (let j = 0; j < strEdit.length; j++) {
                if (j == startSpanIdx) {
                    continue;
                } else if (j < startSpanIdx) {
                    {
                        tmp += s1.charAt(tmp.length);
                    }
                } else {
                    if (strEdit.charAt(j) == "-") {
                        if (strEdit2.charAt(j) == "-") continue;
                        if (nongapfound) {
                            gapmoved = true;
                            break;
                        }
                        tmp += "-";
                        firstgap = false;
                    } else {
                        nongapfound = true;
                        if (firstgap) tmp += s1.charAt(tmp.length + 1); else tmp += s1.charAt(tmp.length + 1);
                    }
                }
            }
            if (singleChar) s1 = tmp + s1.substring(tmp.length + 1); else s1 = tmp + "-" + s1.substring(tmp.length + 1);
            if (s1.length > s2.length) s2 += Array(s1.length - s2.length + 1).join("-"); else if (s2.length > s1.length) s1 += Array(s2.length - s1.length + 1).join("-");
            let historyItem = {
                template: templates[i],
                trg_seq: templates[i].trg_seq,
                tpl_seq: templates[i].tpl_seq,
                currentCursor: {
                    seqno: currentCursor.seqno,
                    seqidx: currentCursor.seqidx
                }
            };
            if (cursorOnKeyseq) {
                templates[i].trg_seq = s1;
                templates[i].tpl_seq = s2;
            } else {
                templates[i].trg_seq = s2;
                templates[i].tpl_seq = s1;
            }
            updatePairwiseStats(templates[i]);
            historyItems[historyItems.length] = historyItem;
        }
        if (historyItems.length > 0) {
            if (deleteLeft) moveCursor("left");
            if (singleChar) {
                $alignTbl.data("startEditRegion", null);
                $alignTbl.data("endEditRegion", null);
            }
            if (addHistory($alignTbl, [ historyItems ])) drawAlignment($alignTbl, null, {
                firstDraw: false
            });
        } else {
            showWarning($cursor, "<div>No valid gaps to be deleted in the selected range</div>");
        }
    }
    function alignTemplates(templateData, onlySelected) {
        let aLength = 0;
        let seqs = [], seqIndex = [], template = [], targetIdx = [];
        let target = "";
        $.each(templateData.templates, function(i, tpl) {
            if (!tpl) return;
            if (!onlySelected || tpl.selected) {
                if (!tpl.trg_seq) return;
                template[template.length] = tpl;
                seqs[seqs.length] = "";
                seqIndex[seqs.length] = 0;
                targetIdx[targetIdx.length] = 0;
                aLength = Math.max(tpl.trg_seq.length, aLength);
            }
        });
        template.sort(function(a, b) {
            return a.selected < b.selected ? -1 : a.selected > b.selected ? 1 : 0;
        });
        if (template.length == 0) return;
        let targetUnaligned = null;
        if (templateData.keyseq_sequence) targetUnaligned = templateData.keyseq_sequence; else {
            targetUnaligned = template[0].trg_seq;
            if (!templateData.templates[0].up_res_num) targetUnaligned = targetUnaligned.replace(/-/g, "");
        }
        let trgIdx = 0;
        while (trgIdx < aLength) {
            let gapSize = 0, g = 0;
            for (let i = 0; i < template.length; i++) {
                g = 0;
                while (template[i].trg_seq.charAt(targetIdx[i] + g) == "-") g++;
                gapSize = Math.max(g, gapSize);
            }
            g = 0;
            while (g < gapSize) {
                target += "-";
                g++;
            }
            target += targetUnaligned.charAt(trgIdx++);
            for (let i = 0; i < template.length; i++) {
                if (!template[i].tpl_seq) continue;
                g = 0;
                while (template[i].trg_seq.charAt(targetIdx[i]) == "-") {
                    seqs[i] += template[i].tpl_seq.charAt(targetIdx[i]++);
                    g++;
                }
                while (g < gapSize) {
                    seqs[i] += "-";
                    g++;
                }
                seqs[i] += template[i].tpl_seq.charAt(targetIdx[i]++);
            }
        }
        seqs.target = target;
        seqs.templates = template;
        updatePairwiseStats(template[0]);
        return seqs;
    }
    function getSeqs(templateData, $alignTbl) {
        let seqs = templateData.templates ? alignTemplates(templateData, true) : null;
        if (!seqs) {
            seqs = [];
            seqs.templates = [ "empty" ];
            if (templateData.keyseq_sequence) seqs.target = templateData.keyseq_sequence; else {
                for (let i = 0; i < templateData.templates.length; i++) {
                    if (templateData.templates[i].trg_seq) {
                        seqs.target = templateData.templates[i].trg_seq.replace(/-/g, "");
                        let gapsRemoved = templateData.templates[i].trg_seq.length - seqs.target.length;
                        if (templateData.templates.length === 1 && gapsRemoved > 0) $("#alignmentWarning_" + $alignTbl.attr("id").substring(8)).html(`<div class="alert alert-warning">
                           <button type="button" class="close" data-dismiss="alert">&times;</button>
                           <strong>Note: </strong> ${gapsRemoved} gaps were removed from your target sequence
                         </div>`);
                        break;
                    }
                }
            }
        }
        return seqs;
    }
    let fastDrawTimers = {};
    function drawAlignment($alignTbl, seqs, opts) {
        let $alignNameTbl = getAlignNameTbl($alignTbl);
        $alignTbl.removeData("clustalColours");
        let templateData = $alignTbl.data()["templateData"];
        if (!templateData) templateData = $("#alignTbl").data()["templateData"];
        if (!templateData) return;
        if (!seqs) seqs = getSeqs(templateData, $alignTbl);
        let resCount = 0;
        for (let i = 0; i < seqs.length; i++) {
            let mtch = seqs[i].match(/[a-zA-Z]/g);
            if (mtch) resCount += mtch.length;
        }
        let defaults = {
            cols: 60,
            fastDraw: false,
            textOnly: resCount > 2e4,
            showResiduePosition: true
        };
        if ($alignTbl.data("options")) opts = $.extend($alignTbl.data("options"), opts); else {
            if (!opts) opts = {};
            if (!opts.cols && $alignTbl.attr("cols_cookie") && $.cookie($alignTbl.attr("cols_cookie"))) opts.cols = $.cookie($alignTbl.attr("cols_cookie"));
            opts = $.extend(defaults, opts);
        }
        if (opts.cols == 0) opts.cols = 50;
        opts.cols = parseInt(opts.cols);
        let alid = $alignTbl.attr("id").substring(8);
        let fastDraw = opts.fastDraw;
        if (opts.textOnly) {
            $("#alignTbl" + alid + "_features").remove();
            $alignTbl.off();
        } else if (fastDraw) {
            $("#alignTbl" + alid + "_features").html("");
            let timer = fastDrawTimers["timer" + alid];
            if (timer) clearTimeout(timer);
            opts["fastDraw"] = false;
            timer = setTimeout(function() {
                drawAlignment($alignTbl, seqs);
            }, 10);
            fastDrawTimers["timer" + alid] = timer;
        }
        $alignTbl.data("options", opts);
        let el = document.getElementById($alignTbl.attr("id"));
        while (el.hasChildNodes()) el.removeChild(el.lastChild);
        el = document.getElementById($alignNameTbl.attr("id"));
        while (el && el.hasChildNodes()) el.removeChild(el.lastChild);
        let templateTypeLabel;
        if (templateData["target_name"]) templateTypeLabel = templateData["target_name"]; else if (templateData["type"] == "model-template") templateTypeLabel = templateData["struc_id"]; else templateTypeLabel = templateData["type"].split("-")[0];
        templateTypeLabel = templateTypeLabel.charAt(0).toUpperCase() + templateTypeLabel.substring(1);
        let seqres = templateTypeLabel.toUpperCase() === "SEQRES";
        let targetSeqLength;
        if (templateData.templates !== undefined) targetSeqLength = templateData.templates[0].modelAtomSeqs ? templateData.templates[0].modelAtomSeqs[0].seq.length : seqs.target.length; else targetSeqLength = seqs.target.length;
        let alignhtml = "", namehtml = "";
        if (templateData.nowrap) {
            alignhtml = '<tr><td style="padding-top:0px;">';
            namehtml = '<tr><td style="padding-top:0px;border-right:none">&nbsp;</td></tr>';
            for (let i = seqs.templates[0].trg_offset; i < targetSeqLength + seqs.templates[0].trg_offset; i++) {
                if (i > 0) {
                    if (i > 0 && i % 10 == 0) {
                        alignhtml += '<span style="position:relative;color:#A2A2A2;">&nbsp;' + '<span style="position:absolute;left:0px;top:8px;font-size:.7em;">|</span>' + `<span style="position:absolute;left:0px;top:0px;font-size:.8em;">${i}</span></span>`;
                    } else alignhtml += "<span>&nbsp;</span>";
                }
            }
        }
        alignhtml += "</td></tr>";
        let keyseqResmap = {};
        let strucResmap = {};
        for (let alIndex = 0; alIndex < targetSeqLength; alIndex += opts.cols) {
            let seqno = 1;
            for (let j = 0; j < templateData["keyseq_chain_name"].length; j++) {
                let resmap = [];
                let targetSeq;
                if (templateData.templates !== undefined) targetSeq = templateData.templates[0].modelAtomSeqs ? templateData.templates[0].modelAtomSeqs[j].seq : seqs.target; else targetSeq = seqs.target;
                let keyseqChnNm = templateData["keyseq_chain_name"][j];
                namehtml += '<tr seqno="' + seqno + '" ' + parseDesc(templateData.keyseq_description) + "><td" + (alIndex > 0 && j == 0 || templateData.charts ? ' style="position:relative; padding-top:10px" ' : "") + ">" + templateTypeLabel + (templateData["keyseq_chain_name"].length > 1 ? ":" + keyseqChnNm : "") + "</td></tr>";
                let trg_offset = seqs.templates[0].trg_offset ? seqs.templates[0].trg_offset : 1;
                if (templateData.templates !== undefined && templateData.templates[0].modelAtomSeqs) trg_offset = templateData.templates[0].modelAtomSeqs[j].trg_offset;
                alignhtml += `<tr seqno="${seqno}" keyseqtype="${templateData["type"]}"`;
                if (templateData.struc_id && (!opts.textOnly || !opts.fastDraw)) alignhtml += ` struc_id="${templateData.struc_id}" keyseqchain="${keyseqChnNm}"`;
                alignhtml += `><td${alIndex > 0 && j == 0 || templateData.charts ? ' style="padding-top:10px" ' : ``}>`;
                if (opts.fastDraw) alignhtml += fastSpan(targetSeq.substring(alIndex, alIndex + opts.cols)); else if (opts.textOnly) alignhtml += `<span>${targetSeq.substring(alIndex, alIndex + opts.cols)}</span>`; else {
                    alignhtml += addSpan(targetSeq.substring(alIndex, alIndex + opts.cols), {
                        alIndex: alIndex,
                        pos: countNonGaps(targetSeq.substring(0, alIndex)),
                        offset: trg_offset,
                        annotation: templateData.keyseq_annotation ? templateData.keyseq_annotation[keyseqChnNm] : null,
                        seqres: seqres,
                        up_res_num: templateData.templates && templateData.templates[j] && templateData.templates[j].up_res_num ? templateData.templates[j].up_res_num : null,
                        resmap: resmap,
                        keyseq: true
                    });
                    if (!keyseqResmap[keyseqChnNm]) keyseqResmap[keyseqChnNm] = [];
                    keyseqResmap[keyseqChnNm] = keyseqResmap[keyseqChnNm].concat(resmap);
                }
                alignhtml += "</td></tr>";
                seqno += 1;
            }
            for (let i = 0; i < seqs.length; i++) {
                let resmap = templateData["type"] != "model-template" && seqs.templates[i].struc_id ? [] : null;
                let targetSeq = templateData.templates[0].modelAtomSeqs ? templateData.templates[0].modelAtomSeqs[i].seq : seqs.target;
                if (!seqs.templates[i].tpl_seq) {
                    seqno += 1;
                    continue;
                }
                namehtml += `<tr seqno="${seqno}" chain="${seqs.templates[i].chain}${parseDesc(seqs.templates[i].description)}"><td nowrap align="left">${seqs.templates[i].qname}</td></tr>`;
                alignhtml += `<tr seqno="${seqno}" ` + ((!opts.textOnly || !opts.fastDraw) && (templateData["type"] != "model-template" && seqs.templates[i].struc_id) ? ' struc_id="' + seqs.templates[i].struc_id + '"' : "") + ' qname="' + seqs.templates[i].qname + '"' + ' chain="' + seqs.templates[i].chain + '"><td nowrap align="left">';
                if (opts.fastDraw) alignhtml += fastSpan(seqs[i].substring(alIndex, alIndex + opts.cols)); else if (opts.textOnly) alignhtml += "<span>" + seqs[i].substring(alIndex, alIndex + opts.cols) + "</span>"; else {
                    alignhtml += addSpan(seqs[i].substring(alIndex, alIndex + opts.cols), {
                        alIndex: alIndex,
                        pos: countNonGaps(seqs[i].substring(0, alIndex)),
                        offset: seqs.templates[i].offset ? seqs.templates[i].offset + 1 : 1,
                        seqres: seqres,
                        annotation: seqs.templates[i].annotation,
                        targetSeq: targetSeq.substring(alIndex, alIndex + opts.cols),
                        pdb_res_num: templateData.templates[i].pdb_res_num ? templateData.templates[i].pdb_res_num.slice(alIndex, alIndex + opts.cols) : null,
                        resmap: resmap,
                        keyseq: false
                    });
                    if (resmap) {
                        if (!strucResmap[seqs.templates[i].struc_id]) strucResmap[seqs.templates[i].struc_id] = {};
                        if (!strucResmap[seqs.templates[i].struc_id][seqs.templates[i].chain]) strucResmap[seqs.templates[i].struc_id][seqs.templates[i].chain] = [];
                        strucResmap[seqs.templates[i].struc_id][seqs.templates[i].chain] = strucResmap[seqs.templates[i].struc_id][seqs.templates[i].chain].concat(resmap);
                    }
                }
                alignhtml += "</td></tr>";
                seqno += 1;
            }
        }
        if (!chainMapping.alignments.includes($alignTbl.attr("id"))) {
            if (templateData["modelLabel"]) {
                if (templateData["type"] == "model-template") {
                    for (let kchn in keyseqResmap) {
                        if (!chainMapping.chainMeta[templateData.struc_id]) chainMapping.chainMeta[templateData.struc_id] = {};
                        if (!chainMapping.chainMeta[templateData.struc_id][kchn]) chainMapping.chainMeta[templateData.struc_id][kchn] = {
                            display: templateData["modelLabel"]
                        };
                    }
                } else {
                    for (let i = 0; i < templateData.templates.length; i++) {
                        if (!chainMapping.chainMeta[templateData.templates[i].struc_id]) chainMapping.chainMeta[templateData.templates[i].struc_id] = {};
                        if (!chainMapping.chainMeta[templateData.templates[i].struc_id][templateData.templates[i].chain]) chainMapping.chainMeta[templateData.templates[i].struc_id][templateData.templates[i].chain] = {
                            display: templateData["modelLabel"]
                        };
                    }
                }
            }
            chainMapping.alignments.push($alignTbl.attr("id"));
            if (templateData["type"] === "model-template") {
                for (let kchn in keyseqResmap) {
                    if (templateData.md5) {
                        if (!chainMapping.targets[templateData.md5]) chainMapping.targets[templateData.md5] = {};
                        if (!chainMapping.targets[templateData.md5][templateData.struc_id]) chainMapping.targets[templateData.md5][templateData.struc_id] = [];
                    }
                    if (!chainMapping.structures[templateData.struc_id]) chainMapping.structures[templateData.struc_id] = {};
                    if (!chainMapping.structures[templateData.struc_id][kchn]) chainMapping.structures[templateData.struc_id][kchn] = [];
                    for (let i = 0; i < keyseqResmap[kchn].length; i++) {
                        if (!keyseqResmap[kchn][i]) continue;
                        let alreadyMapped = false;
                        let res = {};
                        if (chainMapping.structures[templateData.struc_id][kchn][keyseqResmap[kchn][i][1]]) {
                            alreadyMapped = true;
                            res = chainMapping.structures[templateData.struc_id][kchn][keyseqResmap[kchn][i][1]];
                        }
                        res.trgno = keyseqResmap[kchn][i][1];
                        res.trgname = keyseqResmap[kchn][i][0];
                        res.chn = kchn;
                        res.resno = keyseqResmap[kchn][i][1];
                        res.resname = keyseqResmap[kchn][i][0];
                        if (keyseqResmap[kchn][i].length == 3) {
                            if (keyseqResmap[kchn][i][2].qmean) res.qm = keyseqResmap[kchn][i][2].qmean;
                            if (keyseqResmap[kchn][i][2].qmeanBrane) res.qmb = keyseqResmap[kchn][i][2].qmeanbrane;
                            if (keyseqResmap[kchn][i][2].consistency) res.c = keyseqResmap[kchn][i][2].consistency;
                        }
                        if (!alreadyMapped) {
                            if (templateData.md5) {
                                if (!chainMapping.targets[templateData.md5][templateData.struc_id][keyseqResmap[kchn][i][1]]) chainMapping.targets[templateData.md5][templateData.struc_id][keyseqResmap[kchn][i][1]] = [];
                                chainMapping.targets[templateData.md5][templateData.struc_id][keyseqResmap[kchn][i][1]].push(res);
                            }
                            chainMapping.structures[templateData.struc_id][kchn][keyseqResmap[kchn][i][1]] = res;
                        }
                    }
                }
            } else {
                for (let kchn in keyseqResmap) {
                    for (let i = 0; i < keyseqResmap[kchn].length; i++) {
                        for (let struc_id in strucResmap) {
                            if (!chainMapping.targets[templateData.md5]) chainMapping.targets[templateData.md5] = {};
                            if (!chainMapping.targets[templateData.md5][struc_id]) chainMapping.targets[templateData.md5][struc_id] = [];
                            if (!chainMapping.structures[struc_id]) chainMapping.structures[struc_id] = {};
                            for (let chn in strucResmap[struc_id]) {
                                if (!chainMapping.structures[struc_id][chn]) chainMapping.structures[struc_id][chn] = [];
                                if (strucResmap[struc_id][chn][i] == null) continue;
                                let alreadyMapped = false;
                                let res = {};
                                if (chainMapping.structures[struc_id][chn][strucResmap[struc_id][chn][i][1]]) {
                                    alreadyMapped = true;
                                    res = chainMapping.structures[struc_id][chn][strucResmap[struc_id][chn][i][1]];
                                }
                                res.chn = chn;
                                [ res.resname, res.resno ] = strucResmap[struc_id][chn][i];
                                if (keyseqResmap[kchn][i]) {
                                    res.trgname = keyseqResmap[kchn][i][0];
                                    res.trgno = keyseqResmap[kchn][i][1];
                                }
                                if (strucResmap[struc_id][chn][i].length == 3) {
                                    if (strucResmap[struc_id][chn][i][2].qmean) res.qm = strucResmap[struc_id][chn][i][2].qmean;
                                    if (strucResmap[struc_id][chn][i][2].qmeanbrane) res.qmb = strucResmap[struc_id][chn][i][2].qmean;
                                    if (strucResmap[struc_id][chn][i][2].consistency) res.c = strucResmap[struc_id][chn][i][2].consistency;
                                }
                                if (!alreadyMapped) {
                                    if (keyseqResmap[kchn][i]) {
                                        if (!chainMapping.targets[templateData.md5][struc_id][keyseqResmap[kchn][i][1]]) chainMapping.targets[templateData.md5][struc_id][keyseqResmap[kchn][i][1]] = [];
                                        chainMapping.targets[templateData.md5][struc_id][keyseqResmap[kchn][i][1]].push(res);
                                    }
                                    chainMapping.structures[struc_id][chn][strucResmap[struc_id][chn][i][1]] = res;
                                }
                            }
                        }
                    }
                }
            }
        }
        let sstype = $.cookie("secStruct");
        if (!sstype || sstype !== "none") $alignTbl.addClass("ss");
        $alignNameTbl.last().append(namehtml);
        $alignTbl.last().append(alignhtml);
        let padding;
        let firstSeqno = null;
        if (!templateData.nowrap && opts.showResiduePosition) $alignTbl.find("tr:visible").each(function(i, obj) {
            if (firstSeqno && obj.getAttribute("seqno") == firstSeqno) padding = `padding-top:${templateData.barChartHeight + 10}px;`; else padding = "";
            let s = $(this).find("span[s]").last().attr("s");
            if (s) $(this).append(`<td style="padding-left:8px; font-size:.8em; text-align:right; ${padding}">${s}</td>`);
        });
        if (!fastDraw) {
            if (chainMapping.defaultColourScheme) {
                $(`#${chainMapping.defaultColourScheme}Button`).prop("checked", true);
                colourAlignment(chainMapping.defaultColourScheme, $alignTbl);
            } else colourAlignment($.cookie("colourScheme"), $alignTbl);
            seqs = null;
            highlightSelectedResidues();
            if (!opts.firstDraw) drawAnnotation($alignTbl);
        }
    }
    function countNonGaps(str) {
        let m = str.match(/[^-]/g);
        return m ? m.length : 0;
    }
    function parseDesc(desc) {
        return desc ? ` title="${desc}"` : "";
    }
    function fastSpan(txt) {
        let html = "";
        for (let i = 0; i < txt.length; i++) html += `<span>${txt.charAt(i)}</span>`;
        return html;
    }
    function addSpan(txt, args) {
        if (!txt) return "";
        let pos = args.pos;
        let alIndex = args.alIndex;
        let seqres = args.seqres;
        let checktargetMismatch = args.targetSeq != null;
        let spanhtml = [];
        let lastUp = null;
        for (let i = 0; i < txt.length; i++) {
            let res = txt.charAt(i);
            let resno = null;
            let annoIdx = null;
            if (res == "-") {
                spanhtml.push('<span nm="true">-</span>');
                if (args.resmap) args.resmap.push(null);
                continue;
            }
            if (args.pdb_res_num) {
                resno = args.pdb_res_num[i];
                annoIdx = resno - args.offset;
                let prevPDBResNum = args.pdb_res_num[i - 1];
                let thisPDBResNum = args.pdb_res_num[i];
                let nextPDBResNum = args.pdb_res_num[i + 1];
                if (prevPDBResNum != null && prevPDBResNum != thisPDBResNum - 1 && prevPDBResNum != thisPDBResNum) {
                    if (i > 0 && txt.charAt(i - 1) == "-") spanhtml[i - 1] = spanhtml[i - 1].substring(0, spanhtml[i - 1].length - 9) + ' class="insert_r">-</span>';
                    spanhtml.push(`<span s="${resno}" r="${res.toUpperCase()}" ${checktargetMismatch && args.targetSeq.charAt(i) != res ? ' nm="true"' : ""} class="insert_l">${res}</span>`);
                } else if (nextPDBResNum && thisPDBResNum != nextPDBResNum - 1 && nextPDBResNum != thisPDBResNum) {
                    spanhtml.push(`<span s="${resno}" r="${res.toUpperCase()}" ${checktargetMismatch && args.targetSeq.charAt(i) != res ? ' nm="true"' : ""} class="insert_r">${res}</span>`);
                    if (i < txt.length - 2 && txt.charAt(i + 1) == "-") {
                        spanhtml.push('<span nm="true" class="insert_l">-</span>');
                        i++;
                    }
                } else spanhtml.push(`<span s="${resno}" r="${res.toUpperCase()}" ${checktargetMismatch && args.targetSeq.charAt(i) != res ? ' nm="true"' : ""}>${res}</span>`);
            } else if (args.up_res_num) {
                annoIdx = alIndex + i;
                let prevUpResNum = args.up_res_num[alIndex + i - 1];
                let thisUpResNum = args.up_res_num[alIndex + i];
                let nextUpResNum = args.up_res_num[alIndex + i + 1];
                if (thisUpResNum == null) {
                    lastUp += 1;
                    thisUpResNum = lastUp;
                }
                if (nextUpResNum == null) {
                    prevUpResNum = thisUpResNum - 1;
                    nextUpResNum = thisUpResNum + 1;
                }
                resno = thisUpResNum;
                spanhtml.push(`<span ${alIndex + i > 0 && prevUpResNum != thisUpResNum - 1 ? 'class="insert_l"' : ""} ${alIndex + i < alIndex + txt.length - 1 && thisUpResNum != nextUpResNum - 1 ? 'class="insert_r"' : ""} s="${thisUpResNum}" r="${res.toUpperCase()}" ${checktargetMismatch && args.targetSeq.charAt(i) != res ? ' nm="true"' : ""}>${res}</span>`);
                lastUp = thisUpResNum;
            } else if (seqres) {
                annoIdx = alIndex + i;
                resno = alIndex + i + args.offset;
                spanhtml.push(`<span s="${resno}" r="${res.toUpperCase()}" ${checktargetMismatch && args.targetSeq.charAt(i) != res ? ' nm="true"' : ""}>${res}</span>`);
            } else {
                annoIdx = pos;
                resno = pos + args.offset;
                let resMissing = false;
                if (args.annotation && args.annotation.dssp) {
                    if (args.annotation.dssp_string) {
                        if (args.annotation.dssp_string.charAt(resno - 1) === "X" && args.annotation.dssp_string.search(/[^X\-]/) > -1) resMissing = true;
                    } else {
                        if (args.annotation.dssp.charAt(resno - 1) === "X" && args.annotation.dssp.search(/[^X\-]/) > -1) resMissing = true;
                    }
                }
                spanhtml.push(`<span s="${resno}" r="${res.toUpperCase()}"${checktargetMismatch && args.targetSeq.charAt(i) != res ? ' nm="true"' : ""}${resMissing ? ' rx="true"' : ""}>${res}</span>`);
                pos++;
            }
            if (args.resmap) {
                let mappedRes = [ res.toUpperCase(), resno ];
                if (args.annotation) {
                    let anno = {};
                    for (let key in args.annotation) {
                        if (args.annotation[key] && args.annotation[key][annoIdx] != null) anno[key] = args.annotation[key][annoIdx];
                        if (args.keyseq && key === "qmean" && args.annotation.qmean[annoIdx] == null) spanhtml[spanhtml.length - 1] = `<span nores="true" ${spanhtml[spanhtml.length - 1].substring(6)}`;
                    }
                    mappedRes.push(anno);
                }
                args.resmap.push(mappedRes);
            }
        }
        return spanhtml.join("");
    }
    function addRes(clst, obj, selectable) {
        if (clst[selectable][obj]) clst[selectable][obj]++; else clst[selectable][obj] = 1;
    }
    function calculateClustalColours(seqs) {
        let gapCount, res, clst = [];
        for (let resIndex = 0; resIndex < seqs[0].length; resIndex++) {
            clst[resIndex] = {};
            gapCount = 0;
            for (let seqIndex = 0; seqIndex < seqs.length; seqIndex++) {
                res = seqs[seqIndex].charAt(resIndex).toUpperCase();
                if (res == "-") {
                    gapCount++;
                    continue;
                }
                addRes(clst, res, resIndex);
                for (let group of [ "QE", "ED", "WLVIMAFCYHP", "KR", "TS" ]) if (group.indexOf(res) > -1) addRes(clst, group, resIndex);
            }
            clst[resIndex].gapCount = gapCount;
            let keygap = {};
            for (let key in clst[resIndex]) {
                keygap[key + "_nogap"] = (clst[resIndex][key] / seqs.length).toFixed(2);
                clst[resIndex][key] = (clst[resIndex][key] / (seqs.length - gapCount)).toFixed(2);
            }
            for (let key in keygap) {
                clst[resIndex][key] = keygap[key];
            }
        }
        return clst;
    }
    const residueProperties = {
        A: {
            name: "Alanine",
            triplet: "ALA",
            hydrophobic: .7,
            size: .11
        },
        C: {
            name: "Cysteine",
            triplet: "CYS",
            hydrophobic: .78,
            size: .35
        },
        D: {
            name: "Aspartate",
            triplet: "ASP",
            hydrophobic: .11,
            size: .45
        },
        E: {
            name: "Glutamate",
            triplet: "GLU",
            hydrophobic: .11,
            size: .56
        },
        F: {
            name: "Phenylalanine",
            triplet: "PHE",
            hydrophobic: .81,
            size: .7
        },
        G: {
            name: "Glycine",
            triplet: "GLY",
            hydrophobic: .46,
            size: 0
        },
        H: {
            name: "Histidine",
            triplet: "HIS",
            hydrophobic: .14,
            size: .62
        },
        I: {
            name: "Isoleucine",
            triplet: "ILE",
            hydrophobic: 1,
            size: .43
        },
        K: {
            name: "Lysine",
            triplet: "LYS",
            hydrophobic: .07,
            size: .55
        },
        L: {
            name: "Leucine",
            triplet: "LEU",
            hydrophobic: .92,
            size: .43
        },
        M: {
            name: "Methionine",
            triplet: "MET",
            hydrophobic: .71,
            size: .57
        },
        N: {
            name: "Asparagine",
            triplet: "ASN",
            hydrophobic: .11,
            size: .44
        },
        P: {
            name: "Proline",
            triplet: "PRO",
            hydrophobic: .32,
            size: .31
        },
        Q: {
            name: "Glutamine",
            triplet: "GLN",
            hydrophobic: .11,
            size: .55
        },
        R: {
            name: "Arginine",
            triplet: "ARG",
            hydrophobic: 0,
            size: .77
        },
        S: {
            name: "Serine",
            triplet: "SER",
            hydrophobic: .41,
            size: .23
        },
        T: {
            name: "Threonine",
            triplet: "THR",
            hydrophobic: .42,
            size: .34
        },
        V: {
            name: "Valine",
            triplet: "VAL",
            hydrophobic: .97,
            size: .32
        },
        W: {
            name: "Tryptophan",
            triplet: "TRP",
            hydrophobic: .4,
            size: 1
        },
        Y: {
            name: "Tyrosine",
            triplet: "TYR",
            hydrophobic: .36,
            size: .82
        },
        B: {
            name: "Aspartate or Asparagine",
            triplet: "ASX"
        },
        Z: {
            name: "Glutamate or Glutamine",
            triplet: "GLX"
        },
        J: {
            name: "Leucine or Isoleucine",
            triplet: "XLE"
        },
        X: {
            name: "Unknown",
            triplet: "UNK"
        },
        "?": {
            name: "Unknown",
            triplet: "UNK"
        },
        hydrophobic: "LIFWVM",
        polar: "STNQ",
        small: "TDN",
        large: "WYRFH",
        aliphatic: "ILV",
        charged: "HKRED",
        positive: "HKR",
        negative: "ED",
        tiny: "AGS",
        aromatic: "FYWH",
        proline: "P",
        serThr: "ST"
    };
    function colourAlignment(col, $alignTbl) {
        let $spans = $alignTbl.find("span");
        if ($spans.length === 0) return;
        $.each($spans, function(i, span) {
            if (span.style.backgroundColor) span.style.backgroundColor = "";
        });
        if (!$alignTbl.data("features")) $alignTbl.removeData("features");
        if (col == null) col = "rainbow";
        if (([ "clustal", "hydrophobic", "size", "charged", "cysteine" ].indexOf(col) > -1 || residueProperties[col]) && $alignTbl.data("templateData").is_nucleotide) col = "base";
        if ($alignTbl.data("features")) {
            for (let struc_id in $alignTbl.data("features")) {
                let seqno = 0;
                let $rows = $alignTbl.find("tr[struc_id]");
                $rows.filter(`[struc_id="${struc_id}"]`).each(function(r, row) {
                    if (seqno > parseInt(row.getAttribute("seqno"))) return false;
                    seqno = parseInt(row.getAttribute("seqno"));
                    let chn = row.getAttribute("chain");
                    if (!chn) chn = row.getAttribute("keyseqchain");
                    let $spans = $rows.filter("tr[seqno=" + seqno + "]").find(" span[s]");
                    for (let i = 0; i < $alignTbl.data("features")[struc_id].length; i++) {
                        let ft = $alignTbl.data("features")[struc_id][i];
                        let rgb = ft.rgb.split(",").map(Number);
                        if (ft.frmChn != ft.toChn) {
                            console.error("THIS IS AN ERROR " + ft);
                            continue;
                        }
                        if (ft.frmChn.indexOf(chn) == -1) continue;
                        for (let j = ft.resStart; j < ft.resEnd + 1; j++) if (chainMapping.structures[struc_id][chn][j]) chainMapping.structures[struc_id][chn][j].rgb = rgb;
                    }
                    $spans.each(function(j, span) {
                        let s = parseInt(span.getAttribute("s"));
                        if (chainMapping.structures[struc_id][chn][s]) span.style.backgroundColor = "rgba(" + chainMapping.structures[struc_id][chn][s].rgb + ",.6)";
                    });
                });
            }
        } else if (col == "chain" || col == "uniqueChain" || col == "template") {
            let templateData = null;
            if (typeof globalTemplateData != "undefined") {
                templateData = globalTemplateData;
                if (!globalTemplateData["chainColoursAssigned"]) {
                    for (let i = 0; i < globalTemplateData.length; i++) {
                        let tpl = globalTemplateData[i];
                        tpl["chainColours"] = {
                            chain: [],
                            unique: [],
                            template: colourCycle[i % colourCycle.length]
                        };
                        let chattr = tpl.chain_name ? "chain_name" : "chain";
                        for (let t = 0; t < tpl[chattr].length; t++) {
                            tpl["chainColours"]["unique"][tpl[chattr][t].length > 1 ? tpl[chattr][t].join("") : tpl[chattr][t]] = colourCycle[uindex];
                            uindex++;
                            uindex %= 12;
                            for (let c = 0; c < tpl[chattr][t].length; c++) {
                                tpl["chainColours"]["chain"][tpl[chattr][t][c]] = colourCycle[cindex];
                                cindex++;
                                cindex %= colourCycle.length;
                            }
                        }
                    }
                    globalTemplateData["chainColoursAssigned"] = true;
                }
            } else {
                templateData = $alignTbl.data("templateData").templates;
                $(".alignTbl").each(function() {
                    let tdata = $(this).data("templateData");
                    if (!tdata || !tdata.templates) return;
                    if (tdata["chainColoursAssigned"]) return;
                    for (let t = 0; t < tdata.templates.length; t++) {
                        tdata.templates[t]["chainColours"] = {
                            chain: [],
                            unique: []
                        };
                        if (tdata.type == "model-template") {
                            for (let i = 0; i < tdata.keyseq_chain_name.length; i++) {
                                tdata.templates[t]["chainColours"]["chain"][tdata.keyseq_chain_name[i]] = colourCycle[cindex];
                                tdata.templates[t]["chainColours"]["unique"][tdata.keyseq_chain_name[i]] = colourCycle[uindex];
                                cindex++;
                                cindex %= colourCycle.length;
                            }
                        } else {
                            tdata.templates[t]["chainColours"]["chain"][tdata.templates[t]["chain"]] = colourCycle[cindex];
                            tdata.templates[t]["chainColours"]["unique"][tdata.templates[t]["chain"]] = colourCycle[uindex];
                            cindex++;
                            cindex %= colourCycle.length;
                        }
                    }
                    uindex++;
                    uindex %= colourCycle.length;
                    tdata["chainColoursAssigned"] = true;
                });
            }
            if (!templateData) {
                $alignTbl.find("span").each(function(j, span) {
                    span.style.backgroundColor = "rgba(" + colourCycle[$(".alignTbl").index($alignTbl) % colourCycle.length] + ",.6)";
                });
            } else {
                $.each($alignTbl.find("tr[struc_id]"), function(j, row) {
                    let strucId = row.getAttribute("struc_id");
                    let chn = row.getAttribute("keyseqchain");
                    if (!chn) chn = row.getAttribute("chain");
                    if (!chn) return false;
                    let $span = $(this).find("span[s]");
                    let tpl = null;
                    for (let i = 0; i < templateData.length; i++) {
                        if (templateData[i]["struc_id"] == strucId) {
                            tpl = templateData[i];
                            let rgb;
                            if (col == "chain") rgb = tpl["chainColours"]["chain"][chn]; else if (col == "template") rgb = tpl["chainColours"]["template"]; else {
                                for (let chns in tpl["chainColours"]["unique"]) {
                                    if (chns.indexOf(chn) > -1) {
                                        rgb = tpl["chainColours"]["unique"][chns];
                                        break;
                                    }
                                }
                            }
                            if (!rgb) continue;
                            $span.each(function(j, span) {
                                span.style.backgroundColor = "rgb(" + rgb + ",.6)";
                                if (strucId) {
                                    let s = parseInt(span.getAttribute("s"));
                                    if (chainMapping.structures[strucId] && chainMapping.structures[strucId][chn] && chainMapping.structures[strucId][chn][s]) chainMapping.structures[strucId][chn][s].rgb = rgb;
                                }
                            });
                        }
                    }
                });
            }
        } else if (col === "rainbow") {
            let modelTemplate = false;
            let seqIndex = 0;
            let $rows = $alignTbl.find("tr");
            $rows.each(function(i, elm) {
                let strucId = elm.getAttribute("struc_id");
                let chain = elm.getAttribute("chain");
                if (!chain) chain = elm.getAttribute("keyseqchain");
                if (i == 0) modelTemplate = elm.getAttribute("keyseqtype") == "model-template";
                let seqno = parseInt(elm.getAttribute("seqno"));
                if (seqno <= seqIndex) return false;
                seqIndex = seqno;
                if (modelTemplate && elm.getAttribute("keyseqtype") == null) return;
                let $span = $rows.filter("[seqno=" + seqIndex + "]").find("span[s]:not([nores])");
                for (let j = 0; j < $span.length; j++) {
                    let s = $span[j].getAttribute("s");
                    if (s == null) continue;
                    let val = j / $span.length * (1.5 * Math.PI - Math.PI / 2) + Math.PI / 2;
                    let rgb = [ Math.round(255 / 2 * (Math.sin(val + Math.PI) + 1)), Math.round(255 * Math.sin(val - Math.PI / 2)), Math.round(255 / 2 * (Math.sin(val) + 1)) ];
                    $span[j].style.backgroundColor = `rgba(${rgb},.6)`;
                    if (strucId && chainMapping.structures[strucId] && chainMapping.structures[strucId][chain]) {
                        s = parseInt(s);
                        if (chainMapping.structures[strucId][chain][s]) chainMapping.structures[strucId][chain][s].rgb = rgb;
                    }
                }
            });
        } else if (col === "qmean_local") {
            let qm = "qm";
            if (typeof qmeanbraneActive !== "undefined" && qmeanbraneActive) qm = "qmb";
            $alignTbl.find("tr[struc_id]").each(function() {
                let strucId = $(this).attr("struc_id");
                let chain = $(this).attr("chain");
                if (!chain) chain = $(this).attr("keyseqchain");
                if (!chain || !chainMapping.structures[strucId][chain]) return true;
                let rowSpan = $(this).find("span[s]");
                let belowOne = true;
                let hasBfactors = false;
                for (let chn in chainMapping.structures[strucId]) {
                    for (let res in chainMapping.structures[strucId][chn]) {
                        if (chainMapping.structures[strucId][chn][res]) if (chainMapping.structures[strucId][chn][res][qm]) {
                            valExists = true;
                            if (chainMapping.structures[strucId][chn][res][qm] > 1) {
                                belowOne = false;
                                break;
                            }
                        } else if (chainMapping.structures[strucId][chn][res].bfactor) hasBfactors = true;
                    }
                    if (!belowOne) break;
                }
                if (hasBfactors) {
                    colourAnnotation($alignTbl, "bfactor_range", $(this));
                    return true;
                }
                $.each(rowSpan, function(i, span) {
                    let s = parseInt(span.getAttribute("s"));
                    if (!chainMapping.structures[strucId][chain][s]) return true;
                    let q = chainMapping.structures[strucId][chain][s][qm];
                    if (q == null) return true;
                    if (!belowOne) q /= 100;
                    let rgba = qmeanLocalColour(q, .6);
                    span.style.backgroundColor = rgba;
                    let rgb = rgba.substring(5).split(",").slice(0, 3).map(Number);
                    chainMapping.structures[strucId][chain][s].rgb = rgb;
                });
            });
        } else if (col == "confClass") {
            $alignTbl.find("tr[struc_id]").each(function() {
                let strucId = $(this).attr("struc_id");
                let chain = $(this).attr("chain");
                if (!chain) chain = $(this).attr("keyseqchain");
                if (!chain || !chainMapping.structures[strucId][chain]) return true;
                let rowSpan = $(this).find("span[s]");
                let belowOne = true;
                let hasBfactors = false;
                for (let chn in chainMapping.structures[strucId]) {
                    for (let res in chainMapping.structures[strucId][chn]) {
                        if (chainMapping.structures[strucId][chn][res]) if (chainMapping.structures[strucId][chn][res].qm) {
                            valExists = true;
                            if (chainMapping.structures[strucId][chn][res].qm > 1) {
                                belowOne = false;
                                break;
                            }
                        } else if (chainMapping.structures[strucId][chn][res].bfactor) hasBfactors = true;
                    }
                    if (!belowOne) break;
                }
                if (hasBfactors) {
                    colourAnnotation($alignTbl, "bfactor_range", $(this));
                    return true;
                }
                $.each(rowSpan, function(i, span) {
                    let s = parseInt(span.getAttribute("s"));
                    if (!chainMapping.structures[strucId][chain][s]) return true;
                    let q = chainMapping.structures[strucId][chain][s].qm;
                    if (!q) return true;
                    if (belowOne) q *= 100;
                    if (q > 90) rgb = [ 0, 83, 214 ]; else if (q > 70) rgb = [ 101, 203, 243 ]; else if (q > 50) rgb = [ 255, 219, 19 ]; else rgb = [ 255, 125, 69 ];
                    span.style.backgroundColor = "rgba(" + rgb + ",.6)";
                    chainMapping.structures[strucId][chain][s].rgb = rgb;
                });
            });
        } else if (col.indexOf("consistency") > -1) {
            let mn = 0;
            let mx = 1;
            if (col != "consistency") {
                mn = 999;
                mx = -99;
                $alignTbl.find("tr[struc_id]").each(function() {
                    let strucId = $(this).attr("struc_id");
                    let chainName = $(this).attr("chain");
                    if (!chainName) chainName = $(this).attr("keyseqchain");
                    if (!chainName || !chainMapping.structures[strucId][chainName]) return true;
                    for (let i = 0; i < chainMapping.structures[strucId][chainName].length; i++) {
                        if (!chainMapping.structures[strucId][chainName][i] || !chainMapping.structures[strucId][chainName][i].c) continue;
                        mn = Math.min(mn, chainMapping.structures[strucId][chainName][i].c);
                        mx = Math.max(mx, chainMapping.structures[strucId][chainName][i].c);
                    }
                });
            }
            $alignTbl.find("tr[struc_id]").each(function() {
                let strucId = $(this).attr("struc_id");
                let chainName = $(this).attr("chain");
                if (!chainName) chainName = $(this).attr("keyseqchain");
                if (!chainName || !chainMapping.structures[strucId][chainName]) return true;
                let rowSpan = $(this).find("span[s]").toArray();
                for (let i = 0; i < rowSpan.length; i++) {
                    let rgb = [ 127, 195, 45 ];
                    let resno = parseInt(rowSpan[i].getAttribute("s"));
                    let score = chainMapping.structures[strucId][chainName][resno].c;
                    if (!score) continue;
                    if (mx != mn) {
                        score = (parseFloat(score) - mn) / (mx - mn);
                        if (score < .5) {
                            rgb[0] = Math.round(255 * (1 - score));
                            rgb[1] = Math.round(160 * score);
                            rgb[2] = Math.round(90 * score);
                        } else {
                            rgb[0] = Math.round(127);
                            rgb[1] = Math.round(20 + 175 * score);
                            rgb[2] = 45;
                        }
                    }
                    rowSpan[i].style.backgroundColor = "rgba(" + rgb + ",0.6)";
                    if (strucId) chainMapping.structures[strucId][chainName][resno].rgb = rgb;
                }
            });
        } else if (col == "clustal") {
            let clustal = {
                red: [ 238, 51, 17 ],
                blue: [ 17, 119, 238 ],
                green: [ 17, 204, 17 ],
                orange: [ 255, 102, 0 ],
                cyan: [ 17, 187, 187 ],
                pink: [ 238, 119, 119 ],
                magenta: [ 204, 68, 204 ],
                yellow: [ 204, 204, 0 ]
            };
            let clst = $alignTbl.data("clustalColours");
            if (!clst) {
                let seqs = getSeqs($alignTbl.data("templateData"), $alignTbl);
                clst = calculateClustalColours(seqs.concat(seqs.target));
                $alignTbl.data("clustalColours", clst);
            }
            let res;
            let seqno = 0;
            $alignTbl.find("tr").each(function() {
                if (parseInt($(this).attr("seqno")) < seqno) return false;
                let strucId = $(this).attr("struc_id");
                let chain = $(this).attr("chain");
                if (!chain) chain = $(this).attr("keyseqchain");
                seqno = parseInt($(this).attr("seqno"));
                let rowSpan = $alignTbl.find("tr[seqno=" + seqno + "] span");
                $.each(rowSpan, function(column, span) {
                    if (clst[column] == null) return;
                    let rgb = null;
                    if (span.hasAttribute("r")) {
                        res = span.getAttribute("r");
                        switch (res) {
                          case "A":
                          case "I":
                          case "L":
                          case "M":
                          case "F":
                          case "W":
                          case "V":
                            if (clst[column].WLVIMAFCYHP_nogap > .6) rgb = clustal.blue;
                            break;

                          case "C":
                            if (clst[column].WLVIMAFCYHP_nogap > .6) rgb = clustal.blue;
                            if (clst[column].C_nogap > .85) rgb = clustal.pink;
                            break;

                          case "K":
                          case "R":
                            if (clst[column].KR_nogap > .6 || clst[column].K_nogap > .8 || clst[column].R_nogap > .8 || clst[column].Q_nogap > .8) rgb = clustal.red;
                            break;

                          case "E":
                            if (clst[column].KR_nogap > .6 || clst[column].ED_nogap > .5 || clst[column].QE_nogap > .5 || clst[column].E_nogap > .85 || clst[column].Q_nogap > .85 || clst[column].D_nogap > .85) rgb = clustal.magenta;
                            break;

                          case "D":
                            if (clst[column].KR_nogap > .6 || clst[column].ED_nogap > .5 || clst[column].D_nogap > .85 || clst[column].E_nogap > .85 || clst[column].N_nogap > .85) rgb = clustal.magenta;
                            break;

                          case "N":
                            if (clst[column].N_nogap > .5 || clst[column].Y_nogap > .85) rgb = clustal.green;
                            break;

                          case "Q":
                            if (clst[column].KR_nogap > .6 || clst[column].QE_nogap > .5 || clst[column].Q_nogap > .85 || clst[column].E_nogap > .85 || clst[column].K_nogap > .85 || clst[column].R_nogap > .85) rgb = clustal.green;
                            break;

                          case "S":
                          case "T":
                            if (clst[column].WLVIMAFCYHP_nogap > .6 || clst[column].TS_nogap > .5 || clst[column].S_nogap > .85 || clst[column].T_nogap > .85) rgb = clustal.green;
                            break;

                          case "G":
                            rgb = clustal.orange;
                            break;

                          case "P":
                            rgb = clustal.yellow;
                            break;

                          case "H":
                          case "Y":
                            if (clst[column].WLVIMAFCYHP_nogap > .6 || clst[column].W_nogap > .85 || clst[column].Y_nogap > .85 || clst[column].A_nogap > .85 || clst[column].C_nogap > .85 || clst[column].P_nogap > .85 || clst[column].Q_nogap > .85 || clst[column].F_nogap > .85 || clst[column].H_nogap > .85 || clst[column].I_nogap > .85 || clst[column].L_nogap > .85 || clst[column].M_nogap > .85 || clst[column].V_nogap > .85) rgb = clustal.cyan;
                            break;
                        }
                        if (rgb) {
                            span.style.backgroundColor = "rgb(" + rgb + ",.5)";
                            if (strucId && chainMapping.structures[strucId] && chainMapping.structures[strucId][chain]) {
                                let s = parseInt(span.getAttribute("s"));
                                if (chainMapping.structures[strucId][chain][s]) chainMapping.structures[strucId][chain][s].rgb = rgb;
                            }
                        }
                    }
                });
            });
        } else if (col == "secstruc") {
            let sstype = $.cookie("secStruct");
            if (sstype == null) sstype = "dssp";
            let templateData = $alignTbl.data("templateData");
            let $visibleRows = $alignTbl.find("tr");
            let seqres;
            let strucId;
            if (templateData.keyseq_annotation) {
                for (let i = 0; i < templateData.keyseq_chain_name.length; i++) {
                    strucId = templateData.struc_id;
                    let chn = templateData.keyseq_chain_name[i];
                    if (templateData.keyseq_annotation[chn] && templateData.keyseq_annotation[chn][sstype]) {
                        let annot = stringSS(templateData.keyseq_annotation[chn], sstype).replace(/[GI]/g, "H").replace(/B/g, "E");
                        for (let i = 0; i < annot.length; i++) if (chainMapping.structures[strucId][chn][i + 1]) chainMapping.structures[strucId][chn][i + 1].ss = annot[i];
                    } else if (chainMapping.structures[strucId]) for (let i = 0; i < chainMapping.structures[strucId][chn].length; i++) if (chainMapping.structures[strucId][chn][i]) chainMapping.structures[strucId][chn][i].ss = null;
                }
            } else if (templateData.templates) {
                for (let t = 0; t < templateData.templates.length; t++) {
                    let tpl = templateData.templates[t];
                    if (tpl.annotation && tpl.annotation[sstype]) {
                        let annot = stringSS(tpl.annotation, sstype).replace(/[GI]/g, "H").replace(/B/g, "E");
                        if (chainMapping.structures[tpl.struc_id]) {
                            for (let i = 0; i < annot.length; i++) if (chainMapping.structures[tpl.struc_id][tpl.chain] && chainMapping.structures[tpl.struc_id][tpl.chain][i + 1]) chainMapping.structures[tpl.struc_id][tpl.chain][i + 1].ss = annot[i];
                        } else console.warn("no chain mapping found for " + tpl.struc_id + " " + tpl.chain);
                    } else if (tpl.struc_id && sstype != "none") {
                        let beenThere = false;
                        for (let i = 0; !beenThere && i < chainMapping.structures[tpl.struc_id][tpl.chain].length; i++) if (chainMapping.structures[tpl.struc_id][tpl.chain][i] && chainMapping.structures[tpl.struc_id][tpl.chain][i].ss) beenThere = true;
                        if (!beenThere) fetchSecStruc(tpl.struc_id);
                    }
                }
            }
            let seqno = 0;
            for (let r = 0; r < $visibleRows.length; r++) {
                let $this = $visibleRows.eq(r);
                if (parseInt($this.attr("seqno")) < seqno) break;
                seqno = parseInt($this.attr("seqno"));
                let chn = $this.attr("chain");
                if (!chn) chn = $this.attr("keyseqchain");
                strucId = $this.attr("struc_id");
                let annot = null;
                if (!strucId) {
                    if (templateData.type == "model-template") {
                        for (let t = 0; t < templateData.templates.length; t++) {
                            if (templateData.templates[t].chain == chn) {
                                if (templateData.templates[t].annotation && templateData.templates[t].annotation[sstype]) {
                                    annot = stringSS(templateData.templates[t].annotation, sstype);
                                    annot = annot.replace(/[GI]/g, "H").replace(/B/g, "E");
                                }
                                break;
                            }
                        }
                    }
                    if (!annot) continue;
                }
                $row = $visibleRows.filter('[seqno="' + seqno + '"]');
                let span = $row.find("span[s]").toArray();
                $.each(span, function(i, span) {
                    let rgb = null;
                    seqres = parseInt(span.getAttribute("s"));
                    if (annot) {
                        if (annot[seqres - 1] == "E") rgb = [ 51, 204, 51 ]; else if (annot[seqres - 1] == "H") rgb = [ 153, 153, 230 ];
                    } else if (chainMapping.structures[strucId] && chainMapping.structures[strucId][chn] && chainMapping.structures[strucId][chn][seqres]) {
                        if (chainMapping.structures[strucId][chn][seqres].ss == "E") rgb = [ 51, 204, 51 ]; else if (chainMapping.structures[strucId][chn][seqres].ss == "H") rgb = [ 153, 153, 230 ];
                        chainMapping.structures[strucId][chn][seqres].rgb = rgb;
                    }
                    span.style.backgroundColor = "rgba(" + rgb + ",.6)";
                });
            }
        } else if ([ "hydrophobic", "size", "charged", "cysteine" ].indexOf(col) > -1 || residueProperties[col]) {
            $alignTbl.find("tr").each(function() {
                let strucId = $(this).attr("struc_id");
                let chain = $(this).attr("chain");
                if (!chain) chain = $(this).attr("keyseqchain");
                let rowSpan = $(this).find("span");
                $.each(rowSpan, function(i, span) {
                    res = span.getAttribute("r");
                    if (!res || !residueProperties[res]) return;
                    let rgb = residuePropertyColor(col, res);
                    if (!rgb) return true;
                    span.style.backgroundColor = "rgba(" + rgb + ",.6)";
                    if (strucId && chainMapping.structures[strucId] && chainMapping.structures[strucId][chain]) {
                        let s = parseInt(span.getAttribute("s"));
                        if (chainMapping.structures[strucId][chain][s]) chainMapping.structures[strucId][chain][s].rgb = rgb;
                    }
                });
            });
        } else if (col == "base") {
            $alignTbl.find("tr").each(function() {
                let strucId = $(this).attr("struc_id");
                let chain = $(this).attr("chain");
                if (!chain) chain = $(this).attr("keyseqchain");
                let rowSpan = $(this).find("span");
                $.each(rowSpan, function(i, span) {
                    res = span.getAttribute("r");
                    if (!res) return true;
                    let rgb = {
                        A: [ 70, 195, 70 ],
                        C: [ 195, 195, 70 ],
                        G: [ 195, 70, 70 ],
                        T: [ 70, 70, 195 ],
                        U: [ 70, 70, 195 ]
                    }[res];
                    if (!rgb) return true;
                    span.style.backgroundColor = "rgba(" + rgb + ",.6)";
                    if (strucId && chainMapping.structures[strucId] && chainMapping.structures[strucId][chain]) {
                        let s = parseInt(span.getAttribute("s"));
                        if (chainMapping.structures[strucId][chain][s]) chainMapping.structures[strucId][chain][s].rgb = rgb;
                    }
                });
            });
        } else if (col == "soa") {
            colourAnnotation($alignTbl, "soa");
        } else if (col == "lddt") {
            colourAnnotation($alignTbl, col);
        } else if (col.indexOf("bfactor") > -1) {
            colourAnnotation($alignTbl, col);
        } else if (col == "entropy") {
            colourAnnotation($alignTbl, "entropy");
        } else if (col == "indels" && $alignTbl.data("templateData").type == "model-template" && $alignTbl.data("templateData").templates[0].tpl_seq) {
            let tdata = $alignTbl.data("templateData");
            let strucId = tdata.struc_id;
            let mnFrm = 99999999;
            let mxTo = 0;
            let tplInsertions = [];
            let trgInsertions = [];
            for (let i = 0; i < tdata.templates.length; i++) {
                if (tdata.templates[i].struc_id != strucId) continue;
                trgOffset = tdata.templates[i].trg_offset ? tdata.templates[i].trg_offset - 1 : 0;
                if (tdata.templates[i].residue_from) {
                    mnFrm = Math.min(mnFrm, tdata.templates[i].residue_from);
                    mxTo = Math.max(mxTo, tdata.templates[i].residue_to);
                } else {
                    mnFrm = 0;
                    mxTo = 99999999999;
                }
                for (let c = 0; c < tdata.templates[i].trg_seq.length; c++) {
                    if (tdata.templates[i].trg_seq.charAt(c) != "-") trgInsertions[c] = "X";
                    if (tdata.templates[i].tpl_seq.charAt(c) != "-") tplInsertions[c] = "X";
                }
            }
            let lastTrg = null;
            let resno = 1 + trgOffset;
            let insertions = [], deletions = [];
            for (let i = 0; i < tdata.templates[0].trg_seq.length; i++) {
                if (!tplInsertions[i]) for (let c = 0; insertions.indexOf(resno) == -1 && c < tdata.keyseq_chain_name.length; c++) if (resno >= mnFrm && resno <= mxTo) insertions.push(resno);
                if (!trgInsertions[i]) {
                    if (lastTrg != null) for (let c = 0; deletions.indexOf(resno) == -1 && c < tdata.keyseq_chain_name.length; c++) {
                        let chain = tdata.keyseq_chain_name[c];
                        if (resno >= mnFrm && resno <= mxTo) deletions.push(resno);
                        if (resno - 1 >= mnFrm && resno <= mxTo) deletions.push(resno - 1);
                    }
                } else {
                    lastTrg = resno;
                    resno++;
                }
            }
            for (let c = 0; c < tdata.keyseq_chain_name.length; c++) {
                let chain = tdata.keyseq_chain_name[c];
                let $rows = $alignTbl.find("tr[keyseqchain=" + chain + "]");
                $.each($rows.find("span[s]:not([nores])"), function(j, span) {
                    let resno = parseInt(span.getAttribute("s"));
                    if (insertions.indexOf(resno) > -1) rgb = [ 80, 214, 80 ]; else if (deletions.indexOf(resno) > -1) rgb = [ 100, 100, 255 ]; else return true;
                    span.style.backgroundColor = "rgba(" + rgb + ",.6)";
                    chainMapping.structures[strucId][chain][resno].rgb = rgb;
                });
            }
        }
        if ($.cookie("targetMismatch")) {
            if ($.cookie("targetMismatch") == "fade") $spans.filter("[nm]").addClass("targetMismatch"); else if ($.cookie("targetMismatch") == "enhance") {
                $spans.not("[nm]").addClass("targetMismatch");
                $spans.not("[s]").addClass("targetMismatch");
            }
        } else $spans.filter("[nm]").addClass("targetMismatch");
        if (editMode) {
            if (currentCursor) {
                $alignTbl.find(`tr[seqno=${currentCursor.seqno}]`).find("span").eq(currentCursor.seqidx).addClass("cursor");
            }
            if ($alignTbl.data("startEditRegion") && $alignTbl.data("endEditRegion")) {
                let oldstart = $alignTbl.data("startEditRegion");
                let oldend = $alignTbl.data("endEditRegion");
                let oldrow = oldstart.closest("tr");
                let seqno = oldrow.attr("seqno");
                let oldtr = oldrow.siblings("[seqno=" + seqno + "]").addBack();
                let oldspan = oldtr.find("span");
                let newspan = $alignTbl.find("tr[seqno=" + seqno + "]").find("span");
                newspan = newspan.slice(oldspan.index(oldstart), oldspan.index(oldend) + 1).addClass("shadowCursor");
                $alignTbl.data("startEditRegion", newspan.first());
                $alignTbl.data("endEditRegion", newspan.last());
            }
        }
    }
    function residuePropertyColor(col, res) {
        if (col == "hydrophobic") {
            let val = residueProperties[res][col];
            if (val == null) return;
            return [ Math.round(val * 255), 0, 255 - Math.round(val * 255) ];
        } else if (col == "size") {
            let val = residueProperties[res][col];
            if (val == null) return;
            return [ Math.round(val * 255), 255 - Math.round(val * 255), 0 ];
        } else if (col == "charged") {
            if (residueProperties["positive"].indexOf(res) > -1) return [ 100, 100, 255 ]; else if (residueProperties["negative"].indexOf(res) > -1) return [ 255, 100, 100 ];
        } else if (col == "cysteine") {
            return res == "C" ? [ 202, 168, 14 ] : null;
        } else {
            return residueProperties[col].indexOf(res) > -1 ? [ 20, 80, 250 ] : null;
        }
    }
    function colourAnnotation($alignTbl, col, rows) {
        if ($alignTbl.data("options")["textOnly"]) return;
        let valuesExist = false;
        let strucId = null, qname = null;
        let bfactor_range = col == "bfactor_range";
        if (bfactor_range) col = "bfactor";
        (rows ? rows : $alignTbl.find("tr[struc_id]")).each(function() {
            strucId = $(this).attr("struc_id");
            qname = $(this).attr("qname");
            let chain = $(this).attr("chain");
            if (!chain) chain = $(this).attr("keyseqchain");
            if (!chain) return true;
            let rowSpan = $(this).find("span[s]");
            $.each(rowSpan, function(i, span) {
                let s = parseInt(span.getAttribute("s"));
                if (!chainMapping.structures[strucId] || !chainMapping.structures[strucId][chain] || !chainMapping.structures[strucId][chain][s] || !(col in chainMapping.structures[strucId][chain][s])) return true;
                valuesExist = true;
                let val = chainMapping.structures[strucId][chain][s][col];
                let rgb = [ 0, 200, 0 ];
                if (col == "entropy") {
                    if (val < 1.9) rgb = [ Math.floor(val / 1.9 * 256), Math.floor(val / 1.9 * 256), 255 ]; else rgb = [ 255, 256 - Math.floor((val - 1.9) / 1.9 * 256), 256 - Math.floor((val - 1.9) / 1.9 * 256) ];
                } else if (col == "soa") {
                    val = Math.max(0, Math.min(1, val * .01));
                    rgb = [ Math.floor((1 - val) * 255), 0, 255 - Math.floor((1 - val) * 255) ];
                } else if (col == "lddt") {
                    val = (1 - val) * (1.5 * Math.PI - Math.PI / 2) + Math.PI / 2;
                    rgb = [ Math.round(255 / 2 * (Math.sin(val + Math.PI) + 1)), Math.round(255 * Math.sin(val - Math.PI / 2)), Math.round(255 / 2 * (Math.sin(val) + 1)) ];
                } else if (col == "bfactor") {
                    if (bfactor_range) {
                        let bidx = chainMapping.annotations.map(m => m.key).indexOf("bfactor_range");
                        if (chainMapping.annotations[bidx].rangeColours[strucId] && chainMapping.annotations[bidx].rangeColours[strucId][chain]) rgb = chainMapping.annotations[bidx].rangeColours[strucId][chain][s];
                    } else {
                        if (val <= 10) rgb = [ 0, 0, 255 ]; else if (val > 10 && val <= 15) rgb = [ 115, 115, 255 ]; else if (val > 15 && val <= 20) rgb = [ 170, 170, 255 ]; else if (val > 20 && val <= 25) rgb = [ 215, 215, 255 ]; else if (val > 25 && val <= 30) rgb = [ 255, 215, 215 ]; else if (val > 30 && val <= 35) rgb = [ 255, 170, 170 ]; else if (val > 35 && val <= 40) rgb = [ 255, 115, 115 ]; else rgb = [ 255, 0, 0 ];
                    }
                }
                span.style.backgroundColor = "rgba(" + rgb + ",.6)";
                chainMapping.structures[strucId][chain][s].rgb = rgb;
            });
        });
        if (!valuesExist) {
            if (qname && (col == "soa" || col == "entropy") && typeof SCORE_URL !== "undefined") {
                $.getJSON(SCORE_URL.replace("xxxx.0", qname.match(/\w{4}\.\d+/)[0]) + col, function(json, status) {
                    if (json.error) {
                        console.error("Could not fetch " + col + " values");
                        return;
                    }
                    for (let chn in json) {
                        for (let i = 0; i < json[chn].values.length; i++) {
                            if (json[chn].values[i] != null && chainMapping.structures[strucId] && chainMapping.structures[strucId][chn] && chainMapping.structures[strucId][chn][i + json[chn].offset + 1]) chainMapping.structures[strucId][chn][i + json[chn].offset + 1][col] = json[chn].values[i];
                        }
                    }
                    colourAlignment(col, $alignTbl);
                    getColoursForViewer();
                });
            } else if (col == "bfactor") {
                if (fetchBFactors(strucId, "bfactor")) colourAnnotation($alignTbl, col + (bfactor_range ? "_range" : ""));
            }
        }
    }
    function getColoursForViewer(args) {
        if (!args || !args.singleResidue) {
            if (typeof colourRama === "function") colourRama();
            if (typeof colourModelBars === "function") colourModelBars();
        }
        if (typeof currentWin !== "undefined") {
            if (typeof pv_viewer !== "undefined" && $("#pv:visible", currentWin.document).length > 0) {
                getColoursForPV();
            }
            if (ngl_stage != null && $("#ngl:visible", currentWin.document).length > 0) {
                if (args && args.singleResidue) getColoursForNGLSingleResidue(args.singleResidue); else getColoursForNGL(args);
            }
        }
    }
    function getColoursForNGLSingleResidue(singleRes) {
        for (let schemeId in NGL.ColormakerRegistry.userSchemes) if (schemeId.indexOf("singleres") > -1) {
            NGL.ColormakerRegistry.removeScheme(schemeId);
            ngl_stage.eachRepresentation(function(rep) {
                if (rep.repr.colorScheme == schemeId) {
                    for (let schemeId2 in NGL.ColormakerRegistry.userSchemes) if (schemeId2.endsWith(rep.name)) {
                        rep.setColor(schemeId2);
                        return;
                    }
                }
            });
        }
        if (singleRes) {
            let comp = getNGLComponentByName(singleRes.struc_id);
            if (comp) {
                ngl_stage.eachRepresentation(function(rep) {
                    let thisStrucId = comp.name;
                    if (rep.name != thisStrucId) {
                        if (rep.name.endsWith(".dna") && comp.name.startsWith(rep.name.substring(0, 5))) thisStrucId = comp.name + ".dna"; else return;
                    }
                    if (!chainMapping.structures[thisStrucId]) return;
                    let schemeId = NGL.ColormakerRegistry.addScheme(function(params) {
                        for (let chn in chainMapping.structures[thisStrucId]) if (chn === " ") $("#ngl_status").html("NGL cannot colour or select chains named whitespace!");
                        this.atomColor = function(atom) {
                            if (ngl_picked_atom && atom.structure.name === ngl_picked_atom.structure.name && atom.index === ngl_picked_atom.index) return 16711680;
                            if (!chainMapping.structures[thisStrucId][atom.chainname] || !chainMapping.structures[thisStrucId][atom.chainname][atom.resno] || !chainMapping.structures[thisStrucId][atom.chainname][atom.resno].rgb) return 16777215;
                            return chainMapping.structures[thisStrucId][atom.chainname][atom.resno].rgb[0] << 16 | chainMapping.structures[thisStrucId][atom.chainname][atom.resno].rgb[1] << 8 | chainMapping.structures[thisStrucId][atom.chainname][atom.resno].rgb[2];
                        };
                    }, "singleres");
                    rep.setColor(schemeId);
                });
            }
        }
    }
    function getColoursForNGL(args) {
        let visibleStrucIDs = [];
        ngl_stage.eachRepresentation(function(r) {
            visibleStrucIDs.push(r.name);
        });
        let onlyThisStrucId = args && args.strucId ? args.strucId : null;
        let repSet = {};
        for (let comp in ngl_stage.compList) {
            if (ngl_stage.compList[comp].structure === undefined) continue;
            let compName = ngl_stage.compList[comp].name;
            if (!chainMapping.structures[compName]) continue;
            if (onlyThisStrucId && compName != onlyThisStrucId) continue;
            for (let schemeId in NGL.ColormakerRegistry.userSchemes) if (schemeId.indexOf(compName) > -1 && schemeId.indexOf("membrane") == -1) NGL.ColormakerRegistry.removeScheme(schemeId);
            ngl_stage.eachRepresentation(function(rep) {
                if (rep.name != compName && !rep.name.endsWith(".dna") && !rep.name.endsWith(".highlight") && rep.name != "drug.surroundings") return;
                if (repSet[rep.name]) {
                    rep.setColor(repSet[rep.name]);
                    return;
                }
                let strucId;
                if (rep.name.endsWith(".dna")) {
                    if (!compName.startsWith(rep.name.substring(0, 5))) return;
                    if (chainMapping.structures[compName + ".dna"]) strucId = compName + ".dna"; else return;
                } else strucId = compName;
                let schemeId = NGL.ColormakerRegistry.addScheme(function(params) {
                    for (let chn in chainMapping.structures[strucId]) if (chn === " ") $("#ngl_status").html("NGL cannot colour or select chains named whitespace!");
                    this.atomColor = function(atom) {
                        if (!chainMapping.structures[strucId][atom.chainname] || !chainMapping.structures[strucId][atom.chainname][atom.resno] || !chainMapping.structures[strucId][atom.chainname][atom.resno].rgb) {
                            return 16777215;
                        }
                        return chainMapping.structures[strucId][atom.chainname][atom.resno].rgb[0] << 16 | chainMapping.structures[strucId][atom.chainname][atom.resno].rgb[1] << 8 | chainMapping.structures[strucId][atom.chainname][atom.resno].rgb[2];
                    };
                }, strucId);
                repSet[rep.name] = schemeId;
                rep.setColor(schemeId);
            });
        }
    }
    function getColoursForPV() {
        let visibleStrucIDs = [];
        pv_viewer.forEach(function(v) {
            visibleStrucIDs.push(v.name());
        });
        if (visibleStrucIDs.length == 0) return;
        for (let strucId in chainMapping.structures) {
            if (visibleStrucIDs.indexOf(strucId) > -1 && pv_viewer.get(strucId)) pv_viewer.get(strucId).colorBy(new pv.color.ColorOp(function(atom, out, index) {
                let res = atom.residue();
                if (chainMapping.structures[strucId][res.chain().name()] && chainMapping.structures[strucId][res.chain().name()][res.num()] && chainMapping.structures[strucId][res.chain().name()][res.num()].rgb) {
                    out[index] = chainMapping.structures[strucId][res.chain().name()][res.num()].rgb[0] / 255;
                    out[index + 1] = chainMapping.structures[strucId][res.chain().name()][res.num()].rgb[1] / 255;
                    out[index + 2] = chainMapping.structures[strucId][res.chain().name()][res.num()].rgb[2] / 255;
                    out[index + 3] = 1;
                } else {
                    out[index] = .9;
                    out[index + 1] = .9;
                    out[index + 2] = .9;
                    out[index + 3] = 1;
                }
            }));
        }
        pv_viewer.requestRedraw();
    }
    function repositionAnnotation($alignTbl) {
        $("#" + $alignTbl.attr("id") + "_features").css("top", $alignTbl.position().top);
    }
    function drawAnnotation($alignTbl) {
        let sstype = $.cookie("secStruct");
        if (sstype == null) sstype = "dssp";
        let featureId = $alignTbl.attr("id") + "_features";
        $("#" + featureId).html("");
        if ($("#" + featureId).length == 0 || sstype && sstype == "none") {
            $alignTbl.removeClass("ss");
            getAlignNameTbl($alignTbl).removeClass("ss");
            return;
        }
        $alignTbl.addClass("ss");
        getAlignNameTbl($alignTbl).addClass("ss");
        let alTblTop = $alignTbl.position().top;
        let alWidth = $alignTbl.find("td:visible").first().width();
        let sfirst = $alignTbl.find("span:visible");
        if (sfirst.length === 0) return;
        let spanWidth = sfirst[0].getBoundingClientRect().width;
        let spanHeight = sfirst[0].getBoundingClientRect().height;
        let alTblLeft = $alignTbl.position().left;
        if (alWidth < 1) return;
        p = Raphael(featureId, alWidth, $alignTbl.height());
        $("#" + featureId).css("top", $alignTbl.position().top);
        let templateData = $alignTbl.data("templateData");
        if (!templateData) return;
        let $visibleRows = $alignTbl.find("tr:visible");
        if (templateData.keyseq_annotation) {
            let trg_offset = templateData.templates[0].trg_offset ? templateData.templates[0].trg_offset : 1;
            for (let i = 0; i < templateData.keyseq_chain_name.length; i++) {
                let chn = templateData.keyseq_chain_name[i];
                if (templateData.keyseq_annotation[chn]) {
                    if (templateData.keyseq_annotation[chn][sstype]) drawSS(p, cigarSS(templateData.keyseq_annotation[chn], sstype, trg_offset), $visibleRows.filter('[keyseqchain="' + chn + '"]'), $visibleRows.filter('[keyseqchain="' + chn + '"]').find("span[s]").toArray(), alTblTop, alTblLeft, alWidth, spanWidth, spanHeight);
                }
            }
        }
        if (templateData.charts) {
            for (let i = 0; i < templateData.templates.length; i++) {
                let tpl = templateData.templates[i];
                if (!tpl.struc_id || !tpl.annotation) continue;
                let trg_offset = tpl.trg_offset ? tpl.trg_offset : 1;
                function delayBarChart(_p, _chn, _rows) {
                    setTimeout(function() {
                        drawBarCharts(_p, tpl.annotation, sstype, trg_offset, _rows, alTblTop, alWidth, spanWidth, templateData);
                    }, 0);
                }
                delayBarChart(p, tpl.chain, $visibleRows.filter(`[struc_id="${tpl.struc_id}"][chain="${tpl.chain}"]`));
            }
        }
        let seqno = null;
        for (let visrow of $visibleRows) {
            $this = $(visrow);
            let thisSeqno = parseInt($this.attr("seqno"));
            if (seqno !== null && thisSeqno <= seqno) break;
            if (seqno === null) seqno = thisSeqno;
            if ($this.attr("keyseqtype")) continue;
            let chain = $this.attr("chain");
            let strucId = $this.attr("struc_id");
            $row = $visibleRows.filter("[seqno=" + thisSeqno + "]");
            let span = $row.find("span[s]").toArray();
            for (let t = 0; t < templateData.templates.length; t++) {
                if (templateData.templates[t].annotation) cigarSS(templateData.templates[t].annotation, sstype);
                if (strucId ? templateData.templates[t].struc_id == strucId : templateData.templates[t].chain == chain) {
                    if (!strucId && templateData.type == "target-template" && templateData.templates[t].struc_id) continue;
                    if (templateData.templates[t].chain != chain) continue;
                    if (templateData.templates[t].annotation && templateData.templates[t].annotation[sstype]) drawSS(p, cigarSS(templateData.templates[t].annotation, sstype), $row, span, alTblTop, alTblLeft, alWidth, spanWidth, spanHeight);
                }
            }
        }
    }
    function stringSS(object, sstype) {
        let ss = object[sstype];
        if (typeof ss == "string") object[sstype + "_string"] = ss.replace(/[GI]/g, "H").replace(/B/g, "E");
        return object[sstype + "_string"];
    }
    function cigarSS(object, sstype, trg_offset) {
        if (!trg_offset) trg_offset = 1;
        if (typeof object[sstype] === "string") {
            object[sstype + "_string"] = object[sstype];
            ss = object[sstype].replace(/[GI]/g, "H").replace(/B/g, "E");
            let a = null;
            for (let i = 0; i < ss.length; i++) {
                if (ss[i] === "H" || ss[i] === "E") {
                    res_index = i + trg_offset;
                    if (a == null) a = [ [ res_index, res_index, ss[i] ] ]; else if (a[a.length - 1][1] === res_index - 1 && a[a.length - 1][2] === ss[i]) a[a.length - 1][1] = res_index; else a.push([ res_index, res_index, ss[i] ]);
                }
            }
            object[sstype] = a;
        }
        return object[sstype];
    }
    function drawBarCharts(p, anno, sstype, trg_offset, rows, alTblTop, alWidth, spanWidth, templateData) {
        if (!templateData.charts) return;
        let chartsToShow = templateData.charts.split("|");
        let secStructHeight = 0;
        let marginBottom = 5;
        if (chartsToShow.indexOf("dssp") > -1) {
            marginBottom = 2;
            secStructHeight = 12;
            chartsToShow.splice(chartsToShow.indexOf("dssp"), 1);
        }
        let chartsMarginBottom = 6;
        let colouredBars = templateData.pageResidueCount < 5e3;
        if (!trg_offset) trg_offset = 1;
        let st = p.set();
        let tset = p.set();
        let cset = p.set();
        let chartHeight = null;
        let minY, h, y, initY, box;
        rows.each(function(r, row) {
            span = $(row).find("span").toArray();
            if (span.length == 0) return;
            let strucId = row.getAttribute("struc_id");
            let chainName = row.getAttribute("chain");
            let containerBaseY = $(row).position().top - alTblTop + $(span[0]).position().top;
            let alTblLeft = row.getBoundingClientRect().left;
            let lastx = span[span.length - 1].getBoundingClientRect().right - alTblLeft;
            if (lastx === 0) return;
            if (!chartHeight) chartHeight = (containerBaseY - secStructHeight - marginBottom - 10 - chartsMarginBottom * (chartsToShow.length - 1)) / (chartsToShow.length ? chartsToShow.length : 1);
            let chartPaths = [];
            let seq = "";
            for (let chIdx = 0; chIdx < chartsToShow.length; chIdx++) {
                let chart = chartsToShow[chIdx];
                minY = chart === "qm" ? .3 : 0;
                initY = containerBaseY - chartHeight * chIdx - marginBottom - secStructHeight - chIdx * chartsMarginBottom;
                y = initY;
                box = p.path(`M0,${y}L${alWidth},${y}L${alWidth},${y - chartHeight}L0,${y - chartHeight}`);
                box.attr({
                    "stroke-width": .4,
                    stroke: "#888"
                });
                for (let i = Math.round(minY * 10); i < 11; i++) {
                    h = y - chartHeight * ((i - minY * 10) / (10 - minY * 10));
                    let l = p.path("M0," + h + "L" + alWidth + "," + h);
                    if (i == 5) l.attr({
                        "stroke-width": .7,
                        opacity: .8,
                        stroke: "#888",
                        "stroke-dasharray": "-"
                    }); else st.push(l);
                }
                chartPaths.push("");
            }
            for (let i = 0; i < span.length; i++) {
                let sp = span[i];
                let resNo = sp.getAttribute("s");
                if (!resNo) {
                    seq += "-";
                    continue;
                }
                let x = sp.getBoundingClientRect().left - alTblLeft;
                let resno = parseInt(resNo);
                let sindex = resno - trg_offset;
                if (anno[sstype + "_string"] && anno[sstype + "_string"][sindex]) seq += anno[sstype + "_string"][sindex]; else seq += "-";
                for (let chIdx = 0; chIdx < chartsToShow.length; chIdx++) {
                    let chart = chartsToShow[chIdx];
                    minY = chart === "qm" ? .3 : 0;
                    initY = containerBaseY - chartHeight * chIdx - marginBottom - secStructHeight - chIdx * chartsMarginBottom;
                    y = initY;
                    let score = Math.max(minY, chainMapping.structures[strucId][chainName][resno][chart]);
                    h = y - chartHeight * ((score - minY) / (1 - minY));
                    if (colouredBars) {
                        let rgb;
                        if (chart === "qm") {
                            if (score > .5) rgb = [ Math.round((1 - score) * 490), 80, Math.round(score * 255) ]; else rgb = "245,80,0";
                        } else if (chart === "lddt") {
                            score = (1 - score) * (1.5 * Math.PI - Math.PI / 2) + Math.PI / 2;
                            rgb = [ Math.round(255 / 2 * (Math.sin(score + Math.PI) + 1)), Math.round(255 * Math.sin(score - Math.PI / 2)), Math.round(255 / 2 * (Math.sin(score) + 1)) ];
                        }
                        s = p.path(`M${x},${y}L${x},${h}L${x + spanWidth},${h}L${x + spanWidth},${y}`);
                        cset.push(s);
                        s.attr({
                            fill: `rgba(${rgb},.6)`
                        });
                    } else chartPaths[chIdx] += `M${x} ${y}L${x} ${h}L${x + spanWidth} ${h}L${x + spanWidth} ${y}`;
                }
            }
            for (let path of chartPaths) cset.push(p.path(path));
            if (secStructHeight > 0) {
                txt = p.text(1.5, containerBaseY - secStructHeight - marginBottom + secStructHeight / 2, seq);
                tset.push(txt);
                txt.node.setAttribute("textLength", lastx - spanWidth / 2 + 2 + "px");
                txt.node.firstChild.setAttribute("textLength", lastx - spanWidth / 2 + 1.5 + "px");
            }
        });
        tset.attr({
            "text-anchor": "start",
            "font-family": "monospace"
        });
        cset.attr({
            "stroke-width": .7,
            stroke: "#666"
        });
        st.attr({
            "stroke-width": .9,
            stroke: "#888",
            opacity: .6,
            "stroke-dasharray": ". "
        });
    }
    function drawSS(p, ss, $row, span, alTblTop, alTblLeft, alWidth, spanWidth, spanHeight) {
        if (!ss) return;
        let spanIndex = 0;
        let spanLength = span.length;
        for (let i = 0; i < ss.length; i++) {
            if (ss[i][2] != "E" && ss[i][2] != "H") continue;
            let $startRes = null, $endRes = null, resTop = null;
            for (let x = ss[i][0]; x <= ss[i][1]; x++) {
                if (span[spanIndex].getAttribute("s") > x) continue;
                for (let s = spanIndex; s < spanLength; s++) {
                    if (span[s].getAttribute("s") == x) {
                        spanIndex = s;
                        $startRes = $(span[s]);
                        break;
                    }
                }
                if ($startRes) break;
            }
            for (let x = ss[i][1]; x >= ss[i][0]; x--) {
                for (let s = spanIndex; s < spanLength; s++) {
                    if (span[s].getAttribute("s") == x) {
                        spanIndex = s;
                        $endRes = $(span[s]);
                        break;
                    }
                }
                if ($endRes) break;
            }
            if ($startRes == null || $endRes == null) {
                spanIndex = 0;
                continue;
            }
            thisRow = $startRes.closest("tr");
            startRowIdx = $row.index(thisRow);
            endRowIdx = $row.index($endRes.closest("tr"));
            let startResPos = $startRes.position();
            startResLeft = startResPos.left - alTblLeft;
            spanTop = startResPos.top;
            endResLeft = $endRes.position().left - alTblLeft;
            rightLimit = thisRow.width() - spanWidth - spanWidth;
            for (let j = startRowIdx; j < endRowIdx + 1; j++) {
                if (j > startRowIdx) {
                    thisRow = $row.eq(j);
                    spanTop = thisRow.find("span").first().position().top;
                }
                resTop = thisRow.position().top + spanTop - alTblTop;
                if (ss[i][2] == "E") drawSheet(p, resTop, startResLeft - (j - startRowIdx) * alWidth, endResLeft + (endRowIdx - j) * alWidth, spanWidth, spanHeight, rightLimit); else drawHelix(p, resTop, startResLeft - (j - startRowIdx) * alWidth, endResLeft + (endRowIdx - j) * alWidth, spanWidth, spanHeight, rightLimit);
            }
        }
    }
    function drawSheet(p, y, start, end, spanWidth, height, rightLimit) {
        y += 1;
        height -= 2;
        padding = 2;
        start = Math.max(-spanWidth, Math.round(start));
        end = Math.min(rightLimit, Math.round(end + spanWidth));
        let arrowHeight = 1.5;
        let arrowWidth = 6;
        s = p.path("M" + start + "," + (y - padding) + "L" + (end - arrowWidth) + "," + (arrowHeight + y) + "L" + (end - arrowWidth) + "," + (y - padding) + "L" + (end + padding) + "," + (height / 2 + y) + "L" + (end - arrowWidth) + "," + (height + y + padding) + "L" + (end - arrowWidth) + "," + (height - arrowHeight + y) + "L" + start + "," + (height + y + padding) + "L" + start + "," + (y - padding));
        s.attr({
            "stroke-width": .7,
            fill: "#DDD",
            "fill-opacity": .3,
            stroke: "#226",
            "stroke-dasharray": "5 2"
        });
        return s;
    }
    function drawHelix(p, y, start, end, spanWidth, height, rightLimit) {
        y += 2;
        height -= 4;
        padding = 2;
        start = Math.max(-spanWidth, Math.round(start));
        end = Math.min(rightLimit, Math.round(end + spanWidth));
        h = p.rect(start, y - padding, end - start, height + 2 * padding, 4);
        h.attr({
            "stroke-width": .7,
            fill: "#DDD",
            "fill-opacity": .3,
            stroke: "#226"
        });
        return h;
    }
    function updateExportText() {
        let txt = "";
        if ($("#alignmentFormatSelect").val() == "Multi FASTA") {
            $(".alignTbl:visible").each(function() {
                txt += getFormattedAlignment($(this).attr("id").substring(8), "FASTA");
                txt += "\n";
            });
        } else {
            txt = getFormattedAlignment($("#currentAlignmentIndex").val(), $("#alignmentFormatSelect").val());
        }
        $("#formattedAlignmentTextarea").val(txt);
        $("#formattedAlignmentTextarea").focus();
        $("#formattedAlignmentTextarea").select();
    }
    function getFormattedAlignment(alignmentSuffix, format) {
        if (!format) format = "fasta";
        let $nmTbl = $("#alignNameTbl" + alignmentSuffix);
        let $alTbl = $("#alignTbl" + alignmentSuffix);
        let seqnos = [], names = [], sequences = [], desc = [];
        let vis = $("#alignTbl" + alignmentSuffix + " :visible").first().length > 0 ? ":visible" : "";
        $nmTbl.find("tr[seqno]" + vis).each(function(i, obj) {
            let seqno = parseInt(obj.getAttribute("seqno"));
            if (seqnos.length > 0 && seqno <= seqnos[0]) return false;
            seqnos.push(seqno);
            names.push($(this).text());
            desc.push($(this).attr("title"));
            breakCount = 0;
            let thisSeq = "";
            $alTbl.find(`tr[seqno=${seqno}]`).each(function() {
                thisSeq += $(this).find("td").first().text();
                breakCount += $(this).find(".insert_l,.insert_r").length;
            });
            sequences.push(thisSeq);
            if (breakCount > 0) names[i] = `MULTIPLE_SEGMENTS_OF_${names[i]} - WARNING - CONTAINS NON-SEQUENTIAL SEGMENTS`;
        });
        if (format.search(/fasta/i) > -1) return alignmentFASTA(names, sequences, desc); else if (format.search(/clustal/i) > -1) return alignmentClustalW(names, sequences); else return "";
    }
    function alignmentFASTA(names, sequences, desc) {
        if (!sequences || sequences.length == 0) return "No templates selected";
        let linelength = 72;
        let txt = "";
        for (let i = 0; i < names.length; i++) {
            txt += ">" + names[i] + (desc && desc[i] ? " " + desc[i] : "") + "\n";
            for (let j = 0; j < sequences[i].length; j += linelength) txt += sequences[i].substring(j, j + linelength) + "\n";
        }
        return txt;
    }
    function alignmentClustalW(names, sequences) {
        if (!sequences || sequences.length == 0) return "No templates selected";
        let linelength = 60;
        let spaces = 0;
        for (let i = 0; i < names.length; i++) spaces = Math.max(spaces, names[i].length);
        spaces += 2;
        let txt = "CLUSTAL\n\n";
        for (let j = 0; j < sequences[0].length; j += linelength) {
            for (let i = 0; i < names.length; i++) {
                txt += names[i];
                for (let s = names[i].length; s < spaces; s++) txt += " ";
                txt += sequences[i].substring(j, j + linelength) + "\n";
            }
            txt += "\n";
        }
        txt = txt.substring(0, txt.length - 2);
        return txt;
    }
    function html2svg($alignTbl) {
        $alignNameTbl = getAlignNameTbl($alignTbl);
        let svgWidth = $alignTbl.closest(".alignTblContainer").outerWidth();
        let svgHeight = $alignTbl.closest(".alignTblContainer").outerHeight();
        let offset = 10;
        svgWidth += offset;
        let svg = '<svg xmlns="http://www.w3.org/2000/svg" ' + ' width="' + svgWidth + '" height="' + svgHeight + '">' + '<rect x="0" y="0" fill="white" width="' + svgWidth + '" height="' + svgHeight + '"/>';
        let y;
        let nameWidth = $alignNameTbl.outerWidth() + offset;
        let alOffsetLeft = $alignNameTbl.offset().left - offset;
        let $td = $alignTbl.find("td:visible:eq(0)");
        let tblWidth = $td.outerWidth();
        let featuresvg = $("<div>").append($("#" + $alignTbl.attr("id") + "_features" + " svg").first().clone()).html().replace(/name="[\d_]+" /g, "").replace(/r="4" /g, "");
        svg += '<defs><clipPath id="cp2"><rect x="0" y="0" width="' + tblWidth + '" height="' + svgHeight + '"/></clipPath></defs>';
        svg += '<g transform="translate(' + nameWidth + ',0)" clip-path="url(#cp2)">';
        svg += featuresvg.substring(featuresvg.indexOf("</defs>") + 7, featuresvg.indexOf("</svg>"));
        svg += "</g>";
        let $rows = $alignTbl.find("tr:visible");
        $rows.each(function(i, tr) {
            $(this).find("span.ligandContact").each(function(j, sp) {
                y = tr.offsetTop + sp.offsetTop;
                if (!outerWidth) {
                    outerWidth = $(sp).outerWidth();
                    outerHeight = $(sp).outerHeight();
                }
                svg += '<rect style="fill:#CC4444;fill-opacity:0.8;" ' + 'width="' + outerWidth + '" height="' + outerHeight + '" ' + 'x="' + ($(sp).offset().left - alOffsetLeft) + '" y="' + y + '" />';
            });
        });
        svg += '<text text-anchor="end" font-size="' + $td.css("font-size") + "\" font-family='" + $td.css("font-family") + "'>";
        $rows = $alignNameTbl.find("tr:visible");
        $rows.each(function(i, tr) {
            let padding = $(this).find("td").css("padding-top");
            padding = parseInt(padding.substring(0, padding.length - 2));
            y = tr.offsetTop + 12 + padding;
            svg += '<tspan x="' + (nameWidth - 2) + '" y="' + y + '">' + tr.textContent.replace("QMEAN", "") + "</tspan>";
        });
        svg += "</text>";
        $rows = $alignTbl.find("tr:visible");
        let outerWidth, outerHeight;
        $rows.each(function(i, tr) {
            svg += '<text text-anchor="middle" font-size="' + $td.css("font-size") + "\" font-family='" + $td.css("font-family") + "'>";
            $(this).find("span").each(function(j, sp) {
                if (!outerWidth) {
                    outerWidth = $(sp).outerWidth();
                    outerHeight = $(sp).outerHeight();
                }
                if (sp.style.position == "relative") return;
                let style = $(sp).css("color");
                style = "fill:" + style + ";";
                if (sp.style.fontWeight == "bold") style += "font-weight:bold;";
                if (style.length > 0) style = ' style="' + style + '" ';
                y = tr.offsetTop + sp.offsetTop + 12;
                svg += '<tspan x="' + ($(sp).offset().left - alOffsetLeft + outerWidth / 2) + '" y="' + y + '"' + style + ">" + sp.textContent + "</tspan>";
            });
            svg += "</text>";
            td = $(this).find("td")[1];
            if (td) svg += '<text text-anchor="start" font-size="' + $td.css("font-size") + "\" font-family='" + $td.css("font-family") + "'><tspan x=\"" + ($(td).offset().left - alOffsetLeft + 10) + '" y="' + y + '">' + td.textContent + "</tspan></text>";
        });
        $rows.each(function(i, tr) {
            $(this).find("span").not(".ligandContact").each(function(j, sp) {
                y = tr.offsetTop + sp.offsetTop;
                let bg = $(sp).css("backgroundColor");
                if (bg && bg != "transparent" && bg != "rgb(255, 255, 255)") {
                    svg += '<rect x="' + ($(sp).offset().left - alOffsetLeft) + '" y="' + y + '" width="' + outerWidth + '" height="' + outerHeight + '" style="fill:' + bg + ';"/>';
                }
            });
        });
        return svg + "</svg>";
    }
    $.fn.alignment = function(method) {
        if (publicMethods[method]) return publicMethods[method].apply(this, Array.prototype.slice.call(arguments, 1)); else if (typeof method === "object" || !method) return publicMethods.init.apply(this, arguments); else $.error("Method " + method + " does not exist on alignment.min.js");
    };
})(jQuery);

if (!Object.keys) {
    Object.keys = function(obj) {
        let keys = [];
        for (let i in obj) {
            if (obj.hasOwnProperty(i)) {
                keys.push(i);
            }
        }
        return keys;
    };
}

$(document).on("click", ".alignmentOptions input[name=csRadioGroup]", function() {
    delete chainMapping.defaultColourScheme;
    $(".alignTbl:first").alignment("colourAllAlignments", {
        col: $(this).val()
    });
});

$(document).on("click", "#viewerSelect input[name=3dRG]", function() {
    showViewer($(this).val());
});

function highlightSelectedResidues() {
    document.querySelectorAll(".hoverResidue").forEach(el => el.classList.remove("hoverResidue"));
    document.querySelectorAll(".selectResidue").forEach(el => el.classList.remove("selectResidue"));
    for (let struc_id in chainMapping.structures) for (let chn in chainMapping.structures[struc_id]) {
        let selres = chainMapping.structures[struc_id][chn].filter(r => r.selected).map(r => r.resno);
        if (selres.length > 0) {
            let rows = document.querySelectorAll('.alignTbl tr[struc_id^="' + struc_id + '"][chain="' + chn + '"]');
            for (let res of selres) {
                for (let row of rows) {
                    let span = row.querySelector('span[s="' + res + '"]');
                    if (span) {
                        span.classList.add("selectResidue");
                        break;
                    }
                }
            }
        }
    }
}

document.addEventListener("DOMContentLoaded", () => {
    document.body.addEventListener("highlightResidue", e => {
        if (e.detail.src === "alignment") return;
        if (!e.detail.multiple) document.querySelectorAll(".hoverResidue").forEach(el => el.classList.remove("hoverResidue"));
        for (let row of document.querySelectorAll(`.alignTbl tr[struc_id^="${e.detail.struc_id}"]`)) {
            if (!!(row.offsetWidth || row.offsetHeight || row.getClientRects().length)) {
                if (row.hasAttribute("chain") && row.getAttribute("chain") === e.detail.chain || row.hasAttribute("keyseqchain") && row.getAttribute("keyseqchain") === e.detail.chain) {
                    let span = row.querySelector(`span[s="${e.detail.resno}"]`);
                    if (span) {
                        span.classList.add("hoverResidue");
                        break;
                    }
                }
            }
        }
    });
    document.body.addEventListener("selectResidues", e => {
        if ([ "alignment", "index", "comparison", "variance" ].includes(e.detail.src)) for (let chn in chainMapping.structures[e.detail.struc_id]) {
            let selres = e.detail.residues.filter(r => r.chain === chn).map(r => r.resno);
            for (let i = 0; i < chainMapping.structures[e.detail.struc_id][chn].length; i++) {
                let res = chainMapping.structures[e.detail.struc_id][chn][i];
                if (res) {
                    if (selres.includes(i)) res.selected = true; else delete res.selected;
                }
            }
        }
        if ([ "alignment", "structure", "rama", "outlier" ].includes(e.detail.src)) return;
        highlightSelectedResidues();
    });
    document.body.addEventListener("selectResidues", e => {
        if ($("#annotate:visible").length === 0 || [ "structure", "rama", "outlier" ].includes(e.detail.src) || !e.detail.struc_id || !e.detail.residues || e.detail.residues.length === 0) return;
        let mdl = modelData[e.detail.struc_id.substring(6)];
        if (!mdl) return;
        let possibleRanges = [];
        for (let residue of e.detail.residues) {
            let range = possibleRanges.find(r => r.chain === residue.chain && (r.min - 1 === residue.resno || r.max + 1 === residue.resno));
            let res = chainMapping.structures[e.detail.struc_id][residue.chain][residue.resno];
            if (!res) continue;
            if (range) {
                if (range.max < residue.resno) {
                    range.max = residue.resno;
                    range.sequence += res.resname;
                }
                if (range.min > residue.resno) {
                    range.min = residue.resno;
                    range.sequence = res.resname + range.sequence;
                }
            } else possibleRanges.push({
                chain: residue.chain,
                min: residue.resno,
                max: residue.resno,
                sequence: res.resname
            });
        }
        let annovalue = document.getElementById("annotate").value;
        for (let t = 0; t < mdl.targets.length; t++) {
            let chains = mdl.targets[t].id ? mdl.targets[t].id : mdl.targets[t].keyseq_chain_name;
            for (let c of chains) {
                for (let range of possibleRanges.filter(r => r.chain === c)) {
                    let cmd = `motif=${range.sequence} start=${range.min} end=${range.max}`;
                    if (mdl.targets.length > 1) cmd += ` target=${t + 1}`;
                    if (annovalue.includes(cmd)) continue;
                    if (annovalue.length > 0 && !annovalue.endsWith("\n")) annovalue += "\n";
                    annovalue += cmd + "\n";
                }
            }
        }
        document.getElementById("annotate").value = annovalue;
        document.getElementById("annotate").scrollTop = document.getElementById("annotate").scrollHeight;
    });
});

$("body").on("click", function(e) {
    $(".poponthis").each(function() {
        if (!$(this).is(e.target) && $(this).has(e.target).length === 0 && $(".popover").has(e.target).length === 0) {
            (($(this).popover("hide").data("bs.popover") || {}).inState || {}).click = false;
        }
    });
});

function addSettingsPopover() {
    $("a[id^=options_]").css("visibility", "visible").each(function() {
        $(this).popover({
            html: true,
            placement: $(this).attr("pop-placement") ? $(this).attr("pop-placement") : "auto",
            content: $("#alignmentOptions"),
            container: $(this).closest("#viewerWrapper").length == 1 ? "#right-panel" : "body"
        });
    }).on("show.bs.popover", function(e) {
        let alignId = $(this).attr("id").substring(8);
        $("#currentAlignmentIndex").val(alignId);
        if ($("#exportOptions").length == 1) {
            $("#alignmentOptions").show();
            if (alignId == "") $("#exportOptions").hide(); else $("#exportOptions").show();
        } else $("a[id^=options_]").each(function() {
            if ($(this).data("bs.popover") && $(this).data("bs.popover").$tip && $(this).data("bs.popover").$tip.find("#exportOptions").length == 1) {
                if (alignId == "") $(this).data("bs.popover").$tip.find("#exportOptions").hide(); else $(this).data("bs.popover").$tip.find("#exportOptions").show();
            }
        });
    });
}

addSettingsPopover();

if ($("#bgColSlider").length > 0) $("#bgColSlider").slider({
    value: $.cookie("bg_r") ? $.cookie("bg_r") : 255,
    min: 0,
    max: 255,
    step: 1,
    slide: function(event, ui) {
        pv_viewer.options("background", [ ui.value / 255, ui.value / 255, ui.value / 255 ]);
        if (ngl_stage) ngl_stage.setParameters({
            backgroundColor: "rgb(" + ui.value + "," + ui.value + "," + ui.value + ")"
        });
        jQuery.cookie("bg_r", ui.value, {
            expires: 365,
            path: "/"
        });
    },
    change: function(event, ui) {
        pv_viewer.options("background", [ ui.value / 255, ui.value / 255, ui.value / 255 ]);
        if (ngl_stage) ngl_stage.setParameters({
            backgroundColor: "rgb(" + ui.value + "," + ui.value + "," + ui.value + ")"
        });
        jQuery.cookie("bg_r", ui.value, {
            expires: 365,
            path: "/"
        });
        if (ui.value < 180) $(".viewerStatus").addClass("status_dark"); else $(".viewerStatus").removeClass("status_dark");
    }
});

if ($.cookie("bg_r") && $.cookie("bg_r") < 180) $(".viewerStatus").addClass("status_dark");

function resetChainMappingColours() {
    for (let s in chainMapping.structures) for (let c in chainMapping.structures[s]) for (let r in chainMapping.structures[s][c]) delete chainMapping.structures[s][c][r].rgb;
}

var chainMapping = {
    structures: {},
    targets: {},
    alignments: [],
    chainMeta: {},
    annotations: [ {
        key: "qm",
        label: "Confidence",
        active: true
    }, {
        key: "qmb",
        label: "QMEANBrane",
        active: false
    }, {
        key: "c",
        label: "Consistency",
        active: true
    }, {
        key: "entropy",
        label: "Entropy",
        active: true
    }, {
        key: "soa",
        label: "Solvent Accessibility",
        active: true
    }, {
        key: "bfactor",
        label: "B-factor",
        active: true
    }, {
        key: "bfactor_range",
        label: "B-factor range",
        active: true
    } ]
};

$("#annotate").on("keyup", function(e) {
    if (e.key === "Enter") processAnnotations();
});

function processAnnotations(id) {
    if ($("#pv:visible", currentWin.document).length > 0) {
        pv_viewer.rm(`model_${id ? id : "*"}.label`);
        pv_viewer.rm(`model_${id ? id : "*"}.selected`);
    }
    if ($("#ngl:visible", currentWin.document).length > 0) {
        for (const mkey in modelData) {
            if (id && mkey != id) continue;
            let comp = getNGLComponentByName("model_" + mkey);
            if (comp) {
                let removeThese = [];
                for (let x = 0; x < comp.reprList.length; x++) if (comp.reprList[x].name.indexOf(".label") > -1 || comp.reprList[x].name.indexOf(".selected") > -1) removeThese.push(comp.reprList[x]);
                for (let x = 0; x < removeThese.length; x++) comp.removeRepresentation(removeThese[x]);
            }
        }
    }
    let processed = [];
    let lines = $("#annotate").val().split(/\r?\n/);
    let fts = {};
    for (let i = 0; i < lines.length; i++) {
        let motif, rgb, target, chain, start, end, label, labelCol, labelScale, sidechains = false;
        lines[i] = lines[i].trim().replace(/\s{1,}/g, " ").replace(/\s*= */g, "=");
        let line = lines[i];
        if (line.search(/label=\"(.+)\"/i) > -1) {
            let match = line.match(/label=\"(.+)\" ?/i);
            label = match[1];
            line = line.replace(match[0], "").trim();
        }
        let invalid = false;
        let cmds = line.split(" ");
        for (let c = 0; c < cmds.length; c++) {
            let cmd = cmds[c].split("=");
            if (cmd.length != 2) {
                invalid = true;
                continue;
            }
            switch (cmd[0].toUpperCase()) {
              case "MOTIF":
                motif = cmd[1];
                break;

              case "COLOUR":
              case "COLOR":
                rgb = new NGL.Color(cmd[1].toLowerCase());
                rgb = rgb.r * 255 + "," + rgb.g * 255 + "," + rgb.b * 255;
                break;

              case "TARGET":
                target = parseInt(cmd[1]);
                break;

              case "CHAIN":
                chain = cmd[1];
                break;

              case "START":
                start = parseInt(cmd[1]);
                break;

              case "END":
                end = parseInt(cmd[1]);
                break;

              case "LABEL":
                label = cmd[1];
                break;

              case "LABEL_SCALE":
                labelScale = cmd[1];
                break;

              case "LABEL_COLOR":
              case "LABEL_COLOUR":
                labelCol = new NGL.Color(cmd[1].toLowerCase());
                labelCol = labelCol.r * 255 + "," + labelCol.g * 255 + "," + labelCol.b * 255;
                break;

              case "SIDECHAINS":
                sidechains = cmd[1] == "on";
                break;

              default:
                invalid = true;
            }
        }
        if (motif) {
            let validRX = false;
            try {
                rxtest = RegExp(motif, "g");
                validRX = true;
            } catch (e) {}
            if (!validRX) {
                console.error("Motif regex is invalid");
                invalid = true;
            }
        }
        if (rgb == null && label == null && !sidechains) invalid = true;
        if (invalid) {
            if (lines[i].length > 0 && lines[i].charAt(0) != "?") processed.push("?" + lines[i]); else processed.push(lines[i]);
            continue;
        } else processed.push(lines[i]);
        for (const mkey in modelData) {
            if (modelData[mkey].status && modelData[mkey].status != "COMPLETED") continue;
            let viewSelect = {};
            for (let t = 0; t < modelData[mkey].targets.length; t++) {
                if (target && target != t + 1) continue;
                let trg = modelData[mkey].targets[t];
                let trgSeq = trg.target_sequence;
                let chainSeq = trg.id;
                if (!trgSeq) {
                    trgSeq = trg.templates[0].trg_seq;
                    if (trg.templates[0].tpl_seq) trgSeq = trgSeq.replaceAll("-", "");
                    chainSeq = trg.keyseq_chain_name;
                    if (trg.templates[0].trg_offset) trgSeq = "-".repeat(trg.templates[0].trg_offset - 1) + trgSeq;
                }
                if (typeof chainSeq !== "string") chainSeq = chainSeq.join("");
                if (!fts["model_" + mkey]) fts["model_" + mkey] = [];
                let targetMotif = motif;
                if (targetMotif == null) targetMotif = trgSeq.substring(start ? start - 1 : 0, end ? end : trgSeq.length);
                for (let c = 0; c < chainSeq.length; c++) {
                    let chn = chainSeq.charAt(c);
                    if (chain == null || chn == chain) {
                        if (!viewSelect[chn]) viewSelect[chn] = {
                            labelled: [],
                            selected: []
                        };
                        let rx = RegExp(targetMotif, "g");
                        for (const m of trgSeq.matchAll(rx)) {
                            let resStart = m.index + 1;
                            let resEnd = resStart + m[0].length - 1;
                            if (m.length == 4) {
                                resStart += m[1].length;
                                resEnd = resStart + m[2].length - 1;
                            }
                            if (start && start > resEnd) continue;
                            if (end && end < resStart) continue;
                            if (start) resStart = Math.max(resStart, start);
                            if (end) resEnd = Math.min(resEnd, end);
                            if (rgb) fts["model_" + mkey].push({
                                frmChn: chn,
                                toChn: chn,
                                resStart: resStart,
                                resEnd: resEnd,
                                rgb: rgb
                            });
                            if (label || sidechains) {
                                if (label && label.toLowerCase() == "default") label = "default";
                                let midRes = Math.floor((resEnd - resStart) / 2) + resStart;
                                for (let x = resStart; x <= resEnd; x++) {
                                    if (sidechains) viewSelect[chn].selected.push(x);
                                    if (label == "default") viewSelect[chn].labelled.push(x);
                                }
                                if (label && label != "default") viewSelect[chn].labelled.push(midRes);
                            }
                        }
                    }
                }
            }
            if ((label || sidechains) && (id == null || id == mkey)) {
                if ($("#pv:visible", currentWin.document).length > 0) {
                    let struc = pv_viewer.get("model_" + mkey);
                    if (struc) {
                        let selRes = struc.structure().residueSelect(function(res) {
                            if (viewSelect[res.chain().name()] && (viewSelect[res.chain().name()].selected.indexOf(res.num()) > -1 || viewSelect[res.chain().name()].labelled.indexOf(res.num()) > -1)) {
                                if (viewSelect[res.chain().name()].labelled.indexOf(res.num()) > -1) res.eachAtom(function(atom) {
                                    const r0 = atom.residue();
                                    if (atom.name() == "CA") pv_viewer.label(`model_${mkey}.label`, label != "default" ? label : r0.name() + " " + r0.chain().name() + r0.num(), atom.pos(), {
                                        fontSize: 12 * (labelScale ? labelScale : 1),
                                        fontColor: labelCol ? `rgb(${labelCol})` : "#CCC"
                                    });
                                });
                                return viewSelect[res.chain().name()].selected.indexOf(res.num()) > -1;
                            }
                        });
                        if (selRes) {
                            pv_viewer.ballsAndSticks(`model_${mkey}.selected`, selRes);
                        }
                    }
                } else if ($("#ngl:visible", currentWin.document).length > 0) {
                    let comp = getNGLComponentByName("model_" + mkey);
                    if (comp) {
                        let nglSel = "";
                        let hasLabels = false;
                        for (let chn in viewSelect) {
                            if (viewSelect[chn].labelled.length == 0) continue;
                            if (nglSel.length > 0) nglSel += " or ";
                            nglSel += "( :" + chn + " and (";
                            let lastRes = null;
                            for (let i = 0; i < viewSelect[chn].labelled.length; i++) {
                                hasLabels = true;
                                if (lastRes && viewSelect[chn].labelled[i] == lastRes + 1) {
                                    lastRes = viewSelect[chn].labelled[i];
                                    continue;
                                }
                                if (lastRes) nglSel += "-" + lastRes;
                                if (i > 0) nglSel += " or ";
                                nglSel += viewSelect[chn].labelled[i];
                                lastRes = viewSelect[chn].labelled[i];
                            }
                            if (lastRes) nglSel += "-" + lastRes;
                            nglSel += ") and .CA) ";
                        }
                        if (hasLabels) {
                            comp.addRepresentation("label", {
                                name: "model_" + mkey + ".label",
                                sele: nglSel,
                                labelType: "format",
                                labelFormat: label != "default" ? label : "%(resname)s %(chainname)s%(resno)s",
                                fontWeight: "bold",
                                radiusScale: 1.4 * (labelScale ? labelScale : 1),
                                xOffset: .3,
                                color: labelCol ? `rgb(${labelCol})` : "#909090"
                            });
                        }
                        nglSel = "";
                        let hasSelection = false;
                        for (let chn in viewSelect) {
                            if (viewSelect[chn].selected.length == 0) continue;
                            if (nglSel.length > 0) nglSel += " or ";
                            nglSel += "( :" + chn + " and (";
                            let lastRes = null;
                            for (let i = 0; i < viewSelect[chn].selected.length; i++) {
                                hasSelection = true;
                                if (lastRes && viewSelect[chn].selected[i] == lastRes + 1) {
                                    lastRes = viewSelect[chn].selected[i];
                                    continue;
                                }
                                if (lastRes) nglSel += "-" + lastRes;
                                if (i > 0) nglSel += " or ";
                                nglSel += viewSelect[chn].selected[i];
                                lastRes = viewSelect[chn].selected[i];
                            }
                            if (lastRes) nglSel += "-" + lastRes;
                            nglSel += ")) ";
                        }
                        if (hasSelection) {
                            comp.addRepresentation("ball+stick", {
                                name: "model_" + mkey + ".selected",
                                sele: nglSel,
                                colorScheme: "element"
                            });
                        }
                    }
                }
            }
        }
    }
    if (id) return;
    let newText = processed.join("\n");
    if (typeof annotationsURL !== "undefined" && newText.search(/^\?/m) == -1) {
        $.ajax({
            type: "POST",
            url: annotationsURL,
            data: {
                annotations: newText,
                csrfmiddlewaretoken: $.cookie("csrftoken")
            },
            success: function(data) {},
            error: function(response) {
                console.error(response);
            }
        });
    }
    chainMapping.defaultColourScheme = "features";
    if ($("#annotate").val() != newText) $("#annotate").val(newText);
    $(".alignTbl:first").alignment("colourAllAlignments", {
        features: fts
    });
}

window.onload = () => {
    const selectedResiduesRef = {
        current: {}
    };
    document.body.addEventListener("selectResidues", function(e) {
        if (e.detail.src === "index") return;
        if ([ "structure", "rama", "outlier" ].includes(e.detail.src)) {
            let newSelection = {
                ...selectedResiduesRef.current
            };
            if (newSelection && newSelection[e.detail.struc_id]) {
                let resIdx = newSelection[e.detail.struc_id].findIndex(r => r.chain === e.detail.residues[0].chain && r.resno === e.detail.residues[0].resno);
                if (!e.detail.ctrl) {
                    if (resIdx === -1) newSelection[e.detail.struc_id] = e.detail.residues; else newSelection[e.detail.struc_id] = [];
                } else {
                    if (resIdx === -1) newSelection[e.detail.struc_id] = newSelection[e.detail.struc_id].concat(e.detail.residues); else {
                        for (let res of e.detail.residues) {
                            resIdx = newSelection[e.detail.struc_id].findIndex(r => r.chain === res.chain && r.resno === res.resno);
                            newSelection[e.detail.struc_id].splice(resIdx, 1);
                        }
                    }
                }
            } else newSelection = {
                [e.detail.struc_id]: e.detail.residues
            };
            selectedResiduesRef.current = newSelection;
            document.body.dispatchEvent(new CustomEvent("selectResidues", {
                detail: {
                    src: "index",
                    struc_id: e.detail.struc_id,
                    residues: newSelection[e.detail.struc_id]
                }
            }));
        } else selectedResiduesRef.current = {
            [e.detail.struc_id]: e.detail.residues
        };
    });
};

var currentWin = window;

let main_NGL_module = null;

let main_pv = null;

let main_ngl = null;

let plip_bond_num_dash = 5;

let plip_bond_radius = .1;

let plip_higlight_factor = 2;

let plip_higlight_bond_num_dash = 1;

var PLIP = PLIP || {};

PLIP.AbstractViewer = null;

let opacity_fadeout_ligand = .3;

let opacity_fadeout_protein = .3;

function takeSnapshot(filename, scale_factor) {
    if (filename == ".png") filename = "structure.png";
    if ($("#pv:visible", currentWin.document).length > 0) {
        $("#pv canvas")[0].toBlob(function(blob) {
            saveAs(blob, filename);
        });
    }
    if ($("#ngl:visible", currentWin.document).length > 0) {
        scale_factor = scale_factor || 1;
        var transparent = $("#snapshot_bg_transparent_checkbox").is(":checked");
        ngl_stage.makeImage({
            factor: scale_factor,
            antialias: false,
            trim: false,
            transparent: transparent
        }).then(function(blob) {
            NGL.download(blob, filename);
        });
    }
}

function resetStructure() {
    $(".hoverResidue").removeClass("hoverResidue");
    $(".selectResidue").removeClass("selectResidue");
    if ($("#pv:visible", currentWin.document).length > 0) {
        pv_viewer.show("*ligands");
        pv_viewer.hide("*ligands_lines");
        pv_viewer.rm("*res");
        pv_viewer.autoZoom();
        pv_viewer.requestRedraw();
    }
    if ($("#ngl:visible", currentWin.document).length > 0) {
        ngl_stage.eachRepresentation(function(rep) {
            if (rep.name.indexOf("hover") > -1 && rep.repr.selection.string != "none") {
                rep.setSelection("none");
            }
        });
        ngl_stage.autoView(1200);
    }
}

function spin(doSpin) {
    if (typeof pv_viewer !== "undefined") pv_viewer.spin(doSpin);
    if (ngl_stage != null) {
        if (doSpin) ngl_stage.setSpin([ 0, 1, 0 ], -.005); else ngl_stage.setSpin(0);
    }
}

function detach(url) {
    if (currentWin != window) return;
    if ($("#pv:visible", currentWin.document).length > 0) {
        currentWin = window.open(url.replace("xx", "pv"), "_blank", "resizable=1,status=1,toolbar=1,location=1,height=700,width=700");
        main_pv = pv_viewer;
    } else if ($("#ngl:visible", currentWin.document).length > 0) {
        main_ngl = ngl_stage;
        main_NGL_module = NGL;
        currentWin = window.open(url.replace("xx", "ngl"), "_blank", "resizable=1,status=1,toolbar=1,location=1,height=700,width=700");
    } else return;
    if ($("#left-panel").is(".ui-resizable")) $("#left-panel").resizable("destroy");
    $("#left-panel").removeClass().addClass("detached");
    $("#right-panel").hide();
    $(window).trigger("resize");
}

function attachPV() {
    pv_viewer = main_pv;
    attach();
}

function attachNGL() {
    let detachedOrientationMatrix = ngl_stage.viewerControls.getOrientation();
    NGL = main_NGL_module;
    ngl_stage = main_ngl;
    ngl_stage.removeAllComponents();
    showNGL(detachedOrientationMatrix);
    attach();
}

function attach() {
    currentWin = window;
    let classList = $("#right-panel").attr("class");
    if (classList) {
        classList = classList.split(/\s+/);
        for (let i = 0; i < classList.length; i++) {
            size = classList[i].split("-");
            if (size.length == 3) $("#left-panel").addClass(size[0] + "-" + size[1] + "-" + (12 - parseInt(size[2])));
        }
    }
    $("#left-panel").removeClass("detached").resizable({
        handles: "e",
        stop: function() {}
    });
    $("#right-panel").show();
    $(window).trigger("resize");
    if (typeof superposeTemplates != "undefined") superposeTemplates(true);
    refreshBondsAndLigandHighlights();
}

function refreshBondsAndLigandHighlights(strucId) {
    refreshHighlightLigands(strucId);
    refreshBonds(strucId);
}

function refreshBonds(strucId) {
    if (strucId === undefined) {
        plip_bonds.refreshAll();
    } else {
        plip_bonds.refresh(strucId);
    }
}

function refreshHighlightLigands(strucId) {
    if (strucId === undefined) {
        all_ligands.redrawAll();
    } else {
        all_ligands.redraw(strucId);
    }
}

function centerOnLigand(ligands, strucId) {
    if ($("#pv:visible", currentWin.document).length > 0) centerOnLigandPV(ligands, strucId);
    if ($("#ngl:visible", currentWin.document).length > 0) centerOnLigandNGL(ligands, strucId);
}

function highlightLigand(ligands, contacts, strucId) {
    $(".ligandContact").removeClass("ligandContact");
    var allHighlights = {};
    for (var i = 0; i < contacts.length; i++) {
        if (!contacts[i] || contacts[i].charAt(0) == "_") continue;
        res = contacts[i].substring(1);
        chain = contacts[i].charAt(0);
        if (!allHighlights[chain]) allHighlights[chain] = [];
        var obj = allHighlights[chain];
        obj.push(res);
    }
    let $rows = $(".alignTbl tr[struc_id=" + strucId + "]");
    let filter = $rows[0].hasAttribute("keyseqchain") ? "keyseqchain" : "chain";
    for (var chn in allHighlights) {
        let spanArray = $rows.filter("[" + filter + "=" + chn + "]").find("span").toArray();
        for (var i = 0; i < allHighlights[chn].length; i++) {
            for (var s = 0; s < spanArray.length; s++) {
                if (spanArray[s].getAttribute("s") == allHighlights[chn][i]) {
                    spanArray[s].className = "ligandContact";
                    break;
                }
            }
        }
    }
    if ($("#pv:visible", currentWin.document).length > 0) highlightLigandPV(allHighlights, contacts, strucId);
    if ($("#ngl:visible", currentWin.document).length > 0) highlightLigandNGL(ligands, contacts, strucId);
}

document.addEventListener("DOMContentLoaded", () => {
    document.body.addEventListener("highlightResidue", e => {
        if ([ "structure", "varianceChart" ].includes(e.detail.src)) return;
        highlightResidues(e.detail.chain, e.detail.resno, e.detail.struc_id, e.detail.multiple);
    });
    document.body.addEventListener("selectResidues", e => {
        if ([ "structure", "rama", "outlier" ].includes(e.detail.src)) return;
        selectResidues(e);
    });
    document.body.addEventListener("centerResRange", e => {
        if ($("#pv:visible", currentWin.document).length > 0) centerResRangePV(e.detail.chain, e.detail.min, e.detail.max, e.detail.struc_id, e.detail.isDisulfid, e.detail.secondChain);
        if ($("#ngl:visible", currentWin.document).length > 0) centerResRangeNGL(e.detail.chain, e.detail.min, e.detail.max, e.detail.struc_id, e.detail.isDisulfid, e.detail.secondChain);
    });
    document.body.addEventListener("connectResidues", e => {
        if ($("#pv:visible", currentWin.document).length > 0) connectResiduesPV(e.detail.residuePairs);
        if ($("#ngl:visible", currentWin.document).length > 0) connectResiduesNGL(e.detail.residuePairs);
    });
    document.body.addEventListener("mousemove", e => {
        return e.target.matches("#pv canvas") ? doPVMouse(e) : null;
    }, true);
    document.body.addEventListener("click", e => {
        return e.target.matches("#pv canvas") ? doPVMouse(e) : null;
    }, true);
});

function highlightResidues(chain, res_pos, strucId, multiple) {
    if (chain == null && res_pos == null || strucId == null) {
        if ($("#pv:visible", currentWin.document).length > 0) {
            pv_viewer.rm("*_res");
            pv_viewer.requestRedraw();
        }
        if ($("#ngl:visible", currentWin.document).length > 0) {
            if (typeof all_ligands !== "undefined" && typeof plip_residues !== "undefined") highlightLigandNGL(all_ligands.getVisibleLigandIds(strucId), plip_residues.getVisibleResiduesResId(strucId), strucId, multiple); else highlightResiduesNGL([], [], strucId, multiple);
        }
    } else {
        if ($("#pv:visible", currentWin.document).length > 0) return highlightResiduesPV(chain, res_pos, strucId, multiple);
        if ($("#ngl:visible", currentWin.document).length > 0) return highlightResiduesNGL(chain, res_pos, strucId, multiple);
    }
}

function selectResidues(event) {
    if ($("#pv:visible", currentWin.document).length > 0) selectResiduesPV(event);
    if ($("#ngl:visible", currentWin.document).length > 0) selectResiduesNGL(event);
}

function changeRep(rep, show) {
    if (!rep) rep = "cartoon";
    if ($("#pv:visible", currentWin.document).length > 0) {
        changeRepPV(rep, show);
        if (pv_viewer != null && pv_viewer.all().length > 0) if (typeof detached === "undefined" || !detached) updatecols(); else window.opener.updatecols();
    }
    if ($("#ngl:visible", currentWin.document).length > 0) {
        changeRepNGL(rep, show);
        if (typeof detached === "undefined" || !detached) updatecols(); else window.opener.updatecols();
    }
}

function updatecolsSingleRes(singleResidue) {
    $(".alignTbl:first").alignment("updateColoursSingleResidue", singleResidue);
}

function updatecols() {
    $(".alignTbl:first").alignment("viewerInitialised");
}

function fetchSecStruc(strucId) {
    if ($("#pv:visible", currentWin.document).length > 0 && typeof pv_viewer !== "undefined" && pv_viewer !== null) {
        if (pv_viewer.get(strucId)) return fetchSecStrucPV(strucId, pv_viewer.get(strucId).structure());
    } else if ($("#ngl:visible", currentWin.document).length > 0 && typeof ngl_stage !== "undefined" && ngl_stage !== null) {
        if (getNGLComponentByName(strucId)) return fetchSecStrucNGL(strucId, getNGLComponentByName(strucId).structure);
    }
}

function fetchBFactors(strucId, storeAs, checkValidResidues) {
    if ($("#pv:visible", currentWin.document).length > 0 && typeof pv_viewer !== "undefined" && pv_viewer !== null) {
        if (pv_viewer.get(strucId)) return fetchBFactorsPV(strucId, pv_viewer.get(strucId).structure(), storeAs, checkValidResidues);
    } else if ($("#ngl:visible", currentWin.document).length > 0 && typeof ngl_stage !== "undefined" && ngl_stage !== null) {
        if (getNGLComponentByName(strucId)) return fetchBFactorsNGL(strucId, getNGLComponentByName(strucId).structure, storeAs, checkValidResidues);
    }
}

function setBFactorRange(strucId) {
    let mn = 1e10, mx = -1e10, rms = 0, count = 0;
    for (chn in chainMapping.structures[strucId]) for (res in chainMapping.structures[strucId][chn]) {
        if (!chainMapping.structures[strucId][chn][res] || !chainMapping.structures[strucId][chn][res].bfactor) continue;
        let val = chainMapping.structures[strucId][chn][res].bfactor;
        mn = Math.min(val, mn);
        mx = Math.max(val, mx);
        rms += val * val;
        count++;
    }
    rms /= count;
    rms = Math.sqrt(rms);
    mx = 2 * rms;
    let rangeColours = [];
    for (let x = 0; x < 256; x++) rangeColours.push([ x, x, 255 ]);
    for (let x = 256; x < 512; x++) rangeColours.push([ 255, 512 - x, 512 - x ]);
    let bidx = chainMapping.annotations.map(m => m.key).indexOf("bfactor_range");
    if (!chainMapping.annotations[bidx].rangeColours) chainMapping.annotations[bidx].rangeColours = {};
    chainMapping.annotations[bidx].rangeColours[strucId] = {};
    for (chn in chainMapping.structures[strucId]) {
        chainMapping.annotations[bidx].rangeColours[strucId][chn] = [];
        for (res in chainMapping.structures[strucId][chn]) {
            if (!chainMapping.structures[strucId][chn][res] || !chainMapping.structures[strucId][chn][res].bfactor) continue;
            let val = chainMapping.structures[strucId][chn][res].bfactor;
            let idx = Math.round(512 * (val - mn) / (mx - mn)) - 1;
            idx = Math.max(idx, 0);
            idx = Math.min(idx, 511);
            chainMapping.annotations[bidx].rangeColours[strucId][chn][res] = rangeColours[idx];
        }
    }
}

var ngl_stage = null;

var ngl_picked_atom = false;

var ngl_picked_component = false;

function getNGL() {
    let bgCol = "white";
    if ($.cookie("bg_r")) bgCol = "rgb(" + $.cookie("bg_r") + "," + $.cookie("bg_r") + "," + $.cookie("bg_r") + ")";
    ngl_stage = new NGL.Stage("ngl", {
        backgroundColor: bgCol,
        hoverTimeout: .005,
        sampleLevel: 2,
        tooltip: false
    });
    if (!ngl_stage.viewerControls) throw new Error("NGL viewerControls not initialised. How is that possible!?");
    let m = new NGL.Matrix4().makeRotationX(Math.PI / 2);
    let q = new NGL.Quaternion().setFromRotationMatrix(m);
    ngl_stage.viewerControls.rotate(q);
    m = new NGL.Matrix4().makeRotationY(Math.PI);
    q = new NGL.Quaternion().setFromRotationMatrix(m);
    ngl_stage.viewerControls.rotate(q);
    ngl_stage.signals.clicked.add(function(pickingProxy, a) {
        if (pickingProxy && (pickingProxy.atom || pickingProxy.bond)) {
            const atom = pickingProxy.atom || pickingProxy.closestBondAtom;
            if (atom.resno && atom.chainname) pickingProxy.component.autoView(atom.resno + ":" + atom.chainname, 1e3);
            document.body.dispatchEvent(new CustomEvent("selectResidues", {
                detail: {
                    src: "structure",
                    struc_id: pickingProxy.component.name,
                    residues: atom.resno && atom.chainname ? [ {
                        chain: atom.chainname,
                        resno: atom.resno
                    } ] : [],
                    ctrl: pickingProxy.ctrlKey || pickingProxy.metaKey
                }
            }));
        } else ngl_stage.autoView(1200);
    });
    ngl_stage.signals.hovered.add(function(pickingProxy) {
        if (ngl_stage.mouseObserver.pressed) return;
        if (pickingProxy && (pickingProxy.atom || pickingProxy.bond)) {
            let atom = pickingProxy.atom || pickingProxy.closestBondAtom;
            if (ngl_picked_atom && atom.structure.name == ngl_picked_atom.structure.name && atom.index == ngl_picked_atom.index) {
                return;
            } else {
                let cur_component = pickingProxy.component;
                let main_repr = getNGLRepresentationByName(cur_component.name);
                ngl_picked_atom = atom;
                ngl_picked_component = pickingProxy.component;
                let lbl = atom.resname + " " + atom.resno + atom.inscode;
                if (atom.chainname != "_") lbl += " " + atom.chainname;
                if (atom.chainname == "_" || ![ "cartoon", "tube", "rope" ].includes(main_repr.repr.type) || atom.atomname != "CA") lbl += " " + atom.atomname;
                lbl += resAnnoLabel(pickingProxy.component.name, atom.chainname, atom.resno);
                document.getElementById("ngl_status").innerHTML = lbl;
                document.body.dispatchEvent(new CustomEvent("highlightResidue", {
                    detail: {
                        src: "structure",
                        struc_id: pickingProxy.component.name,
                        chain: atom.chainname,
                        resno: atom.resno
                    },
                    bubbles: true
                }));
                if (typeof detached === "undefined" || !detached) updatecolsSingleRes({
                    struc_id: pickingProxy.component.name,
                    chain: atom.chainname,
                    resno: atom.resno
                }); else {
                    window.opener.ngl_picked_atom = ngl_picked_atom;
                    window.opener.updatecolsSingleRes({
                        struc_id: pickingProxy.component.name,
                        chain: atom.chainname,
                        resno: atom.resno
                    });
                }
            }
        } else {
            if (!ngl_picked_atom) {
                return;
            } else {
                ngl_picked_atom = false;
                document.getElementById("ngl_status").innerHTML = "";
                if (typeof detached === "undefined" || !detached) updatecolsSingleRes(); else {
                    window.opener.ngl_picked_atom = ngl_picked_atom;
                    window.opener.updatecolsSingleRes();
                }
            }
        }
    });
    return ngl_stage;
}

function getNGLComponentByName(name) {
    return ngl_stage.getComponentsByName(name).list[0];
}

function getNGLRepresentationByName(name) {
    return ngl_stage.getRepresentationsByName(name).list[0];
}

function removeNGLComponentByName(name) {
    let comp = getNGLComponentByName(name);
    ngl_stage.removeComponent(comp);
}

function highlightResiduesNGL(chain, res_pos, strucId, multiple) {
    let selected = "";
    if (strucId && strucId.endsWith(".dna")) strucId = strucId.replace(".dna", "");
    if (res_pos) {
        ngl_stage.eachRepresentation(function(rep) {
            if (rep.name.indexOf("hover") > -1 && rep.name.indexOf(strucId) == -1 && rep.repr.selection.string != "none") {
                rep.setVisibility(false);
            }
        });
        let repr = getNGLRepresentationByName(strucId + ".hover");
        if (!repr) return;
        repr.setVisibility(true);
        if (multiple) {
            selected = repr.repr.selection.string + " ";
            if (selected.startsWith("none")) selected = selected.substring(5);
        }
        let structure = getNGLComponentByName(strucId).structure;
        let chnString = "";
        for (let i = 0; i < chain.length; i++) {
            if (res_pos.length) {
                for (let j = 0; j < res_pos.length; j++) {
                    if (res_pos[j][0] == res_pos[j][1]) {
                        if (isResidueInStructureNGL(structure, chain[i], res_pos[j][0])) chnString += res_pos[j][0] + ":" + chain[i] + " ";
                    } else {
                        console.warn("Not checking if the selection is part of the structure");
                        chnString += res_pos[j][0] + "-" + res_pos[j][1] + ":" + chain[i] + " ";
                    }
                }
            } else if (isResidueInStructureNGL(structure, chain[i], res_pos)) chnString += res_pos + ":" + chain[i] + " ";
        }
        new_selection = selected + chnString;
        if (new_selection == "") new_selection = "none";
        repr.setSelection(new_selection);
        return new_selection;
    } else {
        let comp = getNGLComponentByName(strucId);
        let chnString = "";
        if (comp) {
            if (chain.length == 1) chnString = (res_pos ? res_pos : "") + ":" + chain; else for (let i = 0; i < chain.length; i++) chnString += (res_pos ? res_pos : "") + ":" + chain[i] + " ";
            comp.autoView(chnString, 1e3);
        }
        return chnString;
    }
}

function selectResiduesNGL(e) {
    let strucId = e.detail.struc_id;
    if (strucId == null || e.detail.residues == null || e.detail.residues.length === 0) {
        ngl_stage.eachRepresentation(rep => {
            if (rep.name.endsWith(".select") || rep.name.endsWith(".hover")) rep.setVisibility(false);
        });
        return;
    }
    let repr = getNGLRepresentationByName(strucId + ".select");
    if (!repr) return;
    repr.setVisibility(true);
    let newSelection = e.detail.residues.map(r => r.resno + ":" + r.chain).join(" ");
    repr.setSelection(newSelection);
    getNGLComponentByName(strucId).autoView(newSelection, 1e3);
    return newSelection;
}

function centerResRangeNGL(chain, start, end, strucId, isDisulfid, secondChain) {
    let chnString = "";
    strucId = strucId.replace(".dna", "");
    let structure = getNGLComponentByName(strucId);
    if (!structure) return;
    structure = structure.structure;
    for (let i = 0; i < chain.length; i++) {
        let lastRes = secondChain ? start : end;
        for (let p = start; p <= lastRes; p++) {
            if (isDisulfid && (p != start && p != lastRes)) continue;
            if (isResidueInStructureNGL(structure, chain[i], p)) chnString += p + ":" + chain[i] + " ";
        }
    }
    if (secondChain) for (let i = 0; i < secondChain.length; i++) if (isResidueInStructureNGL(structure, secondChain[i], end)) chnString += end + ":" + secondChain[i] + " ";
    if (chnString != "") getNGLComponentByName(strucId).autoView(chnString, 1e3);
}

function centerOnLigandNGL(ligands, strucId) {
    let ligString = ligands[0] + ":_";
    for (let i = 1; i < ligands.length; i++) {
        ligString += " " + ligands[i] + ":_";
    }
    for (let comp in ngl_stage.compList) {
        if (ngl_stage.compList[comp].name == strucId) {
            ngl_stage.compList[comp].autoView(ligString, 1e3);
            break;
        }
    }
}

function contactListToNGLSelection(strucId, contacts, ligands) {
    let b = {};
    let chn = null;
    let pos = null;
    if (ligands === undefined || ligands == null) {
        ligands = [];
    }
    if (contacts && contacts.length > 0) {
        let structure = getNGLComponentByName(strucId).structure;
        for (let i = 0; i < contacts.length; i++) {
            if (!contacts[i]) {
                continue;
            }
            chn = contacts[i].charAt(0);
            pos = parseInt(contacts[i].substring(1));
            if (chn == "_") {
                if (ligands.indexOf(pos) == -1) {
                    ligands.push(pos);
                }
            }
            if (isResidueInStructureNGL(structure, chn, pos) && chn != "-") {
                if (!b[chn]) b[chn] = [];
                b[chn][pos] = 1;
            }
        }
    }
    let contactString = "";
    let contactStringExtended = "";
    for (chn in b) {
        contactString += " ( (";
        contactStringExtended += " ( (";
        let firstInChain = true;
        for (let i = 0; i < b[chn].length; i++) {
            if (b[chn][i]) {
                if (!firstInChain) {
                    contactString += " or ";
                    contactStringExtended += " or ";
                } else firstInChain = false;
                contactString += i;
                contactStringExtended += i - 2;
                start = i;
                do {
                    i++;
                } while (b[chn][i]);
                if (i - start > 1) contactString += "-" + (i - 1);
                contactStringExtended += "-" + (i + 1);
            }
        }
        contactString += ") AND :" + chn + ") ";
        contactStringExtended += ") AND :" + chn + ") ";
    }
    return [ contactString, contactStringExtended, ligands ];
}

function isResidueInStructureNGL(structure, chainname, resno) {
    let ResidueFoundException = {};
    try {
        structure.eachChain(function(chain) {
            if (chain.chainname == chainname) {
                first_residue = chain.residueStore.resno[chain.residueOffset];
                last_residue = chain.residueStore.resno[chain.residueEnd];
                chain.eachResidue(function(residue) {
                    if (residue.resno == resno) {
                        throw ResidueFoundException;
                    }
                });
            }
        });
    } catch (e) {
        if (e === ResidueFoundException) {
            return true;
        }
    }
    return false;
}

function highlightLigandNGL(ligands, contacts, strucId, extra_residues, extra_ranges) {
    removeNGLComponentByName(strucId + "ligandBondsHL");
    if (extra_residues === undefined) {
        if (typeof extra_residues_selection === "undefined") {
            extra_residues = [];
        } else {
            extra_residues = extra_residues_selection;
        }
    }
    if (extra_ranges === undefined) {
        if (typeof extra_ranges_selection === "undefined") {
            extra_ranges = [];
        } else {
            extra_ranges = extra_ranges_selection;
        }
    }
    residue_contacts = contacts.concat(extra_residues);
    ranges_contacts = residue_contacts.concat(extra_ranges);
    let residuesStrings = contactListToNGLSelection(strucId, residue_contacts, ligands);
    ligands = residuesStrings[2];
    let rangesStrings = contactListToNGLSelection(strucId, ranges_contacts);
    let contactString = residuesStrings[0];
    let contactStringExtended = rangesStrings[1];
    if (contactString == "") {
        contactString = "none";
    }
    if (contactStringExtended == "") {
        contactStringExtended = "none";
        fadeOutRepresentationNGL(strucId, strucId, 1);
    } else {
        fadeOutRepresentationNGL(strucId, strucId, opacity_fadeout_protein);
    }
    for (let comp in ngl_stage.compList) {
        if (ngl_stage.compList[comp].name == strucId) {
            for (let repr in ngl_stage.compList[comp].reprList) {
                if (ngl_stage.compList[comp].reprList[repr].name == strucId + ".residuesHighlight") {
                    ngl_stage.compList[comp].reprList[repr].setSelection(contactString);
                }
                if (ngl_stage.compList[comp].reprList[repr].name == strucId + ".highlight") {
                    protein_residue_sele_string = "protein AND (" + contactStringExtended + ")";
                    ngl_stage.compList[comp].reprList[repr].setSelection(protein_residue_sele_string);
                }
            }
            break;
        }
    }
    let ligString;
    if (ligands && ligands.length > 0) {
        ligString = ":_ and (" + ligands.join(" or ") + ")";
        fadeOutRepresentationNGL(strucId, strucId + ".ligands", opacity_fadeout_ligand);
    } else if (extra_residues.length > 0 || extra_ranges.length > 0) {
        ligString = "none";
        fadeOutRepresentationNGL(strucId, strucId + ".ligands", opacity_fadeout_ligand);
    } else {
        ligString = "none";
        fadeOutRepresentationNGL(strucId, strucId + ".ligands", 1);
    }
    let ligand_repr = getNGLRepresentationByName(strucId + ".ligandsHighlight");
    if (ligand_repr) {
        ligand_repr.setSelection(ligString);
    }
}

function fadeOutRepresentationNGL(strucId, reprName, opacity) {
    repr = getNGLRepresentationByName(reprName);
    if (repr) repr.setParameters({
        opacity: opacity
    });
}

function connectResiduesNGL(residuePairs) {
    let removeThese = [];
    ngl_stage.compList.forEach(function(c) {
        if (c.name.indexOf("connect") > -1) removeThese.push(c);
    });
    for (let i = 0; i < removeThese.length; i++) ngl_stage.removeComponent(removeThese[i]);
    if (residuePairs) for (let i = 0; i < residuePairs.length; i++) {
        let comp = getNGLComponentByName(residuePairs[i][0].struc_id);
        let shape = new NGL.Shape("connect." + i + ".line", {
            disableImpostor: true
        });
        if (!comp) continue;
        let coo1 = comp.structure.getView(new NGL.Selection(residuePairs[i][0].res_num + ":" + residuePairs[i][0].chain + ".CA")).getAtomData().position;
        let coo2 = comp.structure.getView(new NGL.Selection(residuePairs[i][1].res_num + ":" + residuePairs[i][1].chain + ".CA")).getAtomData().position;
        if (coo1.length > 0 && coo2.length > 0) {
            score = Math.min(1, residuePairs[i][2] * 2);
            shape.addCylinder(coo1, coo2, [ .9, 1 - score, 1 - score ], .2);
            let shapeComp = ngl_stage.addComponentFromObject(shape);
            shapeComp.addRepresentation("buffer");
        }
    }
}

function changeRepNGL(newrep, show) {
    if (newrep === "fog") {
        if (show) ngl_stage.setParameters({
            fogNear: 40,
            fogFar: 60
        }); else ngl_stage.setParameters({
            forNear: 50,
            fogFar: 100
        });
        jQuery.cookie("fog", show, {
            expires: 365,
            path: "/"
        });
        return;
    }
    $("#repBtnLbl").text(newrep.charAt(0).toUpperCase() + newrep.substring(1));
    jQuery.cookie("defaultRepr", newrep, {
        expires: 365,
        path: "/"
    });
    switch (newrep) {
      case "lines":
        newrep = "line";
        break;

      case "ballAndStick":
        newrep = "ball+stick";
        break;

      case "outline":
        return;
    }
    let strucId = null;
    let oldrep = null;
    ngl_stage.eachRepresentation(function(rep) {
        if (!rep.name.match(/(ligands|dna|over|buffer|membrane|select)/)) {
            strucId = rep.name.split(".")[0];
            for (let i in ngl_stage.compList) {
                comp = ngl_stage.compList[i];
                if (comp.name == strucId) {
                    if (oldrep === null) {
                        oldrep = rep.repr.type;
                    }
                    if (oldrep == "surface" && newrep != "surface") {
                        console.debug("Adding residuesHighlight representation");
                        comp.addRepresentation(newrep, {
                            name: strucId + ".residuesHighlight",
                            sele: "none",
                            smoothSheet: true,
                            roughness: 1,
                            surfaceType: "sas",
                            probeRadius: 1.4,
                            side: "front",
                            opacity: rep.getParameters()["opacity"],
                            scaleFactor: Math.min(1.5, Math.max(.1, 5e4 / comp.structure.atomCount))
                        });
                        try {
                            refreshBondsAndLigandHighlights();
                        } catch (err) {}
                    }
                    comp.removeRepresentation(rep);
                    if (newrep == "surface" && rep.name == strucId + ".residuesHighlight") {
                        console.debug("Removing surface residuesHighlight");
                    } else {
                        comp.addRepresentation(newrep, {
                            name: rep.name,
                            sele: rep.getParameters()["sele"],
                            smoothSheet: true,
                            surfaceType: "sas",
                            probeRadius: 1.4,
                            scaleFactor: Math.min(1.5, Math.max(.1, 5e4 / comp.structure.atomCount)),
                            roughness: 1,
                            side: "front",
                            opacity: rep.getParameters()["opacity"]
                        });
                    }
                }
            }
        }
    });
}

function setupNGLRepresentations(comp, strucId, plipEnabled, dnaRepresentations, orientationMatrix, nonPdbized) {
    if (orientationMatrix) {
        ngl_stage.viewerControls.orient(orientationMatrix);
    }
    if (plipEnabled === undefined) plipEnabled = false;
    let repr = $.cookie("defaultRepr");
    if (!repr) repr = "cartoon"; else if (repr === "ballAndStick") repr = "ball+stick"; else if (repr === "lines") repr = "line";
    comp.setName(strucId);
    comp.addRepresentation(repr, {
        name: strucId,
        sele: "not :_ and not nucleic and not :-" + (nonPdbized ? " and not ligand" : ""),
        smoothSheet: true,
        roughness: 1,
        surfaceType: "sas",
        probeRadius: 1.4,
        scaleFactor: Math.min(1.5, Math.max(.1, 5e4 / comp.structure.atomCount)),
        side: "front"
    });
    comp.addRepresentation("ball+stick", {
        name: strucId + ".ligands",
        colorScheme: "element",
        sele: ":_" + (nonPdbized ? " or ligand" : ""),
        side: "front"
    });
    comp.addRepresentation("ball+stick", {
        name: strucId + ".residuesHighlight",
        colorScheme: "element",
        sele: "none",
        aspectRatio: 1
    });
    comp.addRepresentation("ball+stick", {
        name: strucId + ".hover",
        sele: "none"
    });
    comp.addRepresentation("ball+stick", {
        name: strucId + ".select",
        colorScheme: "element",
        sele: "none",
        aspectRatio: 1
    });
    if (plipEnabled) {
        comp.addRepresentation("ball+stick", {
            name: strucId + ".ligandsHighlight",
            colorScheme: "element",
            sele: "none"
        });
        if (repr != "surface") {
            comp.addRepresentation(repr, {
                name: strucId + ".highlight",
                sele: "none",
                smoothSheet: true,
                roughness: 1,
                surfaceType: "sas",
                probeRadius: 1.4,
                scaleFactor: Math.min(1.5, Math.max(.1, 5e4 / comp.structure.atomCount)),
                side: "front"
            });
        }
    }
    if (dnaRepresentations) dnaRepresentations.forEach(function(repr) {
        if (repr == "line") scheme = "element"; else if (repr == "base") scheme = "resname"; else scheme = "moleculetype";
        comp.addRepresentation(repr, {
            name: strucId + ".dna",
            colorScheme: scheme,
            sele: "nucleic and not :_"
        });
    });
    if (orientationMatrix) ngl_stage.autoView(); else if (typeof on_viewer_ready_function == "function") {
        on_viewer_ready_function(strucId, comp);
    } else comp.autoView(1e3);
    updatecols();
    if (plipEnabled) refreshBondsAndLigandHighlights();
}

function resAnnoLabel(baseStrucId, chn, resno) {
    let lbl = "";
    [ baseStrucId, baseStrucId + ".dna" ].forEach(strucId => {
        if (chainMapping.structures[strucId] && chainMapping.structures[strucId][chn] && chainMapping.structures[strucId][chn][resno]) {
            if (chainMapping.chainMeta[strucId] && chainMapping.chainMeta[strucId][chn]) lbl += " (" + chainMapping.chainMeta[strucId][chn].display + ")";
            for (let i = 0; i < chainMapping.annotations.length; i++) if (chainMapping.annotations[i].active && chainMapping.structures[strucId][chn][resno][chainMapping.annotations[i].key] != null) {
                let val = chainMapping.structures[strucId][chn][resno][chainMapping.annotations[i].key];
                lbl += "<br><small>" + chainMapping.annotations[i].label + ": " + (typeof val === "number" ? val.toFixed(chainMapping.annotations[i].key == "ramachandran" ? 4 : 2) : val);
                if (chainMapping.annotations[i].key == "qm" && $.cookie("colourScheme") == "confClass") {
                    if (val < 1) val *= 100;
                    if (val > 90) lbl += " (Very high)"; else if (val > 70) lbl += " (Confident)"; else if (val > 50) lbl += " (Low)"; else lbl += " (Very low)";
                }
                lbl += "</small>";
            }
        }
    });
    return lbl;
}

function fetchBFactorsNGL(strucId, structure, storeAs, checkValidResidues) {
    let total = 0;
    let totalCount = 0;
    if (!chainMapping.structures[strucId]) chainMapping.structures[strucId] = {};
    structure.eachResidue(function(res) {
        let chn = res.chainname;
        if (chn.match(/(_|-)/)) return;
        let tot = 0;
        res.eachAtom(function(a) {
            tot += a.bfactor;
        });
        total += tot / res.atomCount;
        totalCount++;
        coloursAssigned = true;
        if (!chainMapping.structures[strucId][chn]) chainMapping.structures[strucId][chn] = [];
        if (!chainMapping.structures[strucId][chn][res.resno]) chainMapping.structures[strucId][chn][res.resno] = {};
        if (checkValidResidues && !chainMapping.structures[strucId].invalidMapping && chainMapping.structures[strucId][chn][res.resno].resname) if (residueProperties[chainMapping.structures[strucId][chn][res.resno].resname].triplet != res.resname) chainMapping.structures[strucId].invalidMapping = true;
        chainMapping.structures[strucId][chn][res.resno][storeAs] = tot / res.atomCount;
    });
    if (storeAs == "bfactor") setBFactorRange(strucId);
    return total > 0 ? total / totalCount : 0;
}

function fetchSecStrucNGL(strucId, structure) {
    let chn;
    structure.eachResidue(function(res) {
        chn = res.chainname;
        if (chn.match(/(_|-)/)) return;
        if (!chainMapping.structures[strucId]) chainMapping.structures[strucId] = {};
        if (!chainMapping.structures[strucId][chn]) chainMapping.structures[strucId][chn] = [];
        if (!chainMapping.structures[strucId][chn][res.resno]) chainMapping.structures[strucId][chn][res.resno] = {};
        chainMapping.structures[strucId][chn][res.resno].ss = res.sstruc.toUpperCase();
        if (chainMapping.structures[strucId][chn][res.resno].ss == "") chainMapping.structures[strucId][chn][res.resno].ss = "C";
    });
    return true;
}

function setColorForAtom(go, atom, color) {
    let view = go.structure().createEmptyView();
    view.addAtom(atom);
    go.colorBy(pv.color.uniform(color), view);
}

let prevPicked = null;

function doPVMouse(event) {
    if (pv_viewer._mouseHandler._lastMouseDownTime && (pv_viewer._mouseHandler._lastMouseDownTime == null || pv_viewer._mouseHandler._lastMouseUpTime < pv_viewer._mouseHandler._lastMouseDownTime)) return;
    let fastClick = event.type === "click" && pv_viewer._mouseHandler._lastMouseUpTime - pv_viewer._mouseHandler._lastMouseDownTime < 300;
    let rect = pv_viewer.boundingClientRect();
    let picked = pv_viewer.pick({
        x: event.clientX - rect.left,
        y: event.clientY - rect.top
    });
    if (prevPicked !== null && picked !== null && picked.target() === prevPicked.atom) {
        if (event.type === "click") {
            let atom = prevPicked.atom;
            let str = atom.qualifiedName().split(".");
            if (str[0].charAt(0) !== "_" && atom.residue()._isAminoacid) document.body.dispatchEvent(new CustomEvent("selectResidues", {
                detail: {
                    src: "structure",
                    struc_id: picked.object().geom.name(),
                    residues: [ {
                        chain: str[0],
                        resno: atom.residue().num()
                    } ],
                    ctrl: event.ctrlKey || event.metaKey
                }
            }));
            pv_viewer.fitTo(atom.residue());
        }
        return;
    }
    if (prevPicked !== null) setColorForAtom(prevPicked.node, prevPicked.atom, prevPicked.color);
    if (picked !== null) {
        let atom = picked.target();
        let str = atom.qualifiedName().split(".");
        if (str[0].charAt(0) === "_" || !atom.residue().isAminoacid() && !atom.residue().isNucleotide()) {
            document.getElementById("pv_status").innerHTML = atom.residue().name() + " " + atom.residue().num();
        } else {
            let color = [ 0, 0, 0, 0 ];
            picked.node().getColorForAtom(atom, color);
            prevPicked = {
                atom: atom,
                color: color,
                node: picked.node()
            };
            setColorForAtom(picked.node(), atom, "red");
            let strucId = picked.object().geom.name();
            let res = atom.residue().name();
            let resno = atom.residue().num();
            let lbl = res + " " + resno + " " + str[0] + resAnnoLabel(strucId, str[0], resno);
            document.getElementById("pv_status").innerHTML = lbl;
            event.target.dispatchEvent(new CustomEvent("highlightResidue", {
                detail: {
                    src: "structure",
                    struc_id: picked.object().geom.name(),
                    chain: str[0],
                    resno: atom.residue().num()
                },
                bubbles: true
            }));
        }
    } else {
        document.getElementById("pv_status").innerHTML = "";
        prevPicked = null;
        if (fastClick) pv_viewer.autoZoom();
    }
    pv_viewer.requestRedraw();
}

function highlightResiduesPV(chain, res_pos, strucId, multiple) {
    if (!multiple) pv_viewer.rm("*_res");
    if (chainMapping.structures && chainMapping.structures[strucId] && !chainMapping.structures[strucId][chain] && chainMapping.structures[strucId + ".dna"] && chainMapping.structures[strucId][chain + ".dna"]) strucId += ".dna";
    let pv_structure = pv_viewer.get(strucId);
    if (pv_structure == null) return;
    pv_structure = pv_structure.structure();
    if (!res_pos) {
        pv_viewer.fitTo(pv_structure.select({
            cnames: chain.split("")
        }));
        pv_viewer.requestRedraw();
        return;
    }
    let highlightRes = null;
    if (!res_pos.length) {
        highlightRes = pv_structure.select({
            cnames: chain.split("")
        }).residueSelect(function(res) {
            return res.num() == res_pos;
        });
    } else {
        highlightRes = pv_structure.select({
            cnames: chain.split("")
        }).residueSelect(function(res) {
            for (let i = 0; i < res_pos.length; i++) if (res.num() >= res_pos[i][0] && res.num() <= res_pos[i][1]) return true;
            return false;
        });
    }
    pv_viewer.ballsAndSticks((strucId ? strucId + "_" : "") + "res", highlightRes);
    return highlightRes;
}

function selectResiduesPV(e) {
    pv_viewer.rm("*select");
    pv_viewer.rm("*res");
    if (e.detail.residues == null || e.detail.residues.length === 0) {
        pv_viewer.requestRedraw();
        return;
    }
    let strucId = e.detail.struc_id;
    let pv_structure = pv_viewer.get(strucId);
    if (pv_structure == null) return;
    pv_structure = pv_structure.structure();
    let selectRes = null;
    selectRes = pv_structure.select({
        cnames: Array.from(new Set(e.detail.residues.map(r => r.chain)))
    }).residueSelect(res => {
        for (let detRes of e.detail.residues) {
            if (res.chain().name() === detRes.chain && res.num() >= detRes.resno && res.num() <= detRes.resno) return true;
        }
        return false;
    });
    pv_viewer.ballsAndSticks(strucId + ".select", selectRes);
    pv_viewer.fitTo(selectRes);
    pv_viewer.requestRedraw();
    return selectRes;
}

function centerResRangePV(chain, start, end, strucId, isDisulfid, secondChain) {
    if (chainMapping.structures && chainMapping.structures[strucId] && !chainMapping.structures[strucId][chain] && chainMapping.structures[strucId + ".dna"] && chainMapping.structures[strucId][chain + ".dna"]) strucId += ".dna";
    let residues = null;
    while (true) {
        let pv_structure = pv_viewer.get(strucId);
        if (pv_structure == null) return;
        if (chain.length === 1) residues = pv_structure.structure().residueSelect(res => {
            if (secondChain) {
                return res.num() === start && res.chain().name() === chain || res.num() === end && res.chain().name() === secondChain;
            } else {
                if (isDisulfid) return (res.num() === start || res.num() === end) && res.chain().name() === chain; else return res.num() >= start && res.num() <= end && res.chain().name() === chain;
            }
        }); else residues = pv_structure.structure().residueSelect(function(res) {
            if (isDisulfid) return (res.num() == start || res.num() == end) && chain.indexOf(res.chain().name()) > -1; else return res.num() >= start && res.num() <= end && chain.indexOf(res.chain().name()) > -1;
        });
        if (residues.chains().length > 0 || strucId.endsWith(".ligands")) break;
        strucId = strucId.replace(".dna", "") + ".ligands";
    }
    pv_viewer.fitTo(residues);
    pv_viewer.requestRedraw();
}

function centerOnLigandPV(ligands, strucId) {
    var lig_structure = pv_viewer.get(strucId + ".ligands");
    if (lig_structure == null) return;
    lig_structure = lig_structure.structure();
    var residues = lig_structure.residueSelect(function(r) {
        for (var i = 0; i < ligands.length; i++) {
            if (r.num() == ligands[i]) return true;
        }
    });
    pv_viewer.fitTo(residues);
    pv_viewer.requestRedraw();
}

function highlightLigandPV(allHighlights, contacts, strucId) {
    pv_viewer.rm("*ligandContactsHL");
    var lig_structure = pv_viewer.get(strucId + ".ligands");
    if (lig_structure == null) return;
    if (allHighlights) {
        var contacts = pv_viewer.get(strucId).structure().residueSelect(function(res) {
            var chn = res.chain().name();
            var num = res.num();
            var ah = allHighlights[chn];
            if (ah) {
                for (var i = 0; i < ah.length; i++) {
                    if (num == ah[i]) return true;
                }
            }
        });
        pv_viewer.lines(strucId + "ligandContactsHL", contacts);
        pv_viewer.requestRedraw();
    }
}

function connectResiduesPV(residuePairs) {
    pv_viewer.rm("connect.*");
    if (residuePairs) for (let i = 0; i < residuePairs.length; i++) {
        let g = pv_viewer.customMesh("connect." + i + ".line");
        let struc = pv_viewer.get(residuePairs[i][0].struc_id);
        if (!struc) continue;
        struc = struc.structure();
        let ca1 = struc.atom(residuePairs[i][0].chain + "." + residuePairs[i][0].res_num + ".CA");
        let ca2 = struc.atom(residuePairs[i][1].chain + "." + residuePairs[i][1].res_num + ".CA");
        if (ca1 && ca2) {
            let midPoint = pv.vec3.clone(ca1.pos());
            pv.vec3.add(midPoint, midPoint, ca2.pos());
            pv.vec3.scale(midPoint, midPoint, .5);
            score = Math.min(1, residuePairs[i][2] * 2);
            g.addTube(ca1.pos(), ca2.pos(), .1, {
                color: [ .9, 1 - score, 1 - score ]
            });
        }
    }
    pv_viewer.requestRedraw();
}

function changeRepPV(rep, show) {
    if (!rep.match(/(cartoon|tube|trace|lines|outline|fog)/)) rep = "cartoon";
    if (rep == "outline") {
        jQuery.cookie("pv_outline", !pv_viewer.options("outline"), {
            expires: 365,
            path: "/"
        });
        pv_viewer.options("outline", !pv_viewer.options("outline"));
    } else if (rep === "fog") {
        let fogOn = show !== undefined ? show : !pv_viewer.options("fog");
        jQuery.cookie("fog", fogOn, {
            expires: 365,
            path: "/"
        });
        pv_viewer.options("fog", fogOn);
    } else if (rep.match(/(cartoon|tube|trace|lines|ballAndStick)/)) {
        $("#repBtnLbl").text(rep.charAt(0).toUpperCase() + rep.substring(1));
        jQuery.cookie("defaultRepr", rep, {
            expires: 365,
            path: "/"
        });
        let viewsRemoved = [];
        let strucRemoved = [];
        pv_viewer.forEach(function(v) {
            let name = v.name();
            if (!name.match(/(ligand|dna|res|LigandBond|membrane|label|select)/)) {
                viewsRemoved.push(name);
                strucRemoved.push(v.structure());
            }
        });
        for (let i = 0; i < viewsRemoved.length; i++) pv_viewer.rm(viewsRemoved[i]);
        for (let i = 0; i < strucRemoved.length; i++) {
            let name = viewsRemoved[i];
            let pv_structure = strucRemoved[i];
            switch (rep) {
              case "cartoon":
                pv_viewer.cartoon(name, pv_structure, {
                    color: pv.color.uniform("white")
                });
                break;

              case "tube":
                pv_viewer.tube(name, pv_structure, {
                    color: pv.color.uniform("white")
                });
                break;

              case "trace":
                pv_viewer.trace(name, pv_structure, {
                    color: pv.color.uniform("white")
                });
                break;

              case "lines":
                pv_viewer.lines(name, pv_structure, {
                    color: pv.color.uniform("white")
                });
                break;

              case "ballAndstick":
                pv_viewer.ballsAndSticks(name, pv_structure, {
                    color: pv.color.uniform("white")
                });
            }
        }
    }
}

function normalize(a) {
    let out = [];
    let x = a[0], y = a[1], z = a[2];
    let len = x * x + y * y + z * z;
    if (len > 0) {
        len = 1 / Math.sqrt(len);
        out[0] = a[0] * len;
        out[1] = a[1] * len;
        out[2] = a[2] * len;
    }
    return out;
}

function cross(a, b) {
    let out = [];
    let ax = a[0], ay = a[1], az = a[2], bx = b[0], by = b[1], bz = b[2];
    out[0] = ay * bz - az * by;
    out[1] = az * bx - ax * bz;
    out[2] = ax * by - ay * bx;
    return out;
}

function showMembrane(strucid, show) {
    let m = membraneData[strucid];
    if (!m) return;
    let viewer = null;
    if ($("#pv:visible", currentWin.document).length > 0) {
        if (!pv_viewer.get(strucid)) return;
        viewer = "pv";
        if (!show) {
            pv_viewer.hide(strucid + ".membrane");
            pv_viewer.requestRedraw();
            return;
        }
        let mem = pv_viewer.get(strucid + ".membrane");
        if (mem) {
            pv_viewer.show(strucid + ".membrane");
            pv_viewer.requestRedraw();
            return;
        }
    }
    if ($("#ngl:visible", currentWin.document).length > 0) {
        viewer = "ngl";
        if (!getNGLComponentByName(strucid)) return;
        let comp = getNGLComponentByName(strucid + ".membrane");
        if (comp) {
            comp.setVisibility(show);
            return;
        }
        if (!show) return;
    }
    const distance = (a, b) => {
        let x = b[0] - a[0], y = b[1] - a[1], z = b[2] - a[2];
        return Math.sqrt(x * x + y * y + z * z);
    };
    const dot = (a, b) => a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
    let arbitrary_vec = [ 0, 1, 0 ];
    if (Math.abs(1 - Math.abs(dot(m.axis, arbitrary_vec))) <= 1e-5) {
        arbitrary_vec = [ 0, 0, 1 ];
    }
    plane_x = normalize(cross(m.axis, arbitrary_vec));
    plane_y = normalize(cross(m.axis, plane_x));
    let closeToPlane1 = [];
    let closeToPlane2 = [];
    let allResidues = [];
    if (viewer == "pv") {
        pv_viewer.get(strucid).structure().eachAtom(function(atom) {
            if (atom.name() == "CA") allResidues.push(atom.pos());
        });
    } else if (viewer == "ngl") {
        let comp = getNGLComponentByName(strucid);
        comp.structure.eachResidue(function(r) {
            r.eachAtom(function(a) {
                if (a.atomname == "CA") allResidues.push([ a.x, a.y, a.z ]);
            });
        });
    }
    let points = [];
    let coords = "";
    let atomIdx = 1;
    let delta = 1.5;
    let paddedRadius = m.radius / 2 + delta * 16;
    for (let x = -paddedRadius; x < paddedRadius; x += delta) {
        for (let y = -paddedRadius; y < paddedRadius; y += delta) {
            pos = [ m.plane_one_center[0] + x * delta * plane_x[0] + y * delta * plane_y[0], m.plane_one_center[1] + x * delta * plane_x[1] + y * delta * plane_y[1], m.plane_one_center[2] + x * delta * plane_x[2] + y * delta * plane_y[2] ];
            if (distance(pos, m.plane_one_center) > m.radius + 8) continue;
            pos2 = [ m.plane_two_center[0] + x * delta * plane_x[0] + y * delta * plane_y[0], m.plane_two_center[1] + x * delta * plane_x[1] + y * delta * plane_y[1], m.plane_two_center[2] + x * delta * plane_x[2] + y * delta * plane_y[2] ];
            let i = 0;
            while (i < allResidues.length) {
                if (distance(pos, allResidues[i]) < 5) closeToPlane1.push(allResidues.splice(i, 1)[0]); else if (distance(pos2, allResidues[i]) < 5) closeToPlane2.push(allResidues.splice(i, 1)[0]); else i++;
            }
            points.push([ pos, pos2 ]);
        }
    }
    for (let p = 0; p < points.length; p++) {
        let closeToPlane = closeToPlane1;
        let pointIdx = 0;
        do {
            let inplane = false;
            let point = points[p][pointIdx++];
            for (let i = 0; i < closeToPlane.length; i++) {
                if (distance(point, closeToPlane[i]) < 5) {
                    inplane = true;
                    break;
                }
            }
            if (!inplane) coords += "ATOM  " + ("     " + atomIdx++).slice(-5) + "   -  MEM M   1    " + ("        " + point[0].toFixed(3)).slice(-8) + ("        " + point[1].toFixed(3)).slice(-8) + ("        " + point[2].toFixed(3)).slice(-8) + "  1.00  0.00    \n";
            closeToPlane = closeToPlane2;
        } while (pointIdx < 2);
    }
    coords += "END";
    if (viewer == "pv") {
        pv_viewer.points(strucid + ".membrane", pv.io.pdb(coords), {
            color: pv.color.uniform("#8b8b8b"),
            pointSize: 2
        });
        if (m.viewTransform) pv_viewer.setRotation(m.viewTransform);
    } else if (viewer == "ngl") {
        for (schemeId in NGL.ColormakerRegistry.userSchemes) if (schemeId.indexOf(strucid + ".membrane") > -1) NGL.ColormakerRegistry.removeScheme(schemeId);
        if (membraneData[strucid].viewTransform) ngl_stage.viewerControls.orient(membraneData[strucid].viewTransform);
        schemeId = NGL.ColormakerRegistry.addScheme(function(params) {
            this.atomColor = function() {
                return 150 << 16 | 150 << 8 | 150;
            };
        }, strucid + ".membrane");
        let stringBlob = new Blob([ coords ], {
            type: "text/plain"
        });
        ngl_stage.loadFile(stringBlob, {
            ext: "pdb"
        }).then(function(struc) {
            struc.setName(strucid + ".membrane");
            struc.addRepresentation("point", {
                name: strucid + ".membrane",
                pointSize: 2,
                useTexture: true,
                forceTransparent: true,
                edgeBleach: .6,
                color: schemeId
            });
            if (membraneData[strucid].viewTransform) ngl_stage.viewerControls.spin([ 0, 1, 0 ], Math.PI);
            ngl_stage.autoView();
        });
    }
}

function fetchBFactorsPV(strucId, pv_structure, storeAs, checkValidResidues) {
    let total = 0;
    let totalCount = 0;
    pv_structure.eachChain(function(c) {
        chn = c.name();
        if (chn.match(/(_|-)/)) return;
        if (!chainMapping.structures[strucId]) chainMapping.structures[strucId] = {};
        if (!chainMapping.structures[strucId][chn]) chainMapping.structures[strucId][chn] = [];
        c.eachResidue(function(res) {
            if (!res.isAminoacid() && !res.isNucleotide()) return;
            let tot = 0;
            let count = 0;
            res.eachAtom(function(a) {
                tot += a.tempFactor();
                count++;
            });
            total += tot / count;
            totalCount++;
            if (!chainMapping.structures[strucId][chn][res.num()]) chainMapping.structures[strucId][chn][res.num()] = {};
            if (checkValidResidues && !chainMapping.structures[strucId].invalidMapping && chainMapping.structures[strucId][chn][res.num()].resname) if (residueProperties[chainMapping.structures[strucId][chn][res.num()].resname].triplet != res.name()) chainMapping.structures[strucId].invalidMapping = true;
            chainMapping.structures[strucId][chn][res.num()][storeAs] = tot / count;
        });
    });
    if (storeAs == "bfactor") setBFactorRange(strucId);
    return total > 0 ? total / totalCount : 0;
}

function fetchSecStrucPV(strucId, pv_structure) {
    pv_structure.eachResidue(function(res) {
        chn = res.chain().name();
        if (chn.match(/(_|-)/)) return;
        if (!chainMapping.structures[strucId]) chainMapping.structures[strucId] = {};
        if (!chainMapping.structures[strucId][chn]) chainMapping.structures[strucId][chn] = [];
        if (!chainMapping.structures[strucId][chn][res.num()]) chainMapping.structures[strucId][chn][res.num()] = {};
        chainMapping.structures[strucId][chn][res.num()].ss = res.ss();
    });
    return true;
}

function midpoint1d(coo1, coo2, percent) {
    return coo1 + (coo2 - coo1) * percent;
}

function midpoint3d(coo1, coo2, percent) {
    return [ midpoint1d(coo1[0], coo2[0], percent), midpoint1d(coo1[1], coo2[1], percent), midpoint1d(coo1[2], coo2[2], percent) ];
}

function showWaterPV() {
    pv_viewer.ballsAndSticks("water", pv_viewer.get("template").structure().select({
        rnames: [ "HOH" ]
    }));
}

function showWaterNGL() {
    log.console("showWaterNGL not implemented");
}

function showWater() {
    if ($("#pv:visible", currentWin.document).length > 0) showWaterPV();
    if ($("#ngl:visible", currentWin.document).length > 0) showWaterNGL();
}

function toggleBond(strucId, bondId) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_bonds.getBondById(strucId, bondId).toggle();
    deHighlightBond(strucId, bondId, true);
}

function highlightBond(strucId, bondId) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_bonds.getBondById(strucId, bondId).highlight();
}

function deHighlightBond(strucId, bondId, decrementResidueCounter) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    if (decrementResidueCounter == null) {
        decrementResidueCounter = true;
    }
    plip_bonds.getBondById(strucId, bondId).dehighlight(true, decrementResidueCounter);
}

function highlightBonds(strucId, bondIds) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_bonds.getBondsById(strucId, bondIds).forEach(function(bond) {
        bond.highlight(false);
    });
    plip_residues.redraw(strucId);
}

function deHighlightBonds(strucId, bondIds) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_bonds.getBondsById(strucId, bondIds).forEach(function(bond) {
        bond.dehighlight(false);
    });
    plip_residues.redraw(strucId);
}

function clickBonds(strucId, bondIds, bondsId) {
    toggleBonds(strucId, bondIds, bondsId);
    var checkbox = $("#plip_bonds_" + bondsId);
    if (checkbox.prop("checked")) {
        fitToBonds(strucId, bondIds, bondsId);
    }
}

function toggleBonds(strucId, bondIds, bondsId) {
    var checkbox = $("#plip_bonds_" + bondsId);
    var isChecked = checkbox.prop("checked");
    checkbox.prop("checked", !isChecked);
    checkbox.trigger("change");
}

function changeBondsCheckbox(strucId, bondIds, bondsId) {
    var checkbox = $("#plip_bonds_" + bondsId);
    var isChecked = checkbox.prop("checked");
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        checkbox.prop("checked", false);
        return;
    }
    var bonds = plip_bonds.getBondsById(strucId, bondIds);
    bonds.forEach(function(bond) {
        isChecked ? bond.show(false) : bond.hide(false);
        bond.dehighlight(false, true);
    });
    plip_residues.redraw(strucId);
    if (isChecked && plip_interactions_fit_bonds) {
        fitToBonds(strucId, bondIds, bondsId);
    }
}

function showBonds(strucId, bondIds) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    var bonds = plip_bonds.getBondsById(strucId, bondIds);
    bonds.forEach(function(bond) {
        bond.show(false);
        bond.dehighlight(false, true);
    });
    plip_residues.redraw(strucId);
}

function hideBonds(strucId, bondIds) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    var bonds = plip_bonds.getBondsById(strucId, bondIds);
    bonds.forEach(function(bond) {
        bond.hide(false);
        bond.dehighlight(false, true);
    });
    plip_residues.redraw(strucId);
}

function fitToBond(strucId, bondId) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    var bond = plip_bonds.getBondById(strucId, bondId);
    bond.fitView();
}

function fitToResidue(strucId, resId, ligandId) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    residue = plip_residues.getResidue(strucId, resId);
    if (ligandId) {
        ligand = all_ligands.getLigandById(strucId, ligandId);
        PLIP.AbstractViewer.fitView(strucId, [ ligand ], [ residue ]);
    } else {
        PLIP.AbstractViewer.fitView(strucId, null, [ residue ]);
    }
}

function fitToBonds(strucId, bondIds) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_bonds.fitViewBonds(strucId, bondIds);
}

function showResidues(strucId, resIds, ligand_indices) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_residues.getResiduesByResId(strucId, resIds).forEach(function(residue) {
        residue.show(false);
    });
    plip_residues.redraw(strucId);
}

function hideResidues(strucId, resIds, ligand_indices) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    plip_residues.getResiduesByResId(strucId, resIds).forEach(function(residue) {
        residue.hide(false);
    });
    plip_residues.redraw(strucId);
}

function showLigands(strucId, ligand_indices) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    all_ligands.getLigandsById(strucId, ligand_indices).forEach(function(ligand) {
        ligand.show(false, true);
    });
    all_ligands.redraw(strucId);
}

function hideLigands(strucId, ligand_indices) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    all_ligands.getLigandsById(strucId, ligand_indices).forEach(function(ligand) {
        ligand.hide(false, null, true);
    });
    all_ligands.redraw(strucId);
}

PLIP.Bonds = function() {
    var _list = {};
    this.hasStruc = function(strucId) {
        return strucId in _list;
    };
    this.removeStruc = function(strucId) {
        if (strucId in _list) {
            delete _list[strucId];
        }
    };
    this.addStruc = function(strucId) {
        if (!this.hasStruc(strucId)) {
            _list[strucId] = [];
        }
    };
    this.listStruc = function() {
        return Object.keys(_list);
    };
    this.listVisible = function(strucId) {
        if (!this.hasStruc(strucId)) {
            console.error("listVisible: no strucId " + strucId);
            return undefined;
        }
        return _list[strucId].filter(function(bond) {
            return bond.isVisible();
        });
    };
    this.getBondById = function(strucId, id) {
        if (!(strucId in _list)) {
            console.error("getBondById: no strucId " + strucId);
            console.trace();
            return undefined;
        }
        return _list[strucId].find(function(bond) {
            return bond.id === id;
        });
    };
    this.getBondsById = function(strucId, ids) {
        if (!(strucId in _list)) {
            console.error("getBondsById: no strucId " + strucId);
            return undefined;
        }
        return _list[strucId].filter(function(bond) {
            return ids.includes(bond.id);
        });
    };
    this.add = function(bond) {
        this.addStruc(bond.strucId);
        _list[bond.strucId].push(bond);
    };
    var _showBonds = function(bonds, strucId) {
        for (var i = 0; i < bonds.length; i++) {
            if (!bonds[i].isVisible()) {
                bonds[i].show(false);
            }
        }
        plip_residues.redraw(strucId);
    };
    var _hideBonds = function(bonds, strucId) {
        for (var i = 0; i < bonds.length; i++) {
            if (bonds[i].isVisible()) {
                bonds[i].hide(false);
            }
        }
        plip_residues.redraw(strucId);
    };
    this.showAll = function(strucId) {
        var allBonds = _list[strucId];
        _showBonds(allBonds, strucId);
    };
    this.hideAll = function(strucId) {
        var allBonds = _list[strucId];
        _hideBonds(allBonds, strucId);
    };
    this.fitViewBonds = function(strucId, bondIds) {
        var bonds = this.getBondsById(strucId, bondIds);
        var ligands = bonds.map(function(bond) {
            return bond.ligand;
        }).filter(function(x, i, a) {
            return a.indexOf(x) == i;
        });
        var residues = bonds.map(function(bond) {
            return bond.residue;
        }).filter(function(x, i, a) {
            return a.indexOf(x) == i;
        });
        PLIP.AbstractViewer.fitView(strucId, ligands, residues);
    };
    this.refresh = function(strucId) {
        for (var i = 0; i < _list[strucId].length; i++) {
            _list[strucId][i].refresh();
        }
    };
    this.refreshAll = function() {
        for (var i in _list) {
            if (PLIP.AbstractViewer.isStructureEnabled(i)) {
                this.refresh(i);
            }
        }
    };
};

PLIP.Bond = function(id, strucId, ligand, coo1, coo2, coo3, color, residue, show) {
    var _visible = false;
    var _highlighted = false;
    var _bondMesh = undefined;
    var _bondMeshId = "LigandBond" + id;
    var _bondMeshHighlight = undefined;
    var _bondMeshHighlightId = "LigandBondHighlight" + id;
    this.show = function(redraw) {
        if (redraw == null) redraw = true;
        if (this.isVisible()) {
            console.info("Bond already visible");
            return;
        }
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        this.draw();
        if (!(this.residue === undefined)) {
            this.residue.show(false);
        }
        this.ligand.show(redraw);
        _visible = true;
        if (redraw) {
            pv_viewer.requestRedraw();
        }
    };
    this.hide = function(redraw) {
        if (redraw == null) redraw = true;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        if (!this.isVisible()) {
            return;
        }
        this.destroy();
        _visible = false;
        if (!(this.residue === undefined)) {
            this.residue.hide(false);
        }
        this.ligand.hide(redraw);
        if (redraw) {
            pv_viewer.requestRedraw();
        }
    };
    this.draw = function() {
        if (this.coo3 == null) {
            PLIP.AbstractViewer.addBondLine(this.getBondMesh(), this.coo1, this.coo2, plip_bond_num_dash, plip_bond_radius, this.color, true);
        } else {
            PLIP.AbstractViewer.addBondLine(this.getBondMesh(), this.coo1, this.coo3, plip_bond_num_dash, plip_bond_radius, this.color, false);
            PLIP.AbstractViewer.addBondLine(this.getBondMesh(), this.coo2, this.coo3, plip_bond_num_dash, plip_bond_radius, this.color, false);
            PLIP.AbstractViewer.addWaterCircle(this.getBondMesh(), this.coo3, plip_bond_radius * 2, this.color, true);
        }
    };
    this.destroy = function() {
        _bondMesh = PLIP.AbstractViewer.removeMesh(_bondMeshId);
    };
    this.refresh = function() {
        this.destroy();
        if (this.isVisible()) {
            this.draw();
        }
        this.destroyHighlight();
        if (this.isHighlighted()) {
            this.drawHighlight();
        }
    };
    this.toggle = function() {
        if (this.isVisible()) {
            this.hide();
        } else {
            this.show();
        }
    };
    this.isVisible = function() {
        return _visible;
    };
    this.getBondMesh = function() {
        if (_bondMesh == undefined) {
            _bondMesh = PLIP.AbstractViewer.newBondMesh(_bondMeshId);
        }
        return _bondMesh;
    };
    this.highlight = function(redraw) {
        if (redraw == null) redraw = true;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        if (this.isHighlighted()) {
            return;
        }
        this.drawHighlight();
        if (!(this.residue === undefined)) {
            this.residue.show(false);
        }
        this.ligand.show(redraw);
        _highlighted = true;
    };
    this.drawHighlight = function() {
        if (this.coo3 == null) {
            PLIP.AbstractViewer.addBondLine(this.getBondMeshHighlight(), this.coo1, this.coo2, plip_higlight_bond_num_dash, plip_bond_radius * plip_higlight_factor, this.color, true);
        } else {
            PLIP.AbstractViewer.addBondLine(this.getBondMeshHighlight(), this.coo1, this.coo3, plip_higlight_bond_num_dash, plip_bond_radius * plip_higlight_factor, this.color, false);
            PLIP.AbstractViewer.addBondLine(this.getBondMeshHighlight(), this.coo2, this.coo3, plip_higlight_bond_num_dash, plip_bond_radius * plip_higlight_factor, this.color, false);
            PLIP.AbstractViewer.addWaterCircle(this.getBondMeshHighlight(), this.coo3, plip_bond_radius * plip_higlight_factor * 2, this.color, true);
        }
    };
    this.destroyHighlight = function() {
        _bondMeshHighlight = PLIP.AbstractViewer.removeMesh(_bondMeshHighlightId);
    };
    this.dehighlight = function(redraw) {
        if (redraw == null) redraw = true;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        if (!this.isHighlighted()) {
            return;
        }
        if (!this.isHighlighted()) {
            return;
        }
        this.destroyHighlight();
        if (!(this.residue === undefined)) {
            this.residue.hide(false);
        }
        this.ligand.hide(redraw);
        _highlighted = false;
        if (redraw) {
            pv_viewer.requestRedraw();
        }
    };
    this.toggleHighlight = function() {
        if (this.isHighlighted()) {
            this.dehighlight();
        } else {
            this.highlight();
        }
    };
    this.isHighlighted = function() {
        return _highlighted;
    };
    this.getBondMeshHighlight = function() {
        if (_bondMeshHighlight == undefined) {
            _bondMeshHighlight = PLIP.AbstractViewer.newBondMesh(_bondMeshHighlightId);
        }
        return _bondMeshHighlight;
    };
    this.fitView = function() {
        PLIP.AbstractViewer.fitView(this.strucId, [ this.ligand ], [ this.residue ]);
    };
    this.getLigand = function() {
        return this.ligand;
    };
    this.residue = plip_residues.addResidue(strucId, residue[0], residue[1], residue[2]);
    this.residue.addBond(this);
    this.id = id;
    this.strucId = strucId;
    this.ligand = ligand;
    this.coo1 = coo1;
    this.coo2 = coo2;
    this.coo3 = coo3;
    this.color = color;
    if (show) {
        this.show();
    }
};

PLIP.makeResidueId = function(chain, one_letter, position) {
    return chain + ":" + one_letter + position;
};

PLIP.Residue = function(strucId, chain, one_letter, position) {
    var _visible = 0;
    var _bonds = [];
    this.show = function(redraw) {
        if (redraw == null) redraw = true;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        _visible++;
        if (_visible > 0 && redraw) {
            plip_residues.redraw(this.strucId);
        }
    };
    this.hide = function(redraw, decrementCounter) {
        if (redraw == null) redraw = true;
        if (decrementCounter == null) decrementCounter = true;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        if (decrementCounter) {
            _visible = Math.max(_visible - 1, 0);
        } else {
            alert("Did not decrement counter! " + this.id + ": " + _visible);
        }
        if (_visible == 0 && redraw) {
            plip_residues.redraw(this.strucId);
        }
    };
    this.isVisible = function() {
        return _visible > 0;
    };
    this.getOldResId = function() {
        return this.chain + this.position;
    };
    this.addBond = function(bond) {
        _bonds.push(bond);
    };
    this.getBonds = function() {
        return _bonds;
    };
    this.strucId = strucId;
    this.id = PLIP.makeResidueId(chain, one_letter, position);
    this.chain = chain;
    this.one_letter = one_letter;
    this.position = position;
};

PLIP.Residues = function() {
    var _list = {};
    this.hasStruc = function(strucId) {
        return strucId in _list;
    };
    this.removeStruc = function(strucId) {
        if (strucId in _list) {
            delete _list[strucId];
        }
    };
    this.addStruc = function(strucId) {
        if (!this.hasStruc(strucId)) {
            _list[strucId] = [];
        }
    };
    this.listStruc = function() {
        return Object.keys(_list);
    };
    this.getStruc = function(strucId) {
        if (strucId in _list) {
            return _list[strucId];
        } else if (strucId === undefined) {
            console.error("StrucId is undefined");
            console.trace();
        } else {
            console.error("No such strucId: '" + strucId + "'");
            console.trace();
        }
        return undefined;
    };
    this.addResidue = function(strucId, chain, one_letter, position) {
        this.addStruc(strucId);
        var id = PLIP.makeResidueId(chain, one_letter, position);
        var new_residue = this.getResidue(strucId, id);
        if (new_residue === undefined) {
            new_residue = new PLIP.Residue(strucId, chain, one_letter, position);
            this.getStruc(strucId).push(new_residue);
        }
        return new_residue;
    };
    this.getResidue = function(strucId, id) {
        if (strucId in _list) {
            return this.getStruc(strucId).find(function(res) {
                return res.id == id;
            });
        }
        return undefined;
    };
    this.getResidueByOldResId = function(strucId, resId) {
        if (strucId in _list) {
            return this.getStruc(strucId).find(function(res) {
                return res.getOldResId() == resId;
            });
        }
        return undefined;
    };
    this.getResiduesByResId = function(strucId, resIds) {
        if (strucId in _list) {
            return _list[strucId].filter(function(x) {
                return resIds.includes(x.id);
            });
        }
        return undefined;
    };
    this.getVisibleResidues = function(strucId) {
        if (this.hasStruc(strucId)) {
            return this.getStruc(strucId).filter(function(residue) {
                return residue.isVisible();
            });
        } else {
            return [];
        }
    };
    this.getVisibleResiduesResId = function(strucId) {
        return this.getVisibleResidues(strucId).map(function(residue) {
            return residue.getOldResId();
        });
    };
    this.redraw = function(strucId) {
        highlightLigand(all_ligands.getVisibleLigandIds(strucId), this.getVisibleResiduesResId(strucId), strucId);
        pv_viewer.requestRedraw();
    };
    this.redrawAll = function() {
        for (var i in _list) {
            if (PLIP.AbstractViewer.isStructureEnabled(i)) {
                this.redraw(i);
            }
        }
    };
};

PLIP.Ligands = function() {
    var _list = {};
    this.hasStruc = function(strucId) {
        return strucId in _list;
    };
    this.removeStruc = function(strucId) {
        if (this.hasStruc(strucId)) {
            delete _list[strucId];
        }
    };
    this.addStruc = function(strucId) {
        if (!this.hasStruc(strucId)) {
            _list[strucId] = [];
        }
    };
    this.listStruc = function() {
        return Object.keys(_list);
    };
    this.getStruc = function(strucId) {
        if (this.hasStruc(strucId)) {
            return _list[strucId];
        } else if (strucId === undefined) {
            console.error("StrucId " + strucId + " is undefined");
            console.trace();
        } else {
            console.error("No such strucId: '" + strucId + "'");
            console.trace();
        }
        return undefined;
    };
    this.getLigandById = function(strucId, id) {
        if (!this.hasStruc(strucId)) {
            console.error("getLigandById: no strucId " + strucId);
            console.trace();
            return undefined;
        }
        return _list[strucId].find(function(ligand) {
            return ligand.id === id;
        });
    };
    this.getLigandsById = function(strucId, ids) {
        if (!this.hasStruc(strucId)) {
            console.error("getLigandsById: no strucId " + strucId);
            return undefined;
        }
        return _list[strucId].filter(function(ligand) {
            return ids.includes(ligand.id);
        });
    };
    this.addLigand = function(strucId, ligand) {
        this.addStruc(strucId);
        var new_ligand = this.getLigandById(strucId, ligand);
        if (new_ligand === undefined) {
            new_ligand = new PLIP.Ligand(strucId, ligand);
            this.getStruc(strucId).push(new_ligand);
        }
        return new_ligand;
    };
    this.getVisibleLigands = function(strucId) {
        if (this.hasStruc(strucId)) {
            return this.getStruc(strucId).filter(function(ligand) {
                return ligand.isVisible();
            });
        } else {
            return [];
        }
    };
    this.getVisibleLigandIds = function(strucId) {
        return this.getVisibleLigands(strucId).map(function(ligand) {
            return ligand.getLigandId();
        });
    };
    this.redraw = function(strucId) {
        highlightLigand(this.getVisibleLigandIds(strucId), plip_residues.getVisibleResiduesResId(strucId), strucId);
        pv_viewer.requestRedraw();
    };
    this.redrawAll = function() {
        for (var i in _list) {
            if (PLIP.AbstractViewer.isStructureEnabled(i)) {
                this.redraw(i);
            }
        }
    };
};

PLIP.Ligand = function(strucId, position) {
    var _visible = 0;
    var _residues = [];
    var _bonds = [];
    this.show = function(redraw, showResidues) {
        if (redraw == null) redraw = true;
        if (showResidues == null) showResidues = false;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        _visible++;
        if (showResidues) {
            _residues.forEach(function(residue) {
                residue.show(false);
            });
        }
        if (_visible > 0 && redraw) {
            all_ligands.redraw(this.strucId);
        }
    };
    this.hide = function(redraw, decrementCounter, hideResidues) {
        if (redraw == null) redraw = true;
        if (hideResidues == null) hideResidues = false;
        if (decrementCounter == null) decrementCounter = true;
        if (!PLIP.AbstractViewer.isStructureEnabled(this.strucId)) {
            return;
        }
        if (decrementCounter) {
            _visible = Math.max(_visible - 1, 0);
        } else {
            alert("Did not decrement counter! " + this.id + ": " + _visible);
        }
        if (hideResidues) {
            _residues.forEach(function(residue) {
                residue.hide(false);
            });
        }
        if (_visible == 0 && redraw) {
            all_ligands.redraw(this.strucId);
        }
    };
    this.addResidue = function(residue) {
        if (_residues.indexOf(residue) == -1) {
            _residues.push(residue);
        }
    };
    this.addBond = function(bond) {
        if (_bonds.indexOf(bond) == -1) {
            _bonds.push(bond);
        }
    };
    this.isVisible = function() {
        return _visible > 0;
    };
    this.getLigandId = function() {
        return this.id;
    };
    this.getResidues = function() {
        return _residues;
    };
    this.strucId = strucId;
    this.id = position;
    this.position = position;
};

PLIP.AbstractViewerPV = {
    newBondMesh: function(name) {
        return pv_viewer.customMesh(name);
    },
    addWaterCircle: function(customMesh, coo, radius, color, add) {
        customMesh.addSphere(coo, radius, {
            cap: true,
            color: color
        });
    },
    addBondLine: function(customMesh, coo1, coo2, n, radius, color, add) {
        for (var i = 1; i <= n; i++) {
            customMesh.addTube(midpoint3d(coo1, coo2, (2 * i - 2) / (2 * n - 1)), midpoint3d(coo1, coo2, (2 * i - 1) / (2 * n - 1)), radius, {
                cap: true,
                color: color
            });
        }
    },
    removeMesh: function(name) {
        pv_viewer.rm(name);
        return undefined;
    },
    fitView: function(strucId, ligands, residues) {
        if (!this.isStructureEnabled(strucId)) {
            return;
        }
        var queryResidueChains = residues.map(function(residue) {
            return residue === undefined ? null : residue.chain;
        });
        var queryResiduePositions = residues.map(function(residue) {
            return residue === undefined ? null : residue.position;
        });
        if (queryResidueChains.length != queryResiduePositions.length) {
            console.error("queryResidue -Chains and -Positions array lengths don't match: " + queryResidueChains.length + " != " + queryResiduePositions.length);
        }
        var residues = pv_viewer.get(strucId).structure().residueSelect(function(res) {
            var chainName = res.chain().name();
            if (chainName === "_") {
                if (!ligands) {
                    return false;
                }
                ligand_positions = ligands.map(function(ligand) {
                    return ligand.position;
                });
                if (ligand_positions.includes(res.num())) {
                    return true;
                }
            }
            for (var i = 0; i < queryResidueChains.length; i++) {
                if (queryResidueChains[i] === chainName && queryResiduePositions[i] === res.num()) {
                    return true;
                }
            }
            return false;
        });
        pv_viewer.fitTo(residues);
        pv_viewer.requestRedraw();
    },
    isStructureEnabled: function(strucId) {
        return pv_viewer.get(strucId) ? true : false;
    }
};

PLIP.AbstractViewerNGL = {
    newBondMesh: function(name) {
        var shape = new NGL.Shape(name, {
            disableImpostor: true
        });
        return shape;
    },
    addWaterCircle: function(shape, coo, radius, color, add) {
        shape.addEllipsoid(coo, color, radius, [ radius, 0, 0 ], [ 0, radius, 0 ]);
        if (add) {
            let shapeComp = ngl_stage.addComponentFromObject(shape);
            shapeComp.addRepresentation("buffer");
        }
    },
    addBondLine: function(shape, coo1, coo2, n, radius, color, add) {
        for (let i = 1; i <= n; i++) {
            shape.addCylinder(midpoint3d(coo1, coo2, (2 * i - 2) / (2 * n - 1)), midpoint3d(coo1, coo2, (2 * i - 1) / (2 * n - 1)), color, radius);
        }
        if (add) {
            let shapeComp = ngl_stage.addComponentFromObject(shape);
            shapeComp.addRepresentation("buffer");
        }
    },
    removeMesh: function(name) {
        removeNGLComponentByName(name);
        return undefined;
    },
    fitView: function(strucId, ligands, residues) {
        if (!this.isStructureEnabled(strucId)) {
            return;
        }
        var seleStr = "";
        if (ligands) {
            ligand_positions = ligands.map(function(ligand) {
                return ligand.position;
            });
            seleStr += "((" + ligand_positions.join(" or ") + ") and :_)";
        }
        if (!(residues[0] === undefined)) {
            var chain_residue_map = {};
            for (var i in residues) {
                var residue = residues[i];
                var chain = residue.chain;
                if (chain != "-") {
                    var position = residue.position;
                    if (!chain_residue_map[chain]) {
                        chain_residue_map[chain] = [ position ];
                    } else {
                        chain_residue_map[chain].push(position);
                    }
                }
            }
            for (var chain in chain_residue_map) {
                if (seleStr != "") {
                    seleStr += " or ";
                }
                seleStr += "((" + chain_residue_map[chain].join(" or ") + ") and :" + chain + ")";
            }
        }
        getNGLComponentByName(strucId).autoView(seleStr, 1e3);
    },
    isStructureEnabled: function(strucId) {
        return getNGLComponentByName(strucId) ? true : false;
    }
};

function togglePLIPLigandContacts(id, strucId, bondIds) {
    if ($("#togglePLIPLigandContacts_" + id).hasClass("glyphicon-chevron-down")) {
        showPLIPLigandContacts(id, strucId, bondIds);
    } else {
        hidePLIPLigandContacts(id, strucId, bondIds);
    }
}

function showPLIPLigandContacts(id, strucId, bondIds) {
    $("#PLIPligandContacts_" + id).show(100);
    setTimeout(function() {
        $("#togglePLIPLigandContacts_" + id).addClass("glyphicon-chevron-up").removeClass("glyphicon-chevron-down");
    }, 100);
    if (!(bondIds === undefined)) {
        showBonds(strucId, bondIds);
        fitToBonds(strucId, bondIds);
    }
}

function hidePLIPLigandContacts(id, strucId, bondIds) {
    $("#PLIPligandContacts_" + id).hide(100);
    setTimeout(function() {
        $("#togglePLIPLigandContacts_" + id).addClass("glyphicon-chevron-down").removeClass("glyphicon-chevron-up");
    }, 100);
    if (!(bondIds === undefined)) {
        hideBonds(strucId, bondIds);
    }
}

var plip_interactions_fit_bonds = true;

function togglePlipInteractions(id, strucId, bondIds) {
    if ($("#togglePlipInteractions_" + id).hasClass("glyphicon-chevron-down")) {
        $("#togglePlipInteractions_" + id).removeClass("glyphicon-chevron-down").addClass("glyphicon-chevron-up");
        $("#plipInteractions_" + id).show();
        if (PLIP.AbstractViewer.isStructureEnabled(strucId)) {
            plip_interactions_fit_bonds = false;
            $.when($("#plipInteractions_" + id + " input").each(function(index, input) {
                $(input).prop("checked", true).trigger("change");
            })).done(function() {
                fitToBonds(strucId, bondIds);
            });
            plip_interactions_fit_bonds = true;
        }
    } else {
        $("#togglePlipInteractions_" + id).addClass("glyphicon-chevron-down").removeClass("glyphicon-chevron-up");
        $("#plipInteractions_" + id).hide();
        $("#plipInteractions_" + id + " input").each(function(index, input) {
            $(input).prop("checked", false).trigger("change");
        });
    }
}

function changeLigandContacts4Checkbox(strucId, ligandId, resIds, contactId) {
    var checkbox = $("#chainContacts4_" + contactId);
    var isChecked = checkbox.prop("checked");
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        checkbox.prop("checked", false);
        return;
    }
    plip_residues.getResiduesByResId(strucId, resIds).forEach(function(residue) {
        isChecked ? residue.show(false) : residue.hide(false);
    });
    var ligand = all_ligands.getLigandById(strucId, ligandId);
    isChecked ? ligand.show(true, false) : ligand.hide(true, null, false);
    plip_residues.redraw(strucId);
}

function clickLigandContacts4Checkbox(strucId, ligandId, resIds, contactId) {
    if (!PLIP.AbstractViewer.isStructureEnabled(strucId)) {
        return;
    }
    $("#chainContacts4_" + contactId).each(function(index, input) {
        $(input).prop("checked", !$(input).prop("checked")).trigger("change");
    });
    if ($("#chainContacts4_" + contactId).prop("checked")) {
        PLIP.AbstractViewer.fitView(strucId, [ all_ligands.getLigandById(strucId, ligandId) ], plip_residues.getResiduesByResId(strucId, resIds));
    }
}

function toggleResidueContacts(id, ligandId, strucId) {
    if ($("#toggleResidueContacts4_" + id).hasClass("glyphicon-chevron-down")) {
        $("#toggleResidueContacts4_" + id).removeClass("glyphicon-chevron-down").addClass("glyphicon-chevron-up");
        $("#residueContacts4_" + id).show();
        if (PLIP.AbstractViewer.isStructureEnabled(strucId)) {
            $("#residueContacts4_" + id + " input").each(function(index, input) {
                $(input).prop("checked", true).trigger("change");
            });
            ligand = all_ligands.getLigandById(strucId, ligandId);
            PLIP.AbstractViewer.fitView(strucId, [ ligand ], ligand.getResidues());
        }
    } else {
        $("#toggleResidueContacts4_" + id).removeClass("glyphicon-chevron-up").addClass("glyphicon-chevron-down");
        $("#residueContacts4_" + id).hide();
        if (PLIP.AbstractViewer.isStructureEnabled(strucId)) {
            $("#residueContacts4_" + id + " input").each(function(index, input) {
                $(input).prop("checked", false).trigger("change");
            });
        }
    }
    updatecols();
}

function toggleLigandGroup(id, ligands, strucId) {
    if ($("#toggleLigandGroup_" + id).hasClass("glyphicon-chevron-down")) {
        $("#toggleLigandGroup_" + id).addClass("glyphicon-chevron-up").removeClass("glyphicon-chevron-down");
        $("#ligandGroup_" + id).show();
        centerOnLigand(ligands, strucId);
    } else {
        $("#toggleLigandGroup_" + id).addClass("glyphicon-chevron-down").removeClass("glyphicon-chevron-up");
        $("#ligandGroup_" + id + " .glyphicon-chevron-up").each(function(index, elem) {
            $(elem).trigger("click");
        });
        $("#ligandGroup_" + id).hide();
    }
    updatecols();
}

function toggleAllLigands(show) {
    if ($("#toggleAllLink").hasClass("glyphicon-chevron-down")) {
        $("#toggleAllLink,span[id^=toggleLigandGroup_]").removeClass("glyphicon-chevron-down").addClass("glyphicon-chevron-up");
        $("div[id^=ligandGroup_]").show();
    } else {
        $("#toggleAllLink,span[id^=toggleLigandGroup_]").removeClass("glyphicon-chevron-up").addClass("glyphicon-chevron-down");
        $("div[id^=ligandGroup_]").hide();
    }
}

(function() {
    this.canvg = function(target, s, opts) {
        if (target == null && s == null && opts == null) {
            var svgTags = document.querySelectorAll("svg");
            for (var i = 0; i < svgTags.length; i++) {
                var svgTag = svgTags[i];
                var c = document.createElement("canvas");
                c.width = svgTag.clientWidth;
                c.height = svgTag.clientHeight;
                svgTag.parentNode.insertBefore(c, svgTag);
                svgTag.parentNode.removeChild(svgTag);
                var div = document.createElement("div");
                div.appendChild(svgTag);
                canvg(c, div.innerHTML);
            }
            return;
        }
        if (typeof target == "string") {
            target = document.getElementById(target);
        }
        if (target.svg != null) target.svg.stop();
        var svg = build(opts || {});
        if (!(target.childNodes.length == 1 && target.childNodes[0].nodeName == "OBJECT")) target.svg = svg;
        var ctx = target.getContext("2d");
        if (typeof s.documentElement != "undefined") {
            svg.loadXmlDoc(ctx, s);
        } else if (s.substr(0, 1) == "<") {
            svg.loadXml(ctx, s);
        } else {
            svg.load(ctx, s);
        }
    };
    var matchesSelector;
    if (typeof Element.prototype.matches != "undefined") {
        matchesSelector = function(node, selector) {
            return node.matches(selector);
        };
    } else if (typeof Element.prototype.webkitMatchesSelector != "undefined") {
        matchesSelector = function(node, selector) {
            return node.webkitMatchesSelector(selector);
        };
    } else if (typeof Element.prototype.mozMatchesSelector != "undefined") {
        matchesSelector = function(node, selector) {
            return node.mozMatchesSelector(selector);
        };
    } else if (typeof Element.prototype.msMatchesSelector != "undefined") {
        matchesSelector = function(node, selector) {
            return node.msMatchesSelector(selector);
        };
    } else if (typeof Element.prototype.oMatchesSelector != "undefined") {
        matchesSelector = function(node, selector) {
            return node.oMatchesSelector(selector);
        };
    } else {
        if (typeof jQuery === "function" || typeof Zepto === "function") {
            matchesSelector = function(node, selector) {
                return $(node).is(selector);
            };
        }
        if (typeof matchesSelector === "undefined") {
            matchesSelector = Sizzle.matchesSelector;
        }
    }
    var attributeRegex = /(\[[^\]]+\])/g;
    var idRegex = /(#[^\s\+>~\.\[:]+)/g;
    var classRegex = /(\.[^\s\+>~\.\[:]+)/g;
    var pseudoElementRegex = /(::[^\s\+>~\.\[:]+|:first-line|:first-letter|:before|:after)/gi;
    var pseudoClassWithBracketsRegex = /(:[\w-]+\([^\)]*\))/gi;
    var pseudoClassRegex = /(:[^\s\+>~\.\[:]+)/g;
    var elementRegex = /([^\s\+>~\.\[:]+)/g;
    function getSelectorSpecificity(selector) {
        var typeCount = [ 0, 0, 0 ];
        var findMatch = function(regex, type) {
            var matches = selector.match(regex);
            if (matches == null) {
                return;
            }
            typeCount[type] += matches.length;
            selector = selector.replace(regex, " ");
        };
        selector = selector.replace(/:not\(([^\)]*)\)/g, "     $1 ");
        selector = selector.replace(/{[^]*/gm, " ");
        findMatch(attributeRegex, 1);
        findMatch(idRegex, 0);
        findMatch(classRegex, 1);
        findMatch(pseudoElementRegex, 2);
        findMatch(pseudoClassWithBracketsRegex, 1);
        findMatch(pseudoClassRegex, 1);
        selector = selector.replace(/[\*\s\+>~]/g, " ");
        selector = selector.replace(/[#\.]/g, " ");
        findMatch(elementRegex, 2);
        return typeCount.join("");
    }
    function build(opts) {
        var svg = {
            opts: opts
        };
        svg.FRAMERATE = 30;
        svg.MAX_VIRTUAL_PIXELS = 3e4;
        svg.log = function(msg) {};
        if (svg.opts["log"] == true && typeof console != "undefined") {
            svg.log = function(msg) {
                console.log(msg);
            };
        }
        svg.init = function(ctx) {
            var uniqueId = 0;
            svg.UniqueId = function() {
                uniqueId++;
                return "canvg" + uniqueId;
            };
            svg.Definitions = {};
            svg.Styles = {};
            svg.StylesSpecificity = {};
            svg.Animations = [];
            svg.Images = [];
            svg.ctx = ctx;
            svg.ViewPort = new function() {
                this.viewPorts = [];
                this.Clear = function() {
                    this.viewPorts = [];
                };
                this.SetCurrent = function(width, height) {
                    this.viewPorts.push({
                        width: width,
                        height: height
                    });
                };
                this.RemoveCurrent = function() {
                    this.viewPorts.pop();
                };
                this.Current = function() {
                    return this.viewPorts[this.viewPorts.length - 1];
                };
                this.width = function() {
                    return this.Current().width;
                };
                this.height = function() {
                    return this.Current().height;
                };
                this.ComputeSize = function(d) {
                    if (d != null && typeof d == "number") return d;
                    if (d == "x") return this.width();
                    if (d == "y") return this.height();
                    return Math.sqrt(Math.pow(this.width(), 2) + Math.pow(this.height(), 2)) / Math.sqrt(2);
                };
            }();
        };
        svg.init();
        svg.ImagesLoaded = function() {
            for (var i = 0; i < svg.Images.length; i++) {
                if (!svg.Images[i].loaded) return false;
            }
            return true;
        };
        svg.trim = function(s) {
            return s.replace(/^\s+|\s+$/g, "");
        };
        svg.compressSpaces = function(s) {
            return s.replace(/[\s\r\t\n]+/gm, " ");
        };
        svg.ajax = function(url) {
            var AJAX;
            if (window.XMLHttpRequest) {
                AJAX = new XMLHttpRequest();
            } else {
                AJAX = new ActiveXObject("Microsoft.XMLHTTP");
            }
            if (AJAX) {
                AJAX.open("GET", url, false);
                AJAX.send(null);
                return AJAX.responseText;
            }
            return null;
        };
        svg.parseXml = function(xml) {
            if (typeof Windows != "undefined" && typeof Windows.Data != "undefined" && typeof Windows.Data.Xml != "undefined") {
                var xmlDoc = new Windows.Data.Xml.Dom.XmlDocument();
                var settings = new Windows.Data.Xml.Dom.XmlLoadSettings();
                settings.prohibitDtd = false;
                xmlDoc.loadXml(xml, settings);
                return xmlDoc;
            } else if (window.DOMParser) {
                var parser = new DOMParser();
                return parser.parseFromString(xml, "text/xml");
            } else {
                xml = xml.replace(/<!DOCTYPE svg[^>]*>/, "");
                var xmlDoc = new ActiveXObject("Microsoft.XMLDOM");
                xmlDoc.async = "false";
                xmlDoc.loadXML(xml);
                return xmlDoc;
            }
        };
        svg.Property = function(name, value) {
            this.name = name;
            this.value = value;
        };
        svg.Property.prototype.getValue = function() {
            return this.value;
        };
        svg.Property.prototype.hasValue = function() {
            return this.value != null && this.value !== "";
        };
        svg.Property.prototype.numValue = function() {
            if (!this.hasValue()) return 0;
            var n = parseFloat(this.value);
            if ((this.value + "").match(/%$/)) {
                n = n / 100;
            }
            return n;
        };
        svg.Property.prototype.valueOrDefault = function(def) {
            if (this.hasValue()) return this.value;
            return def;
        };
        svg.Property.prototype.numValueOrDefault = function(def) {
            if (this.hasValue()) return this.numValue();
            return def;
        };
        svg.Property.prototype.addOpacity = function(opacityProp) {
            var newValue = this.value;
            if (opacityProp.value != null && opacityProp.value != "" && typeof this.value == "string") {
                var color = new RGBColor(this.value);
                if (color.ok) {
                    newValue = "rgba(" + color.r + ", " + color.g + ", " + color.b + ", " + opacityProp.numValue() + ")";
                }
            }
            return new svg.Property(this.name, newValue);
        };
        svg.Property.prototype.getDefinition = function() {
            var name = this.value.match(/#([^\)'"]+)/);
            if (name) {
                name = name[1];
            }
            if (!name) {
                name = this.value;
            }
            return svg.Definitions[name];
        };
        svg.Property.prototype.isUrlDefinition = function() {
            return this.value.indexOf("url(") == 0;
        };
        svg.Property.prototype.getFillStyleDefinition = function(e, opacityProp) {
            var def = this.getDefinition();
            if (def != null && def.createGradient) {
                return def.createGradient(svg.ctx, e, opacityProp);
            }
            if (def != null && def.createPattern) {
                if (def.getHrefAttribute().hasValue()) {
                    var pt = def.attribute("patternTransform");
                    def = def.getHrefAttribute().getDefinition();
                    if (pt.hasValue()) {
                        def.attribute("patternTransform", true).value = pt.value;
                    }
                }
                return def.createPattern(svg.ctx, e);
            }
            return null;
        };
        svg.Property.prototype.getDPI = function(viewPort) {
            return 96;
        };
        svg.Property.prototype.getEM = function(viewPort) {
            var em = 12;
            var fontSize = new svg.Property("fontSize", svg.Font.Parse(svg.ctx.font).fontSize);
            if (fontSize.hasValue()) em = fontSize.toPixels(viewPort);
            return em;
        };
        svg.Property.prototype.getUnits = function() {
            var s = this.value + "";
            return s.replace(/[0-9\.\-]/g, "");
        };
        svg.Property.prototype.toPixels = function(viewPort, processPercent) {
            if (!this.hasValue()) return 0;
            var s = this.value + "";
            if (s.match(/em$/)) return this.numValue() * this.getEM(viewPort);
            if (s.match(/ex$/)) return this.numValue() * this.getEM(viewPort) / 2;
            if (s.match(/px$/)) return this.numValue();
            if (s.match(/pt$/)) return this.numValue() * this.getDPI(viewPort) * (1 / 72);
            if (s.match(/pc$/)) return this.numValue() * 15;
            if (s.match(/cm$/)) return this.numValue() * this.getDPI(viewPort) / 2.54;
            if (s.match(/mm$/)) return this.numValue() * this.getDPI(viewPort) / 25.4;
            if (s.match(/in$/)) return this.numValue() * this.getDPI(viewPort);
            if (s.match(/%$/)) return this.numValue() * svg.ViewPort.ComputeSize(viewPort);
            var n = this.numValue();
            if (processPercent && n < 1) return n * svg.ViewPort.ComputeSize(viewPort);
            return n;
        };
        svg.Property.prototype.toMilliseconds = function() {
            if (!this.hasValue()) return 0;
            var s = this.value + "";
            if (s.match(/s$/)) return this.numValue() * 1e3;
            if (s.match(/ms$/)) return this.numValue();
            return this.numValue();
        };
        svg.Property.prototype.toRadians = function() {
            if (!this.hasValue()) return 0;
            var s = this.value + "";
            if (s.match(/deg$/)) return this.numValue() * (Math.PI / 180);
            if (s.match(/grad$/)) return this.numValue() * (Math.PI / 200);
            if (s.match(/rad$/)) return this.numValue();
            return this.numValue() * (Math.PI / 180);
        };
        var textBaselineMapping = {
            baseline: "alphabetic",
            "before-edge": "top",
            "text-before-edge": "top",
            middle: "middle",
            central: "middle",
            "after-edge": "bottom",
            "text-after-edge": "bottom",
            ideographic: "ideographic",
            alphabetic: "alphabetic",
            hanging: "hanging",
            mathematical: "alphabetic"
        };
        svg.Property.prototype.toTextBaseline = function() {
            if (!this.hasValue()) return null;
            return textBaselineMapping[this.value];
        };
        svg.Font = new function() {
            this.Styles = "normal|italic|oblique|inherit";
            this.Variants = "normal|small-caps|inherit";
            this.Weights = "normal|bold|bolder|lighter|100|200|300|400|500|600|700|800|900|inherit";
            this.CreateFont = function(fontStyle, fontVariant, fontWeight, fontSize, fontFamily, inherit) {
                var f = inherit != null ? this.Parse(inherit) : this.CreateFont("", "", "", "", "", svg.ctx.font);
                return {
                    fontFamily: fontFamily || f.fontFamily,
                    fontSize: fontSize || f.fontSize,
                    fontStyle: fontStyle || f.fontStyle,
                    fontWeight: fontWeight || f.fontWeight,
                    fontVariant: fontVariant || f.fontVariant,
                    toString: function() {
                        return [ this.fontStyle, this.fontVariant, this.fontWeight, this.fontSize, this.fontFamily ].join(" ");
                    }
                };
            };
            var that = this;
            this.Parse = function(s) {
                var f = {};
                var d = svg.trim(svg.compressSpaces(s || "")).split(" ");
                var set = {
                    fontSize: false,
                    fontStyle: false,
                    fontWeight: false,
                    fontVariant: false
                };
                var ff = "";
                for (var i = 0; i < d.length; i++) {
                    if (!set.fontStyle && that.Styles.indexOf(d[i]) != -1) {
                        if (d[i] != "inherit") f.fontStyle = d[i];
                        set.fontStyle = true;
                    } else if (!set.fontVariant && that.Variants.indexOf(d[i]) != -1) {
                        if (d[i] != "inherit") f.fontVariant = d[i];
                        set.fontStyle = set.fontVariant = true;
                    } else if (!set.fontWeight && that.Weights.indexOf(d[i]) != -1) {
                        if (d[i] != "inherit") f.fontWeight = d[i];
                        set.fontStyle = set.fontVariant = set.fontWeight = true;
                    } else if (!set.fontSize) {
                        if (d[i] != "inherit") f.fontSize = d[i].split("/")[0];
                        set.fontStyle = set.fontVariant = set.fontWeight = set.fontSize = true;
                    } else {
                        if (d[i] != "inherit") ff += d[i];
                    }
                }
                if (ff != "") f.fontFamily = ff;
                return f;
            };
        }();
        svg.ToNumberArray = function(s) {
            var a = svg.trim(svg.compressSpaces((s || "").replace(/,/g, " "))).split(" ");
            for (var i = 0; i < a.length; i++) {
                a[i] = parseFloat(a[i]);
            }
            return a;
        };
        svg.Point = function(x, y) {
            this.x = x;
            this.y = y;
        };
        svg.Point.prototype.angleTo = function(p) {
            return Math.atan2(p.y - this.y, p.x - this.x);
        };
        svg.Point.prototype.applyTransform = function(v) {
            var xp = this.x * v[0] + this.y * v[2] + v[4];
            var yp = this.x * v[1] + this.y * v[3] + v[5];
            this.x = xp;
            this.y = yp;
        };
        svg.CreatePoint = function(s) {
            var a = svg.ToNumberArray(s);
            return new svg.Point(a[0], a[1]);
        };
        svg.CreatePath = function(s) {
            var a = svg.ToNumberArray(s);
            var path = [];
            for (var i = 0; i < a.length; i += 2) {
                path.push(new svg.Point(a[i], a[i + 1]));
            }
            return path;
        };
        svg.BoundingBox = function(x1, y1, x2, y2) {
            this.x1 = Number.NaN;
            this.y1 = Number.NaN;
            this.x2 = Number.NaN;
            this.y2 = Number.NaN;
            this.x = function() {
                return this.x1;
            };
            this.y = function() {
                return this.y1;
            };
            this.width = function() {
                return this.x2 - this.x1;
            };
            this.height = function() {
                return this.y2 - this.y1;
            };
            this.addPoint = function(x, y) {
                if (x != null) {
                    if (isNaN(this.x1) || isNaN(this.x2)) {
                        this.x1 = x;
                        this.x2 = x;
                    }
                    if (x < this.x1) this.x1 = x;
                    if (x > this.x2) this.x2 = x;
                }
                if (y != null) {
                    if (isNaN(this.y1) || isNaN(this.y2)) {
                        this.y1 = y;
                        this.y2 = y;
                    }
                    if (y < this.y1) this.y1 = y;
                    if (y > this.y2) this.y2 = y;
                }
            };
            this.addX = function(x) {
                this.addPoint(x, null);
            };
            this.addY = function(y) {
                this.addPoint(null, y);
            };
            this.addBoundingBox = function(bb) {
                this.addPoint(bb.x1, bb.y1);
                this.addPoint(bb.x2, bb.y2);
            };
            this.addQuadraticCurve = function(p0x, p0y, p1x, p1y, p2x, p2y) {
                var cp1x = p0x + 2 / 3 * (p1x - p0x);
                var cp1y = p0y + 2 / 3 * (p1y - p0y);
                var cp2x = cp1x + 1 / 3 * (p2x - p0x);
                var cp2y = cp1y + 1 / 3 * (p2y - p0y);
                this.addBezierCurve(p0x, p0y, cp1x, cp2x, cp1y, cp2y, p2x, p2y);
            };
            this.addBezierCurve = function(p0x, p0y, p1x, p1y, p2x, p2y, p3x, p3y) {
                var p0 = [ p0x, p0y ], p1 = [ p1x, p1y ], p2 = [ p2x, p2y ], p3 = [ p3x, p3y ];
                this.addPoint(p0[0], p0[1]);
                this.addPoint(p3[0], p3[1]);
                for (i = 0; i <= 1; i++) {
                    var f = function(t) {
                        return Math.pow(1 - t, 3) * p0[i] + 3 * Math.pow(1 - t, 2) * t * p1[i] + 3 * (1 - t) * Math.pow(t, 2) * p2[i] + Math.pow(t, 3) * p3[i];
                    };
                    var b = 6 * p0[i] - 12 * p1[i] + 6 * p2[i];
                    var a = -3 * p0[i] + 9 * p1[i] - 9 * p2[i] + 3 * p3[i];
                    var c = 3 * p1[i] - 3 * p0[i];
                    if (a == 0) {
                        if (b == 0) continue;
                        var t = -c / b;
                        if (0 < t && t < 1) {
                            if (i == 0) this.addX(f(t));
                            if (i == 1) this.addY(f(t));
                        }
                        continue;
                    }
                    var b2ac = Math.pow(b, 2) - 4 * c * a;
                    if (b2ac < 0) continue;
                    var t1 = (-b + Math.sqrt(b2ac)) / (2 * a);
                    if (0 < t1 && t1 < 1) {
                        if (i == 0) this.addX(f(t1));
                        if (i == 1) this.addY(f(t1));
                    }
                    var t2 = (-b - Math.sqrt(b2ac)) / (2 * a);
                    if (0 < t2 && t2 < 1) {
                        if (i == 0) this.addX(f(t2));
                        if (i == 1) this.addY(f(t2));
                    }
                }
            };
            this.isPointInBox = function(x, y) {
                return this.x1 <= x && x <= this.x2 && this.y1 <= y && y <= this.y2;
            };
            this.addPoint(x1, y1);
            this.addPoint(x2, y2);
        };
        svg.Transform = function(v) {
            var that = this;
            this.Type = {};
            this.Type.translate = function(s) {
                this.p = svg.CreatePoint(s);
                this.apply = function(ctx) {
                    ctx.translate(this.p.x || 0, this.p.y || 0);
                };
                this.unapply = function(ctx) {
                    ctx.translate(-1 * this.p.x || 0, -1 * this.p.y || 0);
                };
                this.applyToPoint = function(p) {
                    p.applyTransform([ 1, 0, 0, 1, this.p.x || 0, this.p.y || 0 ]);
                };
            };
            this.Type.rotate = function(s) {
                var a = svg.ToNumberArray(s);
                this.angle = new svg.Property("angle", a[0]);
                this.cx = a[1] || 0;
                this.cy = a[2] || 0;
                this.apply = function(ctx) {
                    ctx.translate(this.cx, this.cy);
                    ctx.rotate(this.angle.toRadians());
                    ctx.translate(-this.cx, -this.cy);
                };
                this.unapply = function(ctx) {
                    ctx.translate(this.cx, this.cy);
                    ctx.rotate(-1 * this.angle.toRadians());
                    ctx.translate(-this.cx, -this.cy);
                };
                this.applyToPoint = function(p) {
                    var a = this.angle.toRadians();
                    p.applyTransform([ 1, 0, 0, 1, this.p.x || 0, this.p.y || 0 ]);
                    p.applyTransform([ Math.cos(a), Math.sin(a), -Math.sin(a), Math.cos(a), 0, 0 ]);
                    p.applyTransform([ 1, 0, 0, 1, -this.p.x || 0, -this.p.y || 0 ]);
                };
            };
            this.Type.scale = function(s) {
                this.p = svg.CreatePoint(s);
                this.apply = function(ctx) {
                    ctx.scale(this.p.x || 1, this.p.y || this.p.x || 1);
                };
                this.unapply = function(ctx) {
                    ctx.scale(1 / this.p.x || 1, 1 / this.p.y || this.p.x || 1);
                };
                this.applyToPoint = function(p) {
                    p.applyTransform([ this.p.x || 0, 0, 0, this.p.y || 0, 0, 0 ]);
                };
            };
            this.Type.matrix = function(s) {
                this.m = svg.ToNumberArray(s);
                this.apply = function(ctx) {
                    ctx.transform(this.m[0], this.m[1], this.m[2], this.m[3], this.m[4], this.m[5]);
                };
                this.unapply = function(ctx) {
                    var a = this.m[0];
                    var b = this.m[2];
                    var c = this.m[4];
                    var d = this.m[1];
                    var e = this.m[3];
                    var f = this.m[5];
                    var g = 0;
                    var h = 0;
                    var i = 1;
                    var det = 1 / (a * (e * i - f * h) - b * (d * i - f * g) + c * (d * h - e * g));
                    ctx.transform(det * (e * i - f * h), det * (f * g - d * i), det * (c * h - b * i), det * (a * i - c * g), det * (b * f - c * e), det * (c * d - a * f));
                };
                this.applyToPoint = function(p) {
                    p.applyTransform(this.m);
                };
            };
            this.Type.SkewBase = function(s) {
                this.base = that.Type.matrix;
                this.base(s);
                this.angle = new svg.Property("angle", s);
            };
            this.Type.SkewBase.prototype = new this.Type.matrix();
            this.Type.skewX = function(s) {
                this.base = that.Type.SkewBase;
                this.base(s);
                this.m = [ 1, 0, Math.tan(this.angle.toRadians()), 1, 0, 0 ];
            };
            this.Type.skewX.prototype = new this.Type.SkewBase();
            this.Type.skewY = function(s) {
                this.base = that.Type.SkewBase;
                this.base(s);
                this.m = [ 1, Math.tan(this.angle.toRadians()), 0, 1, 0, 0 ];
            };
            this.Type.skewY.prototype = new this.Type.SkewBase();
            this.transforms = [];
            this.apply = function(ctx) {
                for (var i = 0; i < this.transforms.length; i++) {
                    this.transforms[i].apply(ctx);
                }
            };
            this.unapply = function(ctx) {
                for (var i = this.transforms.length - 1; i >= 0; i--) {
                    this.transforms[i].unapply(ctx);
                }
            };
            this.applyToPoint = function(p) {
                for (var i = 0; i < this.transforms.length; i++) {
                    this.transforms[i].applyToPoint(p);
                }
            };
            var data = svg.trim(svg.compressSpaces(v)).replace(/\)([a-zA-Z])/g, ") $1").replace(/\)(\s?,\s?)/g, ") ").split(/\s(?=[a-z])/);
            for (var i = 0; i < data.length; i++) {
                var type = svg.trim(data[i].split("(")[0]);
                var s = data[i].split("(")[1].replace(")", "");
                var transform = new this.Type[type](s);
                transform.type = type;
                this.transforms.push(transform);
            }
        };
        svg.AspectRatio = function(ctx, aspectRatio, width, desiredWidth, height, desiredHeight, minX, minY, refX, refY) {
            aspectRatio = svg.compressSpaces(aspectRatio);
            aspectRatio = aspectRatio.replace(/^defer\s/, "");
            var align = aspectRatio.split(" ")[0] || "xMidYMid";
            var meetOrSlice = aspectRatio.split(" ")[1] || "meet";
            var scaleX = width / desiredWidth;
            var scaleY = height / desiredHeight;
            var scaleMin = Math.min(scaleX, scaleY);
            var scaleMax = Math.max(scaleX, scaleY);
            if (meetOrSlice == "meet") {
                desiredWidth *= scaleMin;
                desiredHeight *= scaleMin;
            }
            if (meetOrSlice == "slice") {
                desiredWidth *= scaleMax;
                desiredHeight *= scaleMax;
            }
            refX = new svg.Property("refX", refX);
            refY = new svg.Property("refY", refY);
            if (refX.hasValue() && refY.hasValue()) {
                ctx.translate(-scaleMin * refX.toPixels("x"), -scaleMin * refY.toPixels("y"));
            } else {
                if (align.match(/^xMid/) && (meetOrSlice == "meet" && scaleMin == scaleY || meetOrSlice == "slice" && scaleMax == scaleY)) ctx.translate(width / 2 - desiredWidth / 2, 0);
                if (align.match(/YMid$/) && (meetOrSlice == "meet" && scaleMin == scaleX || meetOrSlice == "slice" && scaleMax == scaleX)) ctx.translate(0, height / 2 - desiredHeight / 2);
                if (align.match(/^xMax/) && (meetOrSlice == "meet" && scaleMin == scaleY || meetOrSlice == "slice" && scaleMax == scaleY)) ctx.translate(width - desiredWidth, 0);
                if (align.match(/YMax$/) && (meetOrSlice == "meet" && scaleMin == scaleX || meetOrSlice == "slice" && scaleMax == scaleX)) ctx.translate(0, height - desiredHeight);
            }
            if (align == "none") ctx.scale(scaleX, scaleY); else if (meetOrSlice == "meet") ctx.scale(scaleMin, scaleMin); else if (meetOrSlice == "slice") ctx.scale(scaleMax, scaleMax);
            ctx.translate(minX == null ? 0 : -minX, minY == null ? 0 : -minY);
        };
        svg.Element = {};
        svg.EmptyProperty = new svg.Property("EMPTY", "");
        svg.Element.ElementBase = function(node) {
            this.attributes = {};
            this.styles = {};
            this.stylesSpecificity = {};
            this.children = [];
            this.attribute = function(name, createIfNotExists) {
                var a = this.attributes[name];
                if (a != null) return a;
                if (createIfNotExists == true) {
                    a = new svg.Property(name, "");
                    this.attributes[name] = a;
                }
                return a || svg.EmptyProperty;
            };
            this.getHrefAttribute = function() {
                for (var a in this.attributes) {
                    if (a == "href" || a.match(/:href$/)) {
                        return this.attributes[a];
                    }
                }
                return svg.EmptyProperty;
            };
            this.style = function(name, createIfNotExists, skipAncestors) {
                var s = this.styles[name];
                if (s != null) return s;
                var a = this.attribute(name);
                if (a != null && a.hasValue()) {
                    this.styles[name] = a;
                    return a;
                }
                if (skipAncestors != true) {
                    var p = this.parent;
                    if (p != null) {
                        var ps = p.style(name);
                        if (ps != null && ps.hasValue()) {
                            return ps;
                        }
                    }
                }
                if (createIfNotExists == true) {
                    s = new svg.Property(name, "");
                    this.styles[name] = s;
                }
                return s || svg.EmptyProperty;
            };
            this.render = function(ctx) {
                if (this.style("display").value == "none") return;
                if (this.style("visibility").value == "hidden") return;
                ctx.save();
                if (this.attribute("mask").hasValue()) {
                    var mask = this.attribute("mask").getDefinition();
                    if (mask != null) mask.apply(ctx, this);
                } else if (this.style("filter").hasValue()) {
                    var filter = this.style("filter").getDefinition();
                    if (filter != null) filter.apply(ctx, this);
                } else {
                    this.setContext(ctx);
                    this.renderChildren(ctx);
                    this.clearContext(ctx);
                }
                ctx.restore();
            };
            this.setContext = function(ctx) {};
            this.clearContext = function(ctx) {};
            this.renderChildren = function(ctx) {
                for (var i = 0; i < this.children.length; i++) {
                    this.children[i].render(ctx);
                }
            };
            this.addChild = function(childNode, create) {
                var child = childNode;
                if (create) child = svg.CreateElement(childNode);
                child.parent = this;
                if (child.type != "title") {
                    this.children.push(child);
                }
            };
            this.addStylesFromStyleDefinition = function() {
                for (var selector in svg.Styles) {
                    if (selector[0] != "@" && matchesSelector(node, selector)) {
                        var styles = svg.Styles[selector];
                        var specificity = svg.StylesSpecificity[selector];
                        if (styles != null) {
                            for (var name in styles) {
                                var existingSpecificity = this.stylesSpecificity[name];
                                if (typeof existingSpecificity == "undefined") {
                                    existingSpecificity = "000";
                                }
                                if (specificity > existingSpecificity) {
                                    this.styles[name] = styles[name];
                                    this.stylesSpecificity[name] = specificity;
                                }
                            }
                        }
                    }
                }
            };
            if (node != null && node.nodeType == 1) {
                for (var i = 0; i < node.attributes.length; i++) {
                    var attribute = node.attributes[i];
                    this.attributes[attribute.nodeName] = new svg.Property(attribute.nodeName, attribute.value);
                }
                this.addStylesFromStyleDefinition();
                if (this.attribute("style").hasValue()) {
                    var styles = this.attribute("style").value.split(";");
                    for (var i = 0; i < styles.length; i++) {
                        if (svg.trim(styles[i]) != "") {
                            var style = styles[i].split(":");
                            var name = svg.trim(style[0]);
                            var value = svg.trim(style[1]);
                            this.styles[name] = new svg.Property(name, value);
                        }
                    }
                }
                if (this.attribute("id").hasValue()) {
                    if (svg.Definitions[this.attribute("id").value] == null) {
                        svg.Definitions[this.attribute("id").value] = this;
                    }
                }
                for (var i = 0; i < node.childNodes.length; i++) {
                    var childNode = node.childNodes[i];
                    if (childNode.nodeType == 1) this.addChild(childNode, true);
                    if (this.captureTextNodes && (childNode.nodeType == 3 || childNode.nodeType == 4)) {
                        var text = childNode.value || childNode.text || childNode.textContent || "";
                        if (svg.compressSpaces(text) != "") {
                            this.addChild(new svg.Element.tspan(childNode), false);
                        }
                    }
                }
            }
        };
        svg.Element.RenderedElementBase = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.setContext = function(ctx) {
                if (this.style("fill").isUrlDefinition()) {
                    var fs = this.style("fill").getFillStyleDefinition(this, this.style("fill-opacity"));
                    if (fs != null) ctx.fillStyle = fs;
                } else if (this.style("fill").hasValue()) {
                    var fillStyle = this.style("fill");
                    if (fillStyle.value == "currentColor") fillStyle.value = this.style("color").value;
                    if (fillStyle.value != "inherit") ctx.fillStyle = fillStyle.value == "none" ? "rgba(0,0,0,0)" : fillStyle.value;
                }
                if (this.style("fill-opacity").hasValue()) {
                    var fillStyle = new svg.Property("fill", ctx.fillStyle);
                    fillStyle = fillStyle.addOpacity(this.style("fill-opacity"));
                    ctx.fillStyle = fillStyle.value;
                }
                if (this.style("stroke").isUrlDefinition()) {
                    var fs = this.style("stroke").getFillStyleDefinition(this, this.style("stroke-opacity"));
                    if (fs != null) ctx.strokeStyle = fs;
                } else if (this.style("stroke").hasValue()) {
                    var strokeStyle = this.style("stroke");
                    if (strokeStyle.value == "currentColor") strokeStyle.value = this.style("color").value;
                    if (strokeStyle.value != "inherit") ctx.strokeStyle = strokeStyle.value == "none" ? "rgba(0,0,0,0)" : strokeStyle.value;
                }
                if (this.style("stroke-opacity").hasValue()) {
                    var strokeStyle = new svg.Property("stroke", ctx.strokeStyle);
                    strokeStyle = strokeStyle.addOpacity(this.style("stroke-opacity"));
                    ctx.strokeStyle = strokeStyle.value;
                }
                if (this.style("stroke-width").hasValue()) {
                    var newLineWidth = this.style("stroke-width").toPixels();
                    ctx.lineWidth = newLineWidth == 0 ? .001 : newLineWidth;
                }
                if (this.style("stroke-linecap").hasValue()) ctx.lineCap = this.style("stroke-linecap").value;
                if (this.style("stroke-linejoin").hasValue()) ctx.lineJoin = this.style("stroke-linejoin").value;
                if (this.style("stroke-miterlimit").hasValue()) ctx.miterLimit = this.style("stroke-miterlimit").value;
                if (this.style("stroke-dasharray").hasValue() && this.style("stroke-dasharray").value != "none") {
                    var gaps = svg.ToNumberArray(this.style("stroke-dasharray").value);
                    if (typeof ctx.setLineDash != "undefined") {
                        ctx.setLineDash(gaps);
                    } else if (typeof ctx.webkitLineDash != "undefined") {
                        ctx.webkitLineDash = gaps;
                    } else if (typeof ctx.mozDash != "undefined" && !(gaps.length == 1 && gaps[0] == 0)) {
                        ctx.mozDash = gaps;
                    }
                    var offset = this.style("stroke-dashoffset").numValueOrDefault(1);
                    if (typeof ctx.lineDashOffset != "undefined") {
                        ctx.lineDashOffset = offset;
                    } else if (typeof ctx.webkitLineDashOffset != "undefined") {
                        ctx.webkitLineDashOffset = offset;
                    } else if (typeof ctx.mozDashOffset != "undefined") {
                        ctx.mozDashOffset = offset;
                    }
                }
                if (typeof ctx.font != "undefined") {
                    ctx.font = svg.Font.CreateFont(this.style("font-style").value, this.style("font-variant").value, this.style("font-weight").value, this.style("font-size").hasValue() ? this.style("font-size").toPixels() + "px" : "", this.style("font-family").value).toString();
                }
                if (this.style("transform", false, true).hasValue()) {
                    var transform = new svg.Transform(this.style("transform", false, true).value);
                    transform.apply(ctx);
                }
                if (this.attribute("clip-path", false, true).hasValue()) {
                    var clip = this.attribute("clip-path", false, true).getDefinition();
                    if (clip != null) clip.apply(ctx);
                }
                if (this.style("opacity").hasValue()) {
                    ctx.globalAlpha = this.style("opacity").numValue();
                }
            };
        };
        svg.Element.RenderedElementBase.prototype = new svg.Element.ElementBase();
        svg.Element.PathElementBase = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.path = function(ctx) {
                if (ctx != null) ctx.beginPath();
                return new svg.BoundingBox();
            };
            this.renderChildren = function(ctx) {
                this.path(ctx);
                svg.Mouse.checkPath(this, ctx);
                if (ctx.fillStyle != "") {
                    if (this.style("fill-rule").valueOrDefault("inherit") != "inherit") {
                        ctx.fill(this.style("fill-rule").value);
                    } else {
                        ctx.fill();
                    }
                }
                if (ctx.strokeStyle != "") ctx.stroke();
                var markers = this.getMarkers();
                if (markers != null) {
                    if (this.style("marker-start").isUrlDefinition()) {
                        var marker = this.style("marker-start").getDefinition();
                        marker.render(ctx, markers[0][0], markers[0][1]);
                    }
                    if (this.style("marker-mid").isUrlDefinition()) {
                        var marker = this.style("marker-mid").getDefinition();
                        for (var i = 1; i < markers.length - 1; i++) {
                            marker.render(ctx, markers[i][0], markers[i][1]);
                        }
                    }
                    if (this.style("marker-end").isUrlDefinition()) {
                        var marker = this.style("marker-end").getDefinition();
                        marker.render(ctx, markers[markers.length - 1][0], markers[markers.length - 1][1]);
                    }
                }
            };
            this.getBoundingBox = function() {
                return this.path();
            };
            this.getMarkers = function() {
                return null;
            };
        };
        svg.Element.PathElementBase.prototype = new svg.Element.RenderedElementBase();
        svg.Element.svg = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.baseClearContext = this.clearContext;
            this.clearContext = function(ctx) {
                this.baseClearContext(ctx);
                svg.ViewPort.RemoveCurrent();
            };
            this.baseSetContext = this.setContext;
            this.setContext = function(ctx) {
                ctx.strokeStyle = "rgba(0,0,0,0)";
                ctx.lineCap = "butt";
                ctx.lineJoin = "miter";
                ctx.miterLimit = 4;
                if (typeof ctx.font != "undefined" && typeof window.getComputedStyle != "undefined") {
                    ctx.font = window.getComputedStyle(ctx.canvas).getPropertyValue("font");
                }
                this.baseSetContext(ctx);
                if (!this.attribute("x").hasValue()) this.attribute("x", true).value = 0;
                if (!this.attribute("y").hasValue()) this.attribute("y", true).value = 0;
                ctx.translate(this.attribute("x").toPixels("x"), this.attribute("y").toPixels("y"));
                var width = svg.ViewPort.width();
                var height = svg.ViewPort.height();
                if (!this.attribute("width").hasValue()) this.attribute("width", true).value = "100%";
                if (!this.attribute("height").hasValue()) this.attribute("height", true).value = "100%";
                if (typeof this.root == "undefined") {
                    width = this.attribute("width").toPixels("x");
                    height = this.attribute("height").toPixels("y");
                    var x = 0;
                    var y = 0;
                    if (this.attribute("refX").hasValue() && this.attribute("refY").hasValue()) {
                        x = -this.attribute("refX").toPixels("x");
                        y = -this.attribute("refY").toPixels("y");
                    }
                    if (this.attribute("overflow").valueOrDefault("hidden") != "visible") {
                        ctx.beginPath();
                        ctx.moveTo(x, y);
                        ctx.lineTo(width, y);
                        ctx.lineTo(width, height);
                        ctx.lineTo(x, height);
                        ctx.closePath();
                        ctx.clip();
                    }
                }
                svg.ViewPort.SetCurrent(width, height);
                if (this.attribute("viewBox").hasValue()) {
                    var viewBox = svg.ToNumberArray(this.attribute("viewBox").value);
                    var minX = viewBox[0];
                    var minY = viewBox[1];
                    width = viewBox[2];
                    height = viewBox[3];
                    svg.AspectRatio(ctx, this.attribute("preserveAspectRatio").value, svg.ViewPort.width(), width, svg.ViewPort.height(), height, minX, minY, this.attribute("refX").value, this.attribute("refY").value);
                    svg.ViewPort.RemoveCurrent();
                    svg.ViewPort.SetCurrent(viewBox[2], viewBox[3]);
                }
            };
        };
        svg.Element.svg.prototype = new svg.Element.RenderedElementBase();
        svg.Element.rect = function(node) {
            this.base = svg.Element.PathElementBase;
            this.base(node);
            this.path = function(ctx) {
                var x = this.attribute("x").toPixels("x");
                var y = this.attribute("y").toPixels("y");
                var width = this.attribute("width").toPixels("x");
                var height = this.attribute("height").toPixels("y");
                var rx = this.attribute("rx").toPixels("x");
                var ry = this.attribute("ry").toPixels("y");
                if (this.attribute("rx").hasValue() && !this.attribute("ry").hasValue()) ry = rx;
                if (this.attribute("ry").hasValue() && !this.attribute("rx").hasValue()) rx = ry;
                rx = Math.min(rx, width / 2);
                ry = Math.min(ry, height / 2);
                if (ctx != null) {
                    ctx.beginPath();
                    ctx.moveTo(x + rx, y);
                    ctx.lineTo(x + width - rx, y);
                    ctx.quadraticCurveTo(x + width, y, x + width, y + ry);
                    ctx.lineTo(x + width, y + height - ry);
                    ctx.quadraticCurveTo(x + width, y + height, x + width - rx, y + height);
                    ctx.lineTo(x + rx, y + height);
                    ctx.quadraticCurveTo(x, y + height, x, y + height - ry);
                    ctx.lineTo(x, y + ry);
                    ctx.quadraticCurveTo(x, y, x + rx, y);
                    ctx.closePath();
                }
                return new svg.BoundingBox(x, y, x + width, y + height);
            };
        };
        svg.Element.rect.prototype = new svg.Element.PathElementBase();
        svg.Element.circle = function(node) {
            this.base = svg.Element.PathElementBase;
            this.base(node);
            this.path = function(ctx) {
                var cx = this.attribute("cx").toPixels("x");
                var cy = this.attribute("cy").toPixels("y");
                var r = this.attribute("r").toPixels();
                if (ctx != null) {
                    ctx.beginPath();
                    ctx.arc(cx, cy, r, 0, Math.PI * 2, true);
                    ctx.closePath();
                }
                return new svg.BoundingBox(cx - r, cy - r, cx + r, cy + r);
            };
        };
        svg.Element.circle.prototype = new svg.Element.PathElementBase();
        svg.Element.ellipse = function(node) {
            this.base = svg.Element.PathElementBase;
            this.base(node);
            this.path = function(ctx) {
                var KAPPA = 4 * ((Math.sqrt(2) - 1) / 3);
                var rx = this.attribute("rx").toPixels("x");
                var ry = this.attribute("ry").toPixels("y");
                var cx = this.attribute("cx").toPixels("x");
                var cy = this.attribute("cy").toPixels("y");
                if (ctx != null) {
                    ctx.beginPath();
                    ctx.moveTo(cx, cy - ry);
                    ctx.bezierCurveTo(cx + KAPPA * rx, cy - ry, cx + rx, cy - KAPPA * ry, cx + rx, cy);
                    ctx.bezierCurveTo(cx + rx, cy + KAPPA * ry, cx + KAPPA * rx, cy + ry, cx, cy + ry);
                    ctx.bezierCurveTo(cx - KAPPA * rx, cy + ry, cx - rx, cy + KAPPA * ry, cx - rx, cy);
                    ctx.bezierCurveTo(cx - rx, cy - KAPPA * ry, cx - KAPPA * rx, cy - ry, cx, cy - ry);
                    ctx.closePath();
                }
                return new svg.BoundingBox(cx - rx, cy - ry, cx + rx, cy + ry);
            };
        };
        svg.Element.ellipse.prototype = new svg.Element.PathElementBase();
        svg.Element.line = function(node) {
            this.base = svg.Element.PathElementBase;
            this.base(node);
            this.getPoints = function() {
                return [ new svg.Point(this.attribute("x1").toPixels("x"), this.attribute("y1").toPixels("y")), new svg.Point(this.attribute("x2").toPixels("x"), this.attribute("y2").toPixels("y")) ];
            };
            this.path = function(ctx) {
                var points = this.getPoints();
                if (ctx != null) {
                    ctx.beginPath();
                    ctx.moveTo(points[0].x, points[0].y);
                    ctx.lineTo(points[1].x, points[1].y);
                }
                return new svg.BoundingBox(points[0].x, points[0].y, points[1].x, points[1].y);
            };
            this.getMarkers = function() {
                var points = this.getPoints();
                var a = points[0].angleTo(points[1]);
                return [ [ points[0], a ], [ points[1], a ] ];
            };
        };
        svg.Element.line.prototype = new svg.Element.PathElementBase();
        svg.Element.polyline = function(node) {
            this.base = svg.Element.PathElementBase;
            this.base(node);
            this.points = svg.CreatePath(this.attribute("points").value);
            this.path = function(ctx) {
                var bb = new svg.BoundingBox(this.points[0].x, this.points[0].y);
                if (ctx != null) {
                    ctx.beginPath();
                    ctx.moveTo(this.points[0].x, this.points[0].y);
                }
                for (var i = 1; i < this.points.length; i++) {
                    bb.addPoint(this.points[i].x, this.points[i].y);
                    if (ctx != null) ctx.lineTo(this.points[i].x, this.points[i].y);
                }
                return bb;
            };
            this.getMarkers = function() {
                var markers = [];
                for (var i = 0; i < this.points.length - 1; i++) {
                    markers.push([ this.points[i], this.points[i].angleTo(this.points[i + 1]) ]);
                }
                markers.push([ this.points[this.points.length - 1], markers[markers.length - 1][1] ]);
                return markers;
            };
        };
        svg.Element.polyline.prototype = new svg.Element.PathElementBase();
        svg.Element.polygon = function(node) {
            this.base = svg.Element.polyline;
            this.base(node);
            this.basePath = this.path;
            this.path = function(ctx) {
                var bb = this.basePath(ctx);
                if (ctx != null) {
                    ctx.lineTo(this.points[0].x, this.points[0].y);
                    ctx.closePath();
                }
                return bb;
            };
        };
        svg.Element.polygon.prototype = new svg.Element.polyline();
        svg.Element.path = function(node) {
            this.base = svg.Element.PathElementBase;
            this.base(node);
            var d = this.attribute("d").value;
            d = d.replace(/,/gm, " ");
            for (var i = 0; i < 2; i++) d = d.replace(/([MmZzLlHhVvCcSsQqTtAa])([^\s])/gm, "$1 $2");
            d = d.replace(/([^\s])([MmZzLlHhVvCcSsQqTtAa])/gm, "$1 $2");
            d = d.replace(/([0-9])([+\-])/gm, "$1 $2");
            for (var i = 0; i < 2; i++) d = d.replace(/(\.[0-9]*)(\.)/gm, "$1 $2");
            d = d.replace(/([Aa](\s+[0-9]+){3})\s+([01])\s*([01])/gm, "$1 $3 $4 ");
            d = svg.compressSpaces(d);
            d = svg.trim(d);
            this.PathParser = new function(d) {
                this.tokens = d.split(" ");
                this.reset = function() {
                    this.i = -1;
                    this.command = "";
                    this.previousCommand = "";
                    this.start = new svg.Point(0, 0);
                    this.control = new svg.Point(0, 0);
                    this.current = new svg.Point(0, 0);
                    this.points = [];
                    this.angles = [];
                };
                this.isEnd = function() {
                    return this.i >= this.tokens.length - 1;
                };
                this.isCommandOrEnd = function() {
                    if (this.isEnd()) return true;
                    return this.tokens[this.i + 1].match(/^[A-Za-z]$/) != null;
                };
                this.isRelativeCommand = function() {
                    switch (this.command) {
                      case "m":
                      case "l":
                      case "h":
                      case "v":
                      case "c":
                      case "s":
                      case "q":
                      case "t":
                      case "a":
                      case "z":
                        return true;
                        break;
                    }
                    return false;
                };
                this.getToken = function() {
                    this.i++;
                    return this.tokens[this.i];
                };
                this.getScalar = function() {
                    return parseFloat(this.getToken());
                };
                this.nextCommand = function() {
                    this.previousCommand = this.command;
                    this.command = this.getToken();
                };
                this.getPoint = function() {
                    var p = new svg.Point(this.getScalar(), this.getScalar());
                    return this.makeAbsolute(p);
                };
                this.getAsControlPoint = function() {
                    var p = this.getPoint();
                    this.control = p;
                    return p;
                };
                this.getAsCurrentPoint = function() {
                    var p = this.getPoint();
                    this.current = p;
                    return p;
                };
                this.getReflectedControlPoint = function() {
                    if (this.previousCommand.toLowerCase() != "c" && this.previousCommand.toLowerCase() != "s" && this.previousCommand.toLowerCase() != "q" && this.previousCommand.toLowerCase() != "t") {
                        return this.current;
                    }
                    var p = new svg.Point(2 * this.current.x - this.control.x, 2 * this.current.y - this.control.y);
                    return p;
                };
                this.makeAbsolute = function(p) {
                    if (this.isRelativeCommand()) {
                        p.x += this.current.x;
                        p.y += this.current.y;
                    }
                    return p;
                };
                this.addMarker = function(p, from, priorTo) {
                    if (priorTo != null && this.angles.length > 0 && this.angles[this.angles.length - 1] == null) {
                        this.angles[this.angles.length - 1] = this.points[this.points.length - 1].angleTo(priorTo);
                    }
                    this.addMarkerAngle(p, from == null ? null : from.angleTo(p));
                };
                this.addMarkerAngle = function(p, a) {
                    this.points.push(p);
                    this.angles.push(a);
                };
                this.getMarkerPoints = function() {
                    return this.points;
                };
                this.getMarkerAngles = function() {
                    for (var i = 0; i < this.angles.length; i++) {
                        if (this.angles[i] == null) {
                            for (var j = i + 1; j < this.angles.length; j++) {
                                if (this.angles[j] != null) {
                                    this.angles[i] = this.angles[j];
                                    break;
                                }
                            }
                        }
                    }
                    return this.angles;
                };
            }(d);
            this.path = function(ctx) {
                var pp = this.PathParser;
                pp.reset();
                var bb = new svg.BoundingBox();
                if (ctx != null) ctx.beginPath();
                while (!pp.isEnd()) {
                    pp.nextCommand();
                    switch (pp.command) {
                      case "M":
                      case "m":
                        var p = pp.getAsCurrentPoint();
                        pp.addMarker(p);
                        bb.addPoint(p.x, p.y);
                        if (ctx != null) ctx.moveTo(p.x, p.y);
                        pp.start = pp.current;
                        while (!pp.isCommandOrEnd()) {
                            var p = pp.getAsCurrentPoint();
                            pp.addMarker(p, pp.start);
                            bb.addPoint(p.x, p.y);
                            if (ctx != null) ctx.lineTo(p.x, p.y);
                        }
                        break;

                      case "L":
                      case "l":
                        while (!pp.isCommandOrEnd()) {
                            var c = pp.current;
                            var p = pp.getAsCurrentPoint();
                            pp.addMarker(p, c);
                            bb.addPoint(p.x, p.y);
                            if (ctx != null) ctx.lineTo(p.x, p.y);
                        }
                        break;

                      case "H":
                      case "h":
                        while (!pp.isCommandOrEnd()) {
                            var newP = new svg.Point((pp.isRelativeCommand() ? pp.current.x : 0) + pp.getScalar(), pp.current.y);
                            pp.addMarker(newP, pp.current);
                            pp.current = newP;
                            bb.addPoint(pp.current.x, pp.current.y);
                            if (ctx != null) ctx.lineTo(pp.current.x, pp.current.y);
                        }
                        break;

                      case "V":
                      case "v":
                        while (!pp.isCommandOrEnd()) {
                            var newP = new svg.Point(pp.current.x, (pp.isRelativeCommand() ? pp.current.y : 0) + pp.getScalar());
                            pp.addMarker(newP, pp.current);
                            pp.current = newP;
                            bb.addPoint(pp.current.x, pp.current.y);
                            if (ctx != null) ctx.lineTo(pp.current.x, pp.current.y);
                        }
                        break;

                      case "C":
                      case "c":
                        while (!pp.isCommandOrEnd()) {
                            var curr = pp.current;
                            var p1 = pp.getPoint();
                            var cntrl = pp.getAsControlPoint();
                            var cp = pp.getAsCurrentPoint();
                            pp.addMarker(cp, cntrl, p1);
                            bb.addBezierCurve(curr.x, curr.y, p1.x, p1.y, cntrl.x, cntrl.y, cp.x, cp.y);
                            if (ctx != null) ctx.bezierCurveTo(p1.x, p1.y, cntrl.x, cntrl.y, cp.x, cp.y);
                        }
                        break;

                      case "S":
                      case "s":
                        while (!pp.isCommandOrEnd()) {
                            var curr = pp.current;
                            var p1 = pp.getReflectedControlPoint();
                            var cntrl = pp.getAsControlPoint();
                            var cp = pp.getAsCurrentPoint();
                            pp.addMarker(cp, cntrl, p1);
                            bb.addBezierCurve(curr.x, curr.y, p1.x, p1.y, cntrl.x, cntrl.y, cp.x, cp.y);
                            if (ctx != null) ctx.bezierCurveTo(p1.x, p1.y, cntrl.x, cntrl.y, cp.x, cp.y);
                        }
                        break;

                      case "Q":
                      case "q":
                        while (!pp.isCommandOrEnd()) {
                            var curr = pp.current;
                            var cntrl = pp.getAsControlPoint();
                            var cp = pp.getAsCurrentPoint();
                            pp.addMarker(cp, cntrl, cntrl);
                            bb.addQuadraticCurve(curr.x, curr.y, cntrl.x, cntrl.y, cp.x, cp.y);
                            if (ctx != null) ctx.quadraticCurveTo(cntrl.x, cntrl.y, cp.x, cp.y);
                        }
                        break;

                      case "T":
                      case "t":
                        while (!pp.isCommandOrEnd()) {
                            var curr = pp.current;
                            var cntrl = pp.getReflectedControlPoint();
                            pp.control = cntrl;
                            var cp = pp.getAsCurrentPoint();
                            pp.addMarker(cp, cntrl, cntrl);
                            bb.addQuadraticCurve(curr.x, curr.y, cntrl.x, cntrl.y, cp.x, cp.y);
                            if (ctx != null) ctx.quadraticCurveTo(cntrl.x, cntrl.y, cp.x, cp.y);
                        }
                        break;

                      case "A":
                      case "a":
                        while (!pp.isCommandOrEnd()) {
                            var curr = pp.current;
                            var rx = pp.getScalar();
                            var ry = pp.getScalar();
                            var xAxisRotation = pp.getScalar() * (Math.PI / 180);
                            var largeArcFlag = pp.getScalar();
                            var sweepFlag = pp.getScalar();
                            var cp = pp.getAsCurrentPoint();
                            var currp = new svg.Point(Math.cos(xAxisRotation) * (curr.x - cp.x) / 2 + Math.sin(xAxisRotation) * (curr.y - cp.y) / 2, -Math.sin(xAxisRotation) * (curr.x - cp.x) / 2 + Math.cos(xAxisRotation) * (curr.y - cp.y) / 2);
                            var l = Math.pow(currp.x, 2) / Math.pow(rx, 2) + Math.pow(currp.y, 2) / Math.pow(ry, 2);
                            if (l > 1) {
                                rx *= Math.sqrt(l);
                                ry *= Math.sqrt(l);
                            }
                            var s = (largeArcFlag == sweepFlag ? -1 : 1) * Math.sqrt((Math.pow(rx, 2) * Math.pow(ry, 2) - Math.pow(rx, 2) * Math.pow(currp.y, 2) - Math.pow(ry, 2) * Math.pow(currp.x, 2)) / (Math.pow(rx, 2) * Math.pow(currp.y, 2) + Math.pow(ry, 2) * Math.pow(currp.x, 2)));
                            if (isNaN(s)) s = 0;
                            var cpp = new svg.Point(s * rx * currp.y / ry, s * -ry * currp.x / rx);
                            var centp = new svg.Point((curr.x + cp.x) / 2 + Math.cos(xAxisRotation) * cpp.x - Math.sin(xAxisRotation) * cpp.y, (curr.y + cp.y) / 2 + Math.sin(xAxisRotation) * cpp.x + Math.cos(xAxisRotation) * cpp.y);
                            var m = function(v) {
                                return Math.sqrt(Math.pow(v[0], 2) + Math.pow(v[1], 2));
                            };
                            var r = function(u, v) {
                                return (u[0] * v[0] + u[1] * v[1]) / (m(u) * m(v));
                            };
                            var a = function(u, v) {
                                return (u[0] * v[1] < u[1] * v[0] ? -1 : 1) * Math.acos(r(u, v));
                            };
                            var a1 = a([ 1, 0 ], [ (currp.x - cpp.x) / rx, (currp.y - cpp.y) / ry ]);
                            var u = [ (currp.x - cpp.x) / rx, (currp.y - cpp.y) / ry ];
                            var v = [ (-currp.x - cpp.x) / rx, (-currp.y - cpp.y) / ry ];
                            var ad = a(u, v);
                            if (r(u, v) <= -1) ad = Math.PI;
                            if (r(u, v) >= 1) ad = 0;
                            var dir = 1 - sweepFlag ? 1 : -1;
                            var ah = a1 + dir * (ad / 2);
                            var halfWay = new svg.Point(centp.x + rx * Math.cos(ah), centp.y + ry * Math.sin(ah));
                            pp.addMarkerAngle(halfWay, ah - dir * Math.PI / 2);
                            pp.addMarkerAngle(cp, ah - dir * Math.PI);
                            bb.addPoint(cp.x, cp.y);
                            if (ctx != null) {
                                var r = rx > ry ? rx : ry;
                                var sx = rx > ry ? 1 : rx / ry;
                                var sy = rx > ry ? ry / rx : 1;
                                ctx.translate(centp.x, centp.y);
                                ctx.rotate(xAxisRotation);
                                ctx.scale(sx, sy);
                                ctx.arc(0, 0, r, a1, a1 + ad, 1 - sweepFlag);
                                ctx.scale(1 / sx, 1 / sy);
                                ctx.rotate(-xAxisRotation);
                                ctx.translate(-centp.x, -centp.y);
                            }
                        }
                        break;

                      case "Z":
                      case "z":
                        if (ctx != null) ctx.closePath();
                        pp.current = pp.start;
                    }
                }
                return bb;
            };
            this.getMarkers = function() {
                var points = this.PathParser.getMarkerPoints();
                var angles = this.PathParser.getMarkerAngles();
                var markers = [];
                for (var i = 0; i < points.length; i++) {
                    markers.push([ points[i], angles[i] ]);
                }
                return markers;
            };
        };
        svg.Element.path.prototype = new svg.Element.PathElementBase();
        svg.Element.pattern = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.createPattern = function(ctx, element) {
                var width = this.attribute("width").toPixels("x", true);
                var height = this.attribute("height").toPixels("y", true);
                var tempSvg = new svg.Element.svg();
                tempSvg.attributes["viewBox"] = new svg.Property("viewBox", this.attribute("viewBox").value);
                tempSvg.attributes["width"] = new svg.Property("width", width + "px");
                tempSvg.attributes["height"] = new svg.Property("height", height + "px");
                tempSvg.attributes["transform"] = new svg.Property("transform", this.attribute("patternTransform").value);
                tempSvg.children = this.children;
                var c = document.createElement("canvas");
                c.width = width;
                c.height = height;
                var cctx = c.getContext("2d");
                if (this.attribute("x").hasValue() && this.attribute("y").hasValue()) {
                    cctx.translate(this.attribute("x").toPixels("x", true), this.attribute("y").toPixels("y", true));
                }
                for (var x = -1; x <= 1; x++) {
                    for (var y = -1; y <= 1; y++) {
                        cctx.save();
                        tempSvg.attributes["x"] = new svg.Property("x", x * c.width);
                        tempSvg.attributes["y"] = new svg.Property("y", y * c.height);
                        tempSvg.render(cctx);
                        cctx.restore();
                    }
                }
                var pattern = ctx.createPattern(c, "repeat");
                return pattern;
            };
        };
        svg.Element.pattern.prototype = new svg.Element.ElementBase();
        svg.Element.marker = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.baseRender = this.render;
            this.render = function(ctx, point, angle) {
                ctx.translate(point.x, point.y);
                if (this.attribute("orient").valueOrDefault("auto") == "auto") ctx.rotate(angle);
                if (this.attribute("markerUnits").valueOrDefault("strokeWidth") == "strokeWidth") ctx.scale(ctx.lineWidth, ctx.lineWidth);
                ctx.save();
                var tempSvg = new svg.Element.svg();
                tempSvg.attributes["viewBox"] = new svg.Property("viewBox", this.attribute("viewBox").value);
                tempSvg.attributes["refX"] = new svg.Property("refX", this.attribute("refX").value);
                tempSvg.attributes["refY"] = new svg.Property("refY", this.attribute("refY").value);
                tempSvg.attributes["width"] = new svg.Property("width", this.attribute("markerWidth").value);
                tempSvg.attributes["height"] = new svg.Property("height", this.attribute("markerHeight").value);
                tempSvg.attributes["fill"] = new svg.Property("fill", this.attribute("fill").valueOrDefault("black"));
                tempSvg.attributes["stroke"] = new svg.Property("stroke", this.attribute("stroke").valueOrDefault("none"));
                tempSvg.children = this.children;
                tempSvg.render(ctx);
                ctx.restore();
                if (this.attribute("markerUnits").valueOrDefault("strokeWidth") == "strokeWidth") ctx.scale(1 / ctx.lineWidth, 1 / ctx.lineWidth);
                if (this.attribute("orient").valueOrDefault("auto") == "auto") ctx.rotate(-angle);
                ctx.translate(-point.x, -point.y);
            };
        };
        svg.Element.marker.prototype = new svg.Element.ElementBase();
        svg.Element.defs = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.render = function(ctx) {};
        };
        svg.Element.defs.prototype = new svg.Element.ElementBase();
        svg.Element.GradientBase = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.stops = [];
            for (var i = 0; i < this.children.length; i++) {
                var child = this.children[i];
                if (child.type == "stop") this.stops.push(child);
            }
            this.getGradient = function() {};
            this.gradientUnits = function() {
                return this.attribute("gradientUnits").valueOrDefault("objectBoundingBox");
            };
            this.attributesToInherit = [ "gradientUnits" ];
            this.inheritStopContainer = function(stopsContainer) {
                for (var i = 0; i < this.attributesToInherit.length; i++) {
                    var attributeToInherit = this.attributesToInherit[i];
                    if (!this.attribute(attributeToInherit).hasValue() && stopsContainer.attribute(attributeToInherit).hasValue()) {
                        this.attribute(attributeToInherit, true).value = stopsContainer.attribute(attributeToInherit).value;
                    }
                }
            };
            this.createGradient = function(ctx, element, parentOpacityProp) {
                var stopsContainer = this;
                if (this.getHrefAttribute().hasValue()) {
                    stopsContainer = this.getHrefAttribute().getDefinition();
                    this.inheritStopContainer(stopsContainer);
                }
                var addParentOpacity = function(color) {
                    if (parentOpacityProp.hasValue()) {
                        var p = new svg.Property("color", color);
                        return p.addOpacity(parentOpacityProp).value;
                    }
                    return color;
                };
                var g = this.getGradient(ctx, element);
                if (g == null) return addParentOpacity(stopsContainer.stops[stopsContainer.stops.length - 1].color);
                for (var i = 0; i < stopsContainer.stops.length; i++) {
                    g.addColorStop(stopsContainer.stops[i].offset, addParentOpacity(stopsContainer.stops[i].color));
                }
                if (this.attribute("gradientTransform").hasValue()) {
                    var rootView = svg.ViewPort.viewPorts[0];
                    var rect = new svg.Element.rect();
                    rect.attributes["x"] = new svg.Property("x", -svg.MAX_VIRTUAL_PIXELS / 3);
                    rect.attributes["y"] = new svg.Property("y", -svg.MAX_VIRTUAL_PIXELS / 3);
                    rect.attributes["width"] = new svg.Property("width", svg.MAX_VIRTUAL_PIXELS);
                    rect.attributes["height"] = new svg.Property("height", svg.MAX_VIRTUAL_PIXELS);
                    var group = new svg.Element.g();
                    group.attributes["transform"] = new svg.Property("transform", this.attribute("gradientTransform").value);
                    group.children = [ rect ];
                    var tempSvg = new svg.Element.svg();
                    tempSvg.attributes["x"] = new svg.Property("x", 0);
                    tempSvg.attributes["y"] = new svg.Property("y", 0);
                    tempSvg.attributes["width"] = new svg.Property("width", rootView.width);
                    tempSvg.attributes["height"] = new svg.Property("height", rootView.height);
                    tempSvg.children = [ group ];
                    var c = document.createElement("canvas");
                    c.width = rootView.width;
                    c.height = rootView.height;
                    var tempCtx = c.getContext("2d");
                    tempCtx.fillStyle = g;
                    tempSvg.render(tempCtx);
                    return tempCtx.createPattern(c, "no-repeat");
                }
                return g;
            };
        };
        svg.Element.GradientBase.prototype = new svg.Element.ElementBase();
        svg.Element.linearGradient = function(node) {
            this.base = svg.Element.GradientBase;
            this.base(node);
            this.attributesToInherit.push("x1");
            this.attributesToInherit.push("y1");
            this.attributesToInherit.push("x2");
            this.attributesToInherit.push("y2");
            this.getGradient = function(ctx, element) {
                var bb = this.gradientUnits() == "objectBoundingBox" ? element.getBoundingBox() : null;
                if (!this.attribute("x1").hasValue() && !this.attribute("y1").hasValue() && !this.attribute("x2").hasValue() && !this.attribute("y2").hasValue()) {
                    this.attribute("x1", true).value = 0;
                    this.attribute("y1", true).value = 0;
                    this.attribute("x2", true).value = 1;
                    this.attribute("y2", true).value = 0;
                }
                var x1 = this.gradientUnits() == "objectBoundingBox" ? bb.x() + bb.width() * this.attribute("x1").numValue() : this.attribute("x1").toPixels("x");
                var y1 = this.gradientUnits() == "objectBoundingBox" ? bb.y() + bb.height() * this.attribute("y1").numValue() : this.attribute("y1").toPixels("y");
                var x2 = this.gradientUnits() == "objectBoundingBox" ? bb.x() + bb.width() * this.attribute("x2").numValue() : this.attribute("x2").toPixels("x");
                var y2 = this.gradientUnits() == "objectBoundingBox" ? bb.y() + bb.height() * this.attribute("y2").numValue() : this.attribute("y2").toPixels("y");
                if (x1 == x2 && y1 == y2) return null;
                return ctx.createLinearGradient(x1, y1, x2, y2);
            };
        };
        svg.Element.linearGradient.prototype = new svg.Element.GradientBase();
        svg.Element.radialGradient = function(node) {
            this.base = svg.Element.GradientBase;
            this.base(node);
            this.attributesToInherit.push("cx");
            this.attributesToInherit.push("cy");
            this.attributesToInherit.push("r");
            this.attributesToInherit.push("fx");
            this.attributesToInherit.push("fy");
            this.getGradient = function(ctx, element) {
                var bb = element.getBoundingBox();
                if (!this.attribute("cx").hasValue()) this.attribute("cx", true).value = "50%";
                if (!this.attribute("cy").hasValue()) this.attribute("cy", true).value = "50%";
                if (!this.attribute("r").hasValue()) this.attribute("r", true).value = "50%";
                var cx = this.gradientUnits() == "objectBoundingBox" ? bb.x() + bb.width() * this.attribute("cx").numValue() : this.attribute("cx").toPixels("x");
                var cy = this.gradientUnits() == "objectBoundingBox" ? bb.y() + bb.height() * this.attribute("cy").numValue() : this.attribute("cy").toPixels("y");
                var fx = cx;
                var fy = cy;
                if (this.attribute("fx").hasValue()) {
                    fx = this.gradientUnits() == "objectBoundingBox" ? bb.x() + bb.width() * this.attribute("fx").numValue() : this.attribute("fx").toPixels("x");
                }
                if (this.attribute("fy").hasValue()) {
                    fy = this.gradientUnits() == "objectBoundingBox" ? bb.y() + bb.height() * this.attribute("fy").numValue() : this.attribute("fy").toPixels("y");
                }
                var r = this.gradientUnits() == "objectBoundingBox" ? (bb.width() + bb.height()) / 2 * this.attribute("r").numValue() : this.attribute("r").toPixels();
                return ctx.createRadialGradient(fx, fy, 0, cx, cy, r);
            };
        };
        svg.Element.radialGradient.prototype = new svg.Element.GradientBase();
        svg.Element.stop = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.offset = this.attribute("offset").numValue();
            if (this.offset < 0) this.offset = 0;
            if (this.offset > 1) this.offset = 1;
            var stopColor = this.style("stop-color", true);
            if (stopColor.value === "") stopColor.value = "#000";
            if (this.style("stop-opacity").hasValue()) stopColor = stopColor.addOpacity(this.style("stop-opacity"));
            this.color = stopColor.value;
        };
        svg.Element.stop.prototype = new svg.Element.ElementBase();
        svg.Element.AnimateBase = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            svg.Animations.push(this);
            this.duration = 0;
            this.begin = this.attribute("begin").toMilliseconds();
            this.maxDuration = this.begin + this.attribute("dur").toMilliseconds();
            this.getProperty = function() {
                var attributeType = this.attribute("attributeType").value;
                var attributeName = this.attribute("attributeName").value;
                if (attributeType == "CSS") {
                    return this.parent.style(attributeName, true);
                }
                return this.parent.attribute(attributeName, true);
            };
            this.initialValue = null;
            this.initialUnits = "";
            this.removed = false;
            this.calcValue = function() {
                return "";
            };
            this.update = function(delta) {
                if (this.initialValue == null) {
                    this.initialValue = this.getProperty().value;
                    this.initialUnits = this.getProperty().getUnits();
                }
                if (this.duration > this.maxDuration) {
                    if (this.attribute("repeatCount").value == "indefinite" || this.attribute("repeatDur").value == "indefinite") {
                        this.duration = 0;
                    } else if (this.attribute("fill").valueOrDefault("remove") == "freeze" && !this.frozen) {
                        this.frozen = true;
                        this.parent.animationFrozen = true;
                        this.parent.animationFrozenValue = this.getProperty().value;
                    } else if (this.attribute("fill").valueOrDefault("remove") == "remove" && !this.removed) {
                        this.removed = true;
                        this.getProperty().value = this.parent.animationFrozen ? this.parent.animationFrozenValue : this.initialValue;
                        return true;
                    }
                    return false;
                }
                this.duration = this.duration + delta;
                var updated = false;
                if (this.begin < this.duration) {
                    var newValue = this.calcValue();
                    if (this.attribute("type").hasValue()) {
                        var type = this.attribute("type").value;
                        newValue = type + "(" + newValue + ")";
                    }
                    this.getProperty().value = newValue;
                    updated = true;
                }
                return updated;
            };
            this.from = this.attribute("from");
            this.to = this.attribute("to");
            this.values = this.attribute("values");
            if (this.values.hasValue()) this.values.value = this.values.value.split(";");
            this.progress = function() {
                var ret = {
                    progress: (this.duration - this.begin) / (this.maxDuration - this.begin)
                };
                if (this.values.hasValue()) {
                    var p = ret.progress * (this.values.value.length - 1);
                    var lb = Math.floor(p), ub = Math.ceil(p);
                    ret.from = new svg.Property("from", parseFloat(this.values.value[lb]));
                    ret.to = new svg.Property("to", parseFloat(this.values.value[ub]));
                    ret.progress = (p - lb) / (ub - lb);
                } else {
                    ret.from = this.from;
                    ret.to = this.to;
                }
                return ret;
            };
        };
        svg.Element.AnimateBase.prototype = new svg.Element.ElementBase();
        svg.Element.animate = function(node) {
            this.base = svg.Element.AnimateBase;
            this.base(node);
            this.calcValue = function() {
                var p = this.progress();
                var newValue = p.from.numValue() + (p.to.numValue() - p.from.numValue()) * p.progress;
                return newValue + this.initialUnits;
            };
        };
        svg.Element.animate.prototype = new svg.Element.AnimateBase();
        svg.Element.animateColor = function(node) {
            this.base = svg.Element.AnimateBase;
            this.base(node);
            this.calcValue = function() {
                var p = this.progress();
                var from = new RGBColor(p.from.value);
                var to = new RGBColor(p.to.value);
                if (from.ok && to.ok) {
                    var r = from.r + (to.r - from.r) * p.progress;
                    var g = from.g + (to.g - from.g) * p.progress;
                    var b = from.b + (to.b - from.b) * p.progress;
                    return "rgb(" + parseInt(r, 10) + "," + parseInt(g, 10) + "," + parseInt(b, 10) + ")";
                }
                return this.attribute("from").value;
            };
        };
        svg.Element.animateColor.prototype = new svg.Element.AnimateBase();
        svg.Element.animateTransform = function(node) {
            this.base = svg.Element.AnimateBase;
            this.base(node);
            this.calcValue = function() {
                var p = this.progress();
                var from = svg.ToNumberArray(p.from.value);
                var to = svg.ToNumberArray(p.to.value);
                var newValue = "";
                for (var i = 0; i < from.length; i++) {
                    newValue += from[i] + (to[i] - from[i]) * p.progress + " ";
                }
                return newValue;
            };
        };
        svg.Element.animateTransform.prototype = new svg.Element.animate();
        svg.Element.font = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.horizAdvX = this.attribute("horiz-adv-x").numValue();
            this.isRTL = false;
            this.isArabic = false;
            this.fontFace = null;
            this.missingGlyph = null;
            this.glyphs = [];
            for (var i = 0; i < this.children.length; i++) {
                var child = this.children[i];
                if (child.type == "font-face") {
                    this.fontFace = child;
                    if (child.style("font-family").hasValue()) {
                        svg.Definitions[child.style("font-family").value] = this;
                    }
                } else if (child.type == "missing-glyph") this.missingGlyph = child; else if (child.type == "glyph") {
                    if (child.arabicForm != "") {
                        this.isRTL = true;
                        this.isArabic = true;
                        if (typeof this.glyphs[child.unicode] == "undefined") this.glyphs[child.unicode] = [];
                        this.glyphs[child.unicode][child.arabicForm] = child;
                    } else {
                        this.glyphs[child.unicode] = child;
                    }
                }
            }
        };
        svg.Element.font.prototype = new svg.Element.ElementBase();
        svg.Element.fontface = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.ascent = this.attribute("ascent").value;
            this.descent = this.attribute("descent").value;
            this.unitsPerEm = this.attribute("units-per-em").numValue();
        };
        svg.Element.fontface.prototype = new svg.Element.ElementBase();
        svg.Element.missingglyph = function(node) {
            this.base = svg.Element.path;
            this.base(node);
            this.horizAdvX = 0;
        };
        svg.Element.missingglyph.prototype = new svg.Element.path();
        svg.Element.glyph = function(node) {
            this.base = svg.Element.path;
            this.base(node);
            this.horizAdvX = this.attribute("horiz-adv-x").numValue();
            this.unicode = this.attribute("unicode").value;
            this.arabicForm = this.attribute("arabic-form").value;
        };
        svg.Element.glyph.prototype = new svg.Element.path();
        svg.Element.text = function(node) {
            this.captureTextNodes = true;
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.baseSetContext = this.setContext;
            this.setContext = function(ctx) {
                this.baseSetContext(ctx);
                var textBaseline = this.style("dominant-baseline").toTextBaseline();
                if (textBaseline == null) textBaseline = this.style("alignment-baseline").toTextBaseline();
                if (textBaseline != null) ctx.textBaseline = textBaseline;
            };
            this.getBoundingBox = function() {
                var x = this.attribute("x").toPixels("x");
                var y = this.attribute("y").toPixels("y");
                var fontSize = this.parent.style("font-size").numValueOrDefault(svg.Font.Parse(svg.ctx.font).fontSize);
                return new svg.BoundingBox(x, y - fontSize, x + Math.floor(fontSize * 2 / 3) * this.children[0].getText().length, y);
            };
            this.renderChildren = function(ctx) {
                this.x = this.attribute("x").toPixels("x");
                this.y = this.attribute("y").toPixels("y");
                if (this.attribute("dx").hasValue()) this.x += this.attribute("dx").toPixels("x");
                if (this.attribute("dy").hasValue()) this.y += this.attribute("dy").toPixels("y");
                this.x += this.getAnchorDelta(ctx, this, 0);
                for (var i = 0; i < this.children.length; i++) {
                    this.renderChild(ctx, this, i);
                }
            };
            this.getAnchorDelta = function(ctx, parent, startI) {
                var textAnchor = this.style("text-anchor").valueOrDefault("start");
                if (textAnchor != "start") {
                    var width = 0;
                    for (var i = startI; i < parent.children.length; i++) {
                        var child = parent.children[i];
                        if (i > startI && child.attribute("x").hasValue()) break;
                        width += child.measureTextRecursive(ctx);
                    }
                    return -1 * (textAnchor == "end" ? width : width / 2);
                }
                return 0;
            };
            this.renderChild = function(ctx, parent, i) {
                var child = parent.children[i];
                if (child.attribute("x").hasValue()) {
                    child.x = child.attribute("x").toPixels("x") + parent.getAnchorDelta(ctx, parent, i);
                    if (child.attribute("dx").hasValue()) child.x += child.attribute("dx").toPixels("x");
                } else {
                    if (child.attribute("dx").hasValue()) parent.x += child.attribute("dx").toPixels("x");
                    child.x = parent.x;
                }
                parent.x = child.x + child.measureText(ctx);
                if (child.attribute("y").hasValue()) {
                    child.y = child.attribute("y").toPixels("y");
                    if (child.attribute("dy").hasValue()) child.y += child.attribute("dy").toPixels("y");
                } else {
                    if (child.attribute("dy").hasValue()) parent.y += child.attribute("dy").toPixels("y");
                    child.y = parent.y;
                }
                parent.y = child.y;
                child.render(ctx);
                for (var i = 0; i < child.children.length; i++) {
                    parent.renderChild(ctx, child, i);
                }
            };
        };
        svg.Element.text.prototype = new svg.Element.RenderedElementBase();
        svg.Element.TextElementBase = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.getGlyph = function(font, text, i) {
                var c = text[i];
                var glyph = null;
                if (font.isArabic) {
                    var arabicForm = "isolated";
                    if ((i == 0 || text[i - 1] == " ") && i < text.length - 2 && text[i + 1] != " ") arabicForm = "terminal";
                    if (i > 0 && text[i - 1] != " " && i < text.length - 2 && text[i + 1] != " ") arabicForm = "medial";
                    if (i > 0 && text[i - 1] != " " && (i == text.length - 1 || text[i + 1] == " ")) arabicForm = "initial";
                    if (typeof font.glyphs[c] != "undefined") {
                        glyph = font.glyphs[c][arabicForm];
                        if (glyph == null && font.glyphs[c].type == "glyph") glyph = font.glyphs[c];
                    }
                } else {
                    glyph = font.glyphs[c];
                }
                if (glyph == null) glyph = font.missingGlyph;
                return glyph;
            };
            this.renderChildren = function(ctx) {
                var customFont = this.parent.style("font-family").getDefinition();
                if (customFont != null) {
                    var fontSize = this.parent.style("font-size").numValueOrDefault(svg.Font.Parse(svg.ctx.font).fontSize);
                    var fontStyle = this.parent.style("font-style").valueOrDefault(svg.Font.Parse(svg.ctx.font).fontStyle);
                    var text = this.getText();
                    if (customFont.isRTL) text = text.split("").reverse().join("");
                    var dx = svg.ToNumberArray(this.parent.attribute("dx").value);
                    for (var i = 0; i < text.length; i++) {
                        var glyph = this.getGlyph(customFont, text, i);
                        var scale = fontSize / customFont.fontFace.unitsPerEm;
                        ctx.translate(this.x, this.y);
                        ctx.scale(scale, -scale);
                        var lw = ctx.lineWidth;
                        ctx.lineWidth = ctx.lineWidth * customFont.fontFace.unitsPerEm / fontSize;
                        if (fontStyle == "italic") ctx.transform(1, 0, .4, 1, 0, 0);
                        glyph.render(ctx);
                        if (fontStyle == "italic") ctx.transform(1, 0, -.4, 1, 0, 0);
                        ctx.lineWidth = lw;
                        ctx.scale(1 / scale, -1 / scale);
                        ctx.translate(-this.x, -this.y);
                        this.x += fontSize * (glyph.horizAdvX || customFont.horizAdvX) / customFont.fontFace.unitsPerEm;
                        if (typeof dx[i] != "undefined" && !isNaN(dx[i])) {
                            this.x += dx[i];
                        }
                    }
                    return;
                }
                if (ctx.fillStyle != "") ctx.fillText(svg.compressSpaces(this.getText()), this.x, this.y);
                if (ctx.strokeStyle != "") ctx.strokeText(svg.compressSpaces(this.getText()), this.x, this.y);
            };
            this.getText = function() {};
            this.measureTextRecursive = function(ctx) {
                var width = this.measureText(ctx);
                for (var i = 0; i < this.children.length; i++) {
                    width += this.children[i].measureTextRecursive(ctx);
                }
                return width;
            };
            this.measureText = function(ctx) {
                var customFont = this.parent.style("font-family").getDefinition();
                if (customFont != null) {
                    var fontSize = this.parent.style("font-size").numValueOrDefault(svg.Font.Parse(svg.ctx.font).fontSize);
                    var measure = 0;
                    var text = this.getText();
                    if (customFont.isRTL) text = text.split("").reverse().join("");
                    var dx = svg.ToNumberArray(this.parent.attribute("dx").value);
                    for (var i = 0; i < text.length; i++) {
                        var glyph = this.getGlyph(customFont, text, i);
                        measure += (glyph.horizAdvX || customFont.horizAdvX) * fontSize / customFont.fontFace.unitsPerEm;
                        if (typeof dx[i] != "undefined" && !isNaN(dx[i])) {
                            measure += dx[i];
                        }
                    }
                    return measure;
                }
                var textToMeasure = svg.compressSpaces(this.getText());
                if (!ctx.measureText) return textToMeasure.length * 10;
                ctx.save();
                this.setContext(ctx);
                var width = ctx.measureText(textToMeasure).width;
                ctx.restore();
                return width;
            };
        };
        svg.Element.TextElementBase.prototype = new svg.Element.RenderedElementBase();
        svg.Element.tspan = function(node) {
            this.captureTextNodes = true;
            this.base = svg.Element.TextElementBase;
            this.base(node);
            this.text = svg.compressSpaces(node.value || node.text || node.textContent || "");
            this.getText = function() {
                if (this.children.length > 0) {
                    return "";
                }
                return this.text;
            };
        };
        svg.Element.tspan.prototype = new svg.Element.TextElementBase();
        svg.Element.tref = function(node) {
            this.base = svg.Element.TextElementBase;
            this.base(node);
            this.getText = function() {
                var element = this.getHrefAttribute().getDefinition();
                if (element != null) return element.children[0].getText();
            };
        };
        svg.Element.tref.prototype = new svg.Element.TextElementBase();
        svg.Element.a = function(node) {
            this.base = svg.Element.TextElementBase;
            this.base(node);
            this.hasText = node.childNodes.length > 0;
            for (var i = 0; i < node.childNodes.length; i++) {
                if (node.childNodes[i].nodeType != 3) this.hasText = false;
            }
            this.text = this.hasText ? node.childNodes[0].value : "";
            this.getText = function() {
                return this.text;
            };
            this.baseRenderChildren = this.renderChildren;
            this.renderChildren = function(ctx) {
                if (this.hasText) {
                    this.baseRenderChildren(ctx);
                    var fontSize = new svg.Property("fontSize", svg.Font.Parse(svg.ctx.font).fontSize);
                    svg.Mouse.checkBoundingBox(this, new svg.BoundingBox(this.x, this.y - fontSize.toPixels("y"), this.x + this.measureText(ctx), this.y));
                } else if (this.children.length > 0) {
                    var g = new svg.Element.g();
                    g.children = this.children;
                    g.parent = this;
                    g.render(ctx);
                }
            };
            this.onclick = function() {
                window.open(this.getHrefAttribute().value);
            };
            this.onmousemove = function() {
                svg.ctx.canvas.style.cursor = "pointer";
            };
        };
        svg.Element.a.prototype = new svg.Element.TextElementBase();
        svg.Element.image = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            var href = this.getHrefAttribute().value;
            if (href == "") {
                return;
            }
            var isSvg = href.match(/\.svg$/);
            svg.Images.push(this);
            this.loaded = false;
            if (!isSvg) {
                this.img = document.createElement("img");
                if (svg.opts["useCORS"] == true) {
                    this.img.crossOrigin = "Anonymous";
                }
                var self = this;
                this.img.onload = function() {
                    self.loaded = true;
                };
                this.img.onerror = function() {
                    svg.log('ERROR: image "' + href + '" not found');
                    self.loaded = true;
                };
                this.img.src = href;
            } else {
                this.img = svg.ajax(href);
                this.loaded = true;
            }
            this.renderChildren = function(ctx) {
                var x = this.attribute("x").toPixels("x");
                var y = this.attribute("y").toPixels("y");
                var width = this.attribute("width").toPixels("x");
                var height = this.attribute("height").toPixels("y");
                if (width == 0 || height == 0) return;
                ctx.save();
                if (isSvg) {
                    ctx.drawSvg(this.img, x, y, width, height);
                } else {
                    ctx.translate(x, y);
                    svg.AspectRatio(ctx, this.attribute("preserveAspectRatio").value, width, this.img.width, height, this.img.height, 0, 0);
                    ctx.drawImage(this.img, 0, 0);
                }
                ctx.restore();
            };
            this.getBoundingBox = function() {
                var x = this.attribute("x").toPixels("x");
                var y = this.attribute("y").toPixels("y");
                var width = this.attribute("width").toPixels("x");
                var height = this.attribute("height").toPixels("y");
                return new svg.BoundingBox(x, y, x + width, y + height);
            };
        };
        svg.Element.image.prototype = new svg.Element.RenderedElementBase();
        svg.Element.g = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.getBoundingBox = function() {
                var bb = new svg.BoundingBox();
                for (var i = 0; i < this.children.length; i++) {
                    bb.addBoundingBox(this.children[i].getBoundingBox());
                }
                return bb;
            };
        };
        svg.Element.g.prototype = new svg.Element.RenderedElementBase();
        svg.Element.symbol = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.render = function(ctx) {};
        };
        svg.Element.symbol.prototype = new svg.Element.RenderedElementBase();
        svg.Element.style = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            var css = "";
            for (var i = 0; i < node.childNodes.length; i++) {
                css += node.childNodes[i].data;
            }
            css = css.replace(/(\/\*([^*]|[\r\n]|(\*+([^*\/]|[\r\n])))*\*+\/)|(^[\s]*\/\/.*)/gm, "");
            css = svg.compressSpaces(css);
            var cssDefs = css.split("}");
            for (var i = 0; i < cssDefs.length; i++) {
                if (svg.trim(cssDefs[i]) != "") {
                    var cssDef = cssDefs[i].split("{");
                    var cssClasses = cssDef[0].split(",");
                    var cssProps = cssDef[1].split(";");
                    for (var j = 0; j < cssClasses.length; j++) {
                        var cssClass = svg.trim(cssClasses[j]);
                        if (cssClass != "") {
                            var props = {};
                            for (var k = 0; k < cssProps.length; k++) {
                                var prop = cssProps[k].indexOf(":");
                                var name = cssProps[k].substr(0, prop);
                                var value = cssProps[k].substr(prop + 1, cssProps[k].length - prop);
                                if (name != null && value != null) {
                                    props[svg.trim(name)] = new svg.Property(svg.trim(name), svg.trim(value));
                                }
                            }
                            svg.Styles[cssClass] = props;
                            svg.StylesSpecificity[cssClass] = getSelectorSpecificity(cssClass);
                            if (cssClass == "@font-face") {
                                var fontFamily = props["font-family"].value.replace(/"/g, "");
                                var srcs = props["src"].value.split(",");
                                for (var s = 0; s < srcs.length; s++) {
                                    if (srcs[s].indexOf('format("svg")') > 0) {
                                        var urlStart = srcs[s].indexOf("url");
                                        var urlEnd = srcs[s].indexOf(")", urlStart);
                                        var url = srcs[s].substr(urlStart + 5, urlEnd - urlStart - 6);
                                        var doc = svg.parseXml(svg.ajax(url));
                                        var fonts = doc.getElementsByTagName("font");
                                        for (var f = 0; f < fonts.length; f++) {
                                            var font = svg.CreateElement(fonts[f]);
                                            svg.Definitions[fontFamily] = font;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        };
        svg.Element.style.prototype = new svg.Element.ElementBase();
        svg.Element.use = function(node) {
            this.base = svg.Element.RenderedElementBase;
            this.base(node);
            this.baseSetContext = this.setContext;
            this.setContext = function(ctx) {
                this.baseSetContext(ctx);
                if (this.attribute("x").hasValue()) ctx.translate(this.attribute("x").toPixels("x"), 0);
                if (this.attribute("y").hasValue()) ctx.translate(0, this.attribute("y").toPixels("y"));
            };
            var element = this.getHrefAttribute().getDefinition();
            this.path = function(ctx) {
                if (element != null) element.path(ctx);
            };
            this.getBoundingBox = function() {
                if (element != null) return element.getBoundingBox();
            };
            this.renderChildren = function(ctx) {
                if (element != null) {
                    var tempSvg = element;
                    if (element.type == "symbol") {
                        tempSvg = new svg.Element.svg();
                        tempSvg.type = "svg";
                        tempSvg.attributes["viewBox"] = new svg.Property("viewBox", element.attribute("viewBox").value);
                        tempSvg.attributes["preserveAspectRatio"] = new svg.Property("preserveAspectRatio", element.attribute("preserveAspectRatio").value);
                        tempSvg.attributes["overflow"] = new svg.Property("overflow", element.attribute("overflow").value);
                        tempSvg.children = element.children;
                    }
                    if (tempSvg.type == "svg") {
                        if (this.attribute("width").hasValue()) tempSvg.attributes["width"] = new svg.Property("width", this.attribute("width").value);
                        if (this.attribute("height").hasValue()) tempSvg.attributes["height"] = new svg.Property("height", this.attribute("height").value);
                    }
                    var oldParent = tempSvg.parent;
                    tempSvg.parent = null;
                    tempSvg.render(ctx);
                    tempSvg.parent = oldParent;
                }
            };
        };
        svg.Element.use.prototype = new svg.Element.RenderedElementBase();
        svg.Element.mask = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.apply = function(ctx, element) {
                var x = this.attribute("x").toPixels("x");
                var y = this.attribute("y").toPixels("y");
                var width = this.attribute("width").toPixels("x");
                var height = this.attribute("height").toPixels("y");
                if (width == 0 && height == 0) {
                    var bb = new svg.BoundingBox();
                    for (var i = 0; i < this.children.length; i++) {
                        bb.addBoundingBox(this.children[i].getBoundingBox());
                    }
                    var x = Math.floor(bb.x1);
                    var y = Math.floor(bb.y1);
                    var width = Math.floor(bb.width());
                    var height = Math.floor(bb.height());
                }
                var mask = element.attribute("mask").value;
                element.attribute("mask").value = "";
                var cMask = document.createElement("canvas");
                cMask.width = x + width;
                cMask.height = y + height;
                var maskCtx = cMask.getContext("2d");
                this.renderChildren(maskCtx);
                var c = document.createElement("canvas");
                c.width = x + width;
                c.height = y + height;
                var tempCtx = c.getContext("2d");
                element.render(tempCtx);
                tempCtx.globalCompositeOperation = "destination-in";
                tempCtx.fillStyle = maskCtx.createPattern(cMask, "no-repeat");
                tempCtx.fillRect(0, 0, x + width, y + height);
                ctx.fillStyle = tempCtx.createPattern(c, "no-repeat");
                ctx.fillRect(0, 0, x + width, y + height);
                element.attribute("mask").value = mask;
            };
            this.render = function(ctx) {};
        };
        svg.Element.mask.prototype = new svg.Element.ElementBase();
        svg.Element.clipPath = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.apply = function(ctx) {
                var oldBeginPath = CanvasRenderingContext2D.prototype.beginPath;
                CanvasRenderingContext2D.prototype.beginPath = function() {};
                var oldClosePath = CanvasRenderingContext2D.prototype.closePath;
                CanvasRenderingContext2D.prototype.closePath = function() {};
                oldBeginPath.call(ctx);
                for (var i = 0; i < this.children.length; i++) {
                    var child = this.children[i];
                    if (typeof child.path != "undefined") {
                        var transform = null;
                        if (child.style("transform", false, true).hasValue()) {
                            transform = new svg.Transform(child.style("transform", false, true).value);
                            transform.apply(ctx);
                        }
                        child.path(ctx);
                        CanvasRenderingContext2D.prototype.closePath = oldClosePath;
                        if (transform) {
                            transform.unapply(ctx);
                        }
                    }
                }
                oldClosePath.call(ctx);
                ctx.clip();
                CanvasRenderingContext2D.prototype.beginPath = oldBeginPath;
                CanvasRenderingContext2D.prototype.closePath = oldClosePath;
            };
            this.render = function(ctx) {};
        };
        svg.Element.clipPath.prototype = new svg.Element.ElementBase();
        svg.Element.filter = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.apply = function(ctx, element) {
                var bb = element.getBoundingBox();
                var x = Math.floor(bb.x1);
                var y = Math.floor(bb.y1);
                var width = Math.floor(bb.width());
                var height = Math.floor(bb.height());
                var filter = element.style("filter").value;
                element.style("filter").value = "";
                var px = 0, py = 0;
                for (var i = 0; i < this.children.length; i++) {
                    var efd = this.children[i].extraFilterDistance || 0;
                    px = Math.max(px, efd);
                    py = Math.max(py, efd);
                }
                var c = document.createElement("canvas");
                c.width = width + 2 * px;
                c.height = height + 2 * py;
                var tempCtx = c.getContext("2d");
                tempCtx.translate(-x + px, -y + py);
                element.render(tempCtx);
                for (var i = 0; i < this.children.length; i++) {
                    if (typeof this.children[i].apply === "function") {
                        this.children[i].apply(tempCtx, 0, 0, width + 2 * px, height + 2 * py);
                    }
                }
                ctx.drawImage(c, 0, 0, width + 2 * px, height + 2 * py, x - px, y - py, width + 2 * px, height + 2 * py);
                element.style("filter", true).value = filter;
            };
            this.render = function(ctx) {};
        };
        svg.Element.filter.prototype = new svg.Element.ElementBase();
        svg.Element.feMorphology = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.apply = function(ctx, x, y, width, height) {};
        };
        svg.Element.feMorphology.prototype = new svg.Element.ElementBase();
        svg.Element.feComposite = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.apply = function(ctx, x, y, width, height) {};
        };
        svg.Element.feComposite.prototype = new svg.Element.ElementBase();
        svg.Element.feColorMatrix = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            var matrix = svg.ToNumberArray(this.attribute("values").value);
            switch (this.attribute("type").valueOrDefault("matrix")) {
              case "saturate":
                var s = matrix[0];
                matrix = [ .213 + .787 * s, .715 - .715 * s, .072 - .072 * s, 0, 0, .213 - .213 * s, .715 + .285 * s, .072 - .072 * s, 0, 0, .213 - .213 * s, .715 - .715 * s, .072 + .928 * s, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1 ];
                break;

              case "hueRotate":
                var a = matrix[0] * Math.PI / 180;
                var c = function(m1, m2, m3) {
                    return m1 + Math.cos(a) * m2 + Math.sin(a) * m3;
                };
                matrix = [ c(.213, .787, -.213), c(.715, -.715, -.715), c(.072, -.072, .928), 0, 0, c(.213, -.213, .143), c(.715, .285, .14), c(.072, -.072, -.283), 0, 0, c(.213, -.213, -.787), c(.715, -.715, .715), c(.072, .928, .072), 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1 ];
                break;

              case "luminanceToAlpha":
                matrix = [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, .2125, .7154, .0721, 0, 0, 0, 0, 0, 0, 1 ];
                break;
            }
            function imGet(img, x, y, width, height, rgba) {
                return img[y * width * 4 + x * 4 + rgba];
            }
            function imSet(img, x, y, width, height, rgba, val) {
                img[y * width * 4 + x * 4 + rgba] = val;
            }
            function m(i, v) {
                var mi = matrix[i];
                return mi * (mi < 0 ? v - 255 : v);
            }
            this.apply = function(ctx, x, y, width, height) {
                var srcData = ctx.getImageData(0, 0, width, height);
                for (var y = 0; y < height; y++) {
                    for (var x = 0; x < width; x++) {
                        var r = imGet(srcData.data, x, y, width, height, 0);
                        var g = imGet(srcData.data, x, y, width, height, 1);
                        var b = imGet(srcData.data, x, y, width, height, 2);
                        var a = imGet(srcData.data, x, y, width, height, 3);
                        imSet(srcData.data, x, y, width, height, 0, m(0, r) + m(1, g) + m(2, b) + m(3, a) + m(4, 1));
                        imSet(srcData.data, x, y, width, height, 1, m(5, r) + m(6, g) + m(7, b) + m(8, a) + m(9, 1));
                        imSet(srcData.data, x, y, width, height, 2, m(10, r) + m(11, g) + m(12, b) + m(13, a) + m(14, 1));
                        imSet(srcData.data, x, y, width, height, 3, m(15, r) + m(16, g) + m(17, b) + m(18, a) + m(19, 1));
                    }
                }
                ctx.clearRect(0, 0, width, height);
                ctx.putImageData(srcData, 0, 0);
            };
        };
        svg.Element.feColorMatrix.prototype = new svg.Element.ElementBase();
        svg.Element.feGaussianBlur = function(node) {
            this.base = svg.Element.ElementBase;
            this.base(node);
            this.blurRadius = Math.floor(this.attribute("stdDeviation").numValue());
            this.extraFilterDistance = this.blurRadius;
            this.apply = function(ctx, x, y, width, height) {
                if (typeof stackBlurCanvasRGBA == "undefined") {
                    svg.log("ERROR: StackBlur.js must be included for blur to work");
                    return;
                }
                ctx.canvas.id = svg.UniqueId();
                ctx.canvas.style.display = "none";
                document.body.appendChild(ctx.canvas);
                stackBlurCanvasRGBA(ctx.canvas.id, x, y, width, height, this.blurRadius);
                document.body.removeChild(ctx.canvas);
            };
        };
        svg.Element.feGaussianBlur.prototype = new svg.Element.ElementBase();
        svg.Element.title = function(node) {};
        svg.Element.title.prototype = new svg.Element.ElementBase();
        svg.Element.desc = function(node) {};
        svg.Element.desc.prototype = new svg.Element.ElementBase();
        svg.Element.MISSING = function(node) {
            svg.log("ERROR: Element '" + node.nodeName + "' not yet implemented.");
        };
        svg.Element.MISSING.prototype = new svg.Element.ElementBase();
        svg.CreateElement = function(node) {
            var className = node.nodeName.replace(/^[^:]+:/, "");
            className = className.replace(/\-/g, "");
            var e = null;
            if (typeof svg.Element[className] != "undefined") {
                e = new svg.Element[className](node);
            } else {
                e = new svg.Element.MISSING(node);
            }
            e.type = node.nodeName;
            return e;
        };
        svg.load = function(ctx, url) {
            svg.loadXml(ctx, svg.ajax(url));
        };
        svg.loadXml = function(ctx, xml) {
            svg.loadXmlDoc(ctx, svg.parseXml(xml));
        };
        svg.loadXmlDoc = function(ctx, dom) {
            svg.init(ctx);
            var mapXY = function(p) {
                var e = ctx.canvas;
                while (e) {
                    p.x -= e.offsetLeft;
                    p.y -= e.offsetTop;
                    e = e.offsetParent;
                }
                if (window.scrollX) p.x += window.scrollX;
                if (window.scrollY) p.y += window.scrollY;
                return p;
            };
            if (svg.opts["ignoreMouse"] != true) {
                ctx.canvas.onclick = function(e) {
                    var p = mapXY(new svg.Point(e != null ? e.clientX : event.clientX, e != null ? e.clientY : event.clientY));
                    svg.Mouse.onclick(p.x, p.y);
                };
                ctx.canvas.onmousemove = function(e) {
                    var p = mapXY(new svg.Point(e != null ? e.clientX : event.clientX, e != null ? e.clientY : event.clientY));
                    svg.Mouse.onmousemove(p.x, p.y);
                };
            }
            var e = svg.CreateElement(dom.documentElement);
            e.root = true;
            e.addStylesFromStyleDefinition();
            var isFirstRender = true;
            var draw = function() {
                svg.ViewPort.Clear();
                if (ctx.canvas.parentNode) svg.ViewPort.SetCurrent(ctx.canvas.parentNode.clientWidth, ctx.canvas.parentNode.clientHeight);
                if (svg.opts["ignoreDimensions"] != true) {
                    if (e.style("width").hasValue()) {
                        ctx.canvas.width = e.style("width").toPixels("x");
                        ctx.canvas.style.width = ctx.canvas.width + "px";
                    }
                    if (e.style("height").hasValue()) {
                        ctx.canvas.height = e.style("height").toPixels("y");
                        ctx.canvas.style.height = ctx.canvas.height + "px";
                    }
                }
                var cWidth = ctx.canvas.clientWidth || ctx.canvas.width;
                var cHeight = ctx.canvas.clientHeight || ctx.canvas.height;
                if (svg.opts["ignoreDimensions"] == true && e.style("width").hasValue() && e.style("height").hasValue()) {
                    cWidth = e.style("width").toPixels("x");
                    cHeight = e.style("height").toPixels("y");
                }
                svg.ViewPort.SetCurrent(cWidth, cHeight);
                if (svg.opts["offsetX"] != null) e.attribute("x", true).value = svg.opts["offsetX"];
                if (svg.opts["offsetY"] != null) e.attribute("y", true).value = svg.opts["offsetY"];
                if (svg.opts["scaleWidth"] != null || svg.opts["scaleHeight"] != null) {
                    var xRatio = null, yRatio = null, viewBox = svg.ToNumberArray(e.attribute("viewBox").value);
                    if (svg.opts["scaleWidth"] != null) {
                        if (e.attribute("width").hasValue()) xRatio = e.attribute("width").toPixels("x") / svg.opts["scaleWidth"]; else if (!isNaN(viewBox[2])) xRatio = viewBox[2] / svg.opts["scaleWidth"];
                    }
                    if (svg.opts["scaleHeight"] != null) {
                        if (e.attribute("height").hasValue()) yRatio = e.attribute("height").toPixels("y") / svg.opts["scaleHeight"]; else if (!isNaN(viewBox[3])) yRatio = viewBox[3] / svg.opts["scaleHeight"];
                    }
                    if (xRatio == null) {
                        xRatio = yRatio;
                    }
                    if (yRatio == null) {
                        yRatio = xRatio;
                    }
                    e.attribute("width", true).value = svg.opts["scaleWidth"];
                    e.attribute("height", true).value = svg.opts["scaleHeight"];
                    e.style("transform", true, true).value += " scale(" + 1 / xRatio + "," + 1 / yRatio + ")";
                }
                if (svg.opts["ignoreClear"] != true) {
                    ctx.clearRect(0, 0, cWidth, cHeight);
                }
                e.render(ctx);
                if (isFirstRender) {
                    isFirstRender = false;
                    if (typeof svg.opts["renderCallback"] == "function") svg.opts["renderCallback"](dom);
                }
            };
            var waitingForImages = true;
            if (svg.ImagesLoaded()) {
                waitingForImages = false;
                draw();
            }
            svg.intervalID = setInterval(function() {
                var needUpdate = false;
                if (waitingForImages && svg.ImagesLoaded()) {
                    waitingForImages = false;
                    needUpdate = true;
                }
                if (svg.opts["ignoreMouse"] != true) {
                    needUpdate = needUpdate | svg.Mouse.hasEvents();
                }
                if (svg.opts["ignoreAnimation"] != true) {
                    for (var i = 0; i < svg.Animations.length; i++) {
                        needUpdate = needUpdate | svg.Animations[i].update(1e3 / svg.FRAMERATE);
                    }
                }
                if (typeof svg.opts["forceRedraw"] == "function") {
                    if (svg.opts["forceRedraw"]() == true) needUpdate = true;
                }
                if (needUpdate) {
                    draw();
                    svg.Mouse.runEvents();
                }
            }, 1e3 / svg.FRAMERATE);
        };
        svg.stop = function() {
            if (svg.intervalID) {
                clearInterval(svg.intervalID);
            }
        };
        svg.Mouse = new function() {
            this.events = [];
            this.hasEvents = function() {
                return this.events.length != 0;
            };
            this.onclick = function(x, y) {
                this.events.push({
                    type: "onclick",
                    x: x,
                    y: y,
                    run: function(e) {
                        if (e.onclick) e.onclick();
                    }
                });
            };
            this.onmousemove = function(x, y) {
                this.events.push({
                    type: "onmousemove",
                    x: x,
                    y: y,
                    run: function(e) {
                        if (e.onmousemove) e.onmousemove();
                    }
                });
            };
            this.eventElements = [];
            this.checkPath = function(element, ctx) {
                for (var i = 0; i < this.events.length; i++) {
                    var e = this.events[i];
                    if (ctx.isPointInPath && ctx.isPointInPath(e.x, e.y)) this.eventElements[i] = element;
                }
            };
            this.checkBoundingBox = function(element, bb) {
                for (var i = 0; i < this.events.length; i++) {
                    var e = this.events[i];
                    if (bb.isPointInBox(e.x, e.y)) this.eventElements[i] = element;
                }
            };
            this.runEvents = function() {
                svg.ctx.canvas.style.cursor = "";
                for (var i = 0; i < this.events.length; i++) {
                    var e = this.events[i];
                    var element = this.eventElements[i];
                    while (element) {
                        e.run(element);
                        element = element.parent;
                    }
                }
                this.events = [];
                this.eventElements = [];
            };
        }();
        return svg;
    }
})();

if (typeof CanvasRenderingContext2D != "undefined") {
    CanvasRenderingContext2D.prototype.drawSvg = function(s, dx, dy, dw, dh) {
        canvg(this.canvas, s, {
            ignoreMouse: true,
            ignoreAnimation: true,
            ignoreDimensions: true,
            ignoreClear: true,
            offsetX: dx,
            offsetY: dy,
            scaleWidth: dw,
            scaleHeight: dh
        });
    };
}

function RGBColor(color_string) {
    this.ok = false;
    if (color_string.charAt(0) == "#") {
        color_string = color_string.substr(1, 6);
    }
    color_string = color_string.replace(/ /g, "");
    color_string = color_string.toLowerCase();
    var simple_colors = {
        aliceblue: "f0f8ff",
        antiquewhite: "faebd7",
        aqua: "00ffff",
        aquamarine: "7fffd4",
        azure: "f0ffff",
        beige: "f5f5dc",
        bisque: "ffe4c4",
        black: "000000",
        blanchedalmond: "ffebcd",
        blue: "0000ff",
        blueviolet: "8a2be2",
        brown: "a52a2a",
        burlywood: "deb887",
        cadetblue: "5f9ea0",
        chartreuse: "7fff00",
        chocolate: "d2691e",
        coral: "ff7f50",
        cornflowerblue: "6495ed",
        cornsilk: "fff8dc",
        crimson: "dc143c",
        cyan: "00ffff",
        darkblue: "00008b",
        darkcyan: "008b8b",
        darkgoldenrod: "b8860b",
        darkgray: "a9a9a9",
        darkgreen: "006400",
        darkkhaki: "bdb76b",
        darkmagenta: "8b008b",
        darkolivegreen: "556b2f",
        darkorange: "ff8c00",
        darkorchid: "9932cc",
        darkred: "8b0000",
        darksalmon: "e9967a",
        darkseagreen: "8fbc8f",
        darkslateblue: "483d8b",
        darkslategray: "2f4f4f",
        darkturquoise: "00ced1",
        darkviolet: "9400d3",
        deeppink: "ff1493",
        deepskyblue: "00bfff",
        dimgray: "696969",
        dodgerblue: "1e90ff",
        feldspar: "d19275",
        firebrick: "b22222",
        floralwhite: "fffaf0",
        forestgreen: "228b22",
        fuchsia: "ff00ff",
        gainsboro: "dcdcdc",
        ghostwhite: "f8f8ff",
        gold: "ffd700",
        goldenrod: "daa520",
        gray: "808080",
        green: "008000",
        greenyellow: "adff2f",
        honeydew: "f0fff0",
        hotpink: "ff69b4",
        indianred: "cd5c5c",
        indigo: "4b0082",
        ivory: "fffff0",
        khaki: "f0e68c",
        lavender: "e6e6fa",
        lavenderblush: "fff0f5",
        lawngreen: "7cfc00",
        lemonchiffon: "fffacd",
        lightblue: "add8e6",
        lightcoral: "f08080",
        lightcyan: "e0ffff",
        lightgoldenrodyellow: "fafad2",
        lightgrey: "d3d3d3",
        lightgreen: "90ee90",
        lightpink: "ffb6c1",
        lightsalmon: "ffa07a",
        lightseagreen: "20b2aa",
        lightskyblue: "87cefa",
        lightslateblue: "8470ff",
        lightslategray: "778899",
        lightsteelblue: "b0c4de",
        lightyellow: "ffffe0",
        lime: "00ff00",
        limegreen: "32cd32",
        linen: "faf0e6",
        magenta: "ff00ff",
        maroon: "800000",
        mediumaquamarine: "66cdaa",
        mediumblue: "0000cd",
        mediumorchid: "ba55d3",
        mediumpurple: "9370d8",
        mediumseagreen: "3cb371",
        mediumslateblue: "7b68ee",
        mediumspringgreen: "00fa9a",
        mediumturquoise: "48d1cc",
        mediumvioletred: "c71585",
        midnightblue: "191970",
        mintcream: "f5fffa",
        mistyrose: "ffe4e1",
        moccasin: "ffe4b5",
        navajowhite: "ffdead",
        navy: "000080",
        oldlace: "fdf5e6",
        olive: "808000",
        olivedrab: "6b8e23",
        orange: "ffa500",
        orangered: "ff4500",
        orchid: "da70d6",
        palegoldenrod: "eee8aa",
        palegreen: "98fb98",
        paleturquoise: "afeeee",
        palevioletred: "d87093",
        papayawhip: "ffefd5",
        peachpuff: "ffdab9",
        peru: "cd853f",
        pink: "ffc0cb",
        plum: "dda0dd",
        powderblue: "b0e0e6",
        purple: "800080",
        red: "ff0000",
        rosybrown: "bc8f8f",
        royalblue: "4169e1",
        saddlebrown: "8b4513",
        salmon: "fa8072",
        sandybrown: "f4a460",
        seagreen: "2e8b57",
        seashell: "fff5ee",
        sienna: "a0522d",
        silver: "c0c0c0",
        skyblue: "87ceeb",
        slateblue: "6a5acd",
        slategray: "708090",
        snow: "fffafa",
        springgreen: "00ff7f",
        steelblue: "4682b4",
        tan: "d2b48c",
        teal: "008080",
        thistle: "d8bfd8",
        tomato: "ff6347",
        turquoise: "40e0d0",
        violet: "ee82ee",
        violetred: "d02090",
        wheat: "f5deb3",
        white: "ffffff",
        whitesmoke: "f5f5f5",
        yellow: "ffff00",
        yellowgreen: "9acd32"
    };
    for (var key in simple_colors) {
        if (color_string == key) {
            color_string = simple_colors[key];
        }
    }
    var color_defs = [ {
        re: /^rgb\((\d{1,3}),\s*(\d{1,3}),\s*(\d{1,3})\)$/,
        example: [ "rgb(123, 234, 45)", "rgb(255,234,245)" ],
        process: function(bits) {
            return [ parseInt(bits[1]), parseInt(bits[2]), parseInt(bits[3]) ];
        }
    }, {
        re: /^(\w{2})(\w{2})(\w{2})$/,
        example: [ "#00ff00", "336699" ],
        process: function(bits) {
            return [ parseInt(bits[1], 16), parseInt(bits[2], 16), parseInt(bits[3], 16) ];
        }
    }, {
        re: /^(\w{1})(\w{1})(\w{1})$/,
        example: [ "#fb0", "f0f" ],
        process: function(bits) {
            return [ parseInt(bits[1] + bits[1], 16), parseInt(bits[2] + bits[2], 16), parseInt(bits[3] + bits[3], 16) ];
        }
    } ];
    for (var i = 0; i < color_defs.length; i++) {
        var re = color_defs[i].re;
        var processor = color_defs[i].process;
        var bits = re.exec(color_string);
        if (bits) {
            channels = processor(bits);
            this.r = channels[0];
            this.g = channels[1];
            this.b = channels[2];
            this.ok = true;
        }
    }
    this.r = this.r < 0 || isNaN(this.r) ? 0 : this.r > 255 ? 255 : this.r;
    this.g = this.g < 0 || isNaN(this.g) ? 0 : this.g > 255 ? 255 : this.g;
    this.b = this.b < 0 || isNaN(this.b) ? 0 : this.b > 255 ? 255 : this.b;
    this.toRGB = function() {
        return "rgb(" + this.r + ", " + this.g + ", " + this.b + ")";
    };
    this.toHex = function() {
        var r = this.r.toString(16);
        var g = this.g.toString(16);
        var b = this.b.toString(16);
        if (r.length == 1) r = "0" + r;
        if (g.length == 1) g = "0" + g;
        if (b.length == 1) b = "0" + b;
        return "#" + r + g + b;
    };
    this.getHelpXML = function() {
        var examples = new Array();
        for (var i = 0; i < color_defs.length; i++) {
            var example = color_defs[i].example;
            for (var j = 0; j < example.length; j++) {
                examples[examples.length] = example[j];
            }
        }
        for (var sc in simple_colors) {
            examples[examples.length] = sc;
        }
        var xml = document.createElement("ul");
        xml.setAttribute("id", "rgbcolor-examples");
        for (var i = 0; i < examples.length; i++) {
            try {
                var list_item = document.createElement("li");
                var list_color = new RGBColor(examples[i]);
                var example_div = document.createElement("div");
                example_div.style.cssText = "margin: 3px; " + "border: 1px solid black; " + "background:" + list_color.toHex() + "; " + "color:" + list_color.toHex();
                example_div.appendChild(document.createTextNode("test"));
                var list_item_value = document.createTextNode(" " + examples[i] + " -> " + list_color.toRGB() + " -> " + list_color.toHex());
                list_item.appendChild(example_div);
                list_item.appendChild(list_item_value);
                xml.appendChild(list_item);
            } catch (e) {}
        }
        return xml;
    };
}

var mul_table = [ 512, 512, 456, 512, 328, 456, 335, 512, 405, 328, 271, 456, 388, 335, 292, 512, 454, 405, 364, 328, 298, 271, 496, 456, 420, 388, 360, 335, 312, 292, 273, 512, 482, 454, 428, 405, 383, 364, 345, 328, 312, 298, 284, 271, 259, 496, 475, 456, 437, 420, 404, 388, 374, 360, 347, 335, 323, 312, 302, 292, 282, 273, 265, 512, 497, 482, 468, 454, 441, 428, 417, 405, 394, 383, 373, 364, 354, 345, 337, 328, 320, 312, 305, 298, 291, 284, 278, 271, 265, 259, 507, 496, 485, 475, 465, 456, 446, 437, 428, 420, 412, 404, 396, 388, 381, 374, 367, 360, 354, 347, 341, 335, 329, 323, 318, 312, 307, 302, 297, 292, 287, 282, 278, 273, 269, 265, 261, 512, 505, 497, 489, 482, 475, 468, 461, 454, 447, 441, 435, 428, 422, 417, 411, 405, 399, 394, 389, 383, 378, 373, 368, 364, 359, 354, 350, 345, 341, 337, 332, 328, 324, 320, 316, 312, 309, 305, 301, 298, 294, 291, 287, 284, 281, 278, 274, 271, 268, 265, 262, 259, 257, 507, 501, 496, 491, 485, 480, 475, 470, 465, 460, 456, 451, 446, 442, 437, 433, 428, 424, 420, 416, 412, 408, 404, 400, 396, 392, 388, 385, 381, 377, 374, 370, 367, 363, 360, 357, 354, 350, 347, 344, 341, 338, 335, 332, 329, 326, 323, 320, 318, 315, 312, 310, 307, 304, 302, 299, 297, 294, 292, 289, 287, 285, 282, 280, 278, 275, 273, 271, 269, 267, 265, 263, 261, 259 ];

var shg_table = [ 9, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 17, 17, 17, 17, 17, 17, 17, 18, 18, 18, 18, 18, 18, 18, 18, 18, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24 ];

function premultiplyAlpha(imageData) {
    var pixels = imageData.data;
    var size = imageData.width * imageData.height * 4;
    for (var i = 0; i < size; i += 4) {
        var a = pixels[i + 3] / 255;
        pixels[i] *= a;
        pixels[i + 1] *= a;
        pixels[i + 2] *= a;
    }
}

function unpremultiplyAlpha(imageData) {
    var pixels = imageData.data;
    var size = imageData.width * imageData.height * 4;
    for (var i = 0; i < size; i += 4) {
        var a = pixels[i + 3];
        if (a != 0) {
            a = 255 / a;
            pixels[i] *= a;
            pixels[i + 1] *= a;
            pixels[i + 2] *= a;
        }
    }
}

function stackBlurImage(imageID, canvasID, radius, blurAlphaChannel) {
    var img = document.getElementById(imageID);
    var w = img.naturalWidth;
    var h = img.naturalHeight;
    var canvas = document.getElementById(canvasID);
    canvas.style.width = w + "px";
    canvas.style.height = h + "px";
    canvas.width = w;
    canvas.height = h;
    var context = canvas.getContext("2d");
    context.clearRect(0, 0, w, h);
    context.drawImage(img, 0, 0);
    if (isNaN(radius) || radius < 1) return;
    if (blurAlphaChannel) stackBlurCanvasRGBA(canvasID, 0, 0, w, h, radius); else stackBlurCanvasRGB(canvasID, 0, 0, w, h, radius);
}

function stackBlurCanvasRGBA(id, top_x, top_y, width, height, radius) {
    if (isNaN(radius) || radius < 1) return;
    radius |= 0;
    var canvas = document.getElementById(id);
    var context = canvas.getContext("2d");
    var imageData;
    try {
        try {
            imageData = context.getImageData(top_x, top_y, width, height);
        } catch (e) {
            try {
                netscape.security.PrivilegeManager.enablePrivilege("UniversalBrowserRead");
                imageData = context.getImageData(top_x, top_y, width, height);
            } catch (e) {
                alert("Cannot access local image");
                throw new Error("unable to access local image data: " + e);
                return;
            }
        }
    } catch (e) {
        alert("Cannot access image");
        throw new Error("unable to access image data: " + e);
    }
    premultiplyAlpha(imageData);
    var pixels = imageData.data;
    var x, y, i, p, yp, yi, yw, r_sum, g_sum, b_sum, a_sum, r_out_sum, g_out_sum, b_out_sum, a_out_sum, r_in_sum, g_in_sum, b_in_sum, a_in_sum, pr, pg, pb, pa, rbs;
    var div = radius + radius + 1;
    var w4 = width << 2;
    var widthMinus1 = width - 1;
    var heightMinus1 = height - 1;
    var radiusPlus1 = radius + 1;
    var sumFactor = radiusPlus1 * (radiusPlus1 + 1) / 2;
    var stackStart = new BlurStack();
    var stack = stackStart;
    for (i = 1; i < div; i++) {
        stack = stack.next = new BlurStack();
        if (i == radiusPlus1) var stackEnd = stack;
    }
    stack.next = stackStart;
    var stackIn = null;
    var stackOut = null;
    yw = yi = 0;
    var mul_sum = mul_table[radius];
    var shg_sum = shg_table[radius];
    for (y = 0; y < height; y++) {
        r_in_sum = g_in_sum = b_in_sum = a_in_sum = r_sum = g_sum = b_sum = a_sum = 0;
        r_out_sum = radiusPlus1 * (pr = pixels[yi]);
        g_out_sum = radiusPlus1 * (pg = pixels[yi + 1]);
        b_out_sum = radiusPlus1 * (pb = pixels[yi + 2]);
        a_out_sum = radiusPlus1 * (pa = pixels[yi + 3]);
        r_sum += sumFactor * pr;
        g_sum += sumFactor * pg;
        b_sum += sumFactor * pb;
        a_sum += sumFactor * pa;
        stack = stackStart;
        for (i = 0; i < radiusPlus1; i++) {
            stack.r = pr;
            stack.g = pg;
            stack.b = pb;
            stack.a = pa;
            stack = stack.next;
        }
        for (i = 1; i < radiusPlus1; i++) {
            p = yi + ((widthMinus1 < i ? widthMinus1 : i) << 2);
            r_sum += (stack.r = pr = pixels[p]) * (rbs = radiusPlus1 - i);
            g_sum += (stack.g = pg = pixels[p + 1]) * rbs;
            b_sum += (stack.b = pb = pixels[p + 2]) * rbs;
            a_sum += (stack.a = pa = pixels[p + 3]) * rbs;
            r_in_sum += pr;
            g_in_sum += pg;
            b_in_sum += pb;
            a_in_sum += pa;
            stack = stack.next;
        }
        stackIn = stackStart;
        stackOut = stackEnd;
        for (x = 0; x < width; x++) {
            pixels[yi] = r_sum * mul_sum >> shg_sum;
            pixels[yi + 1] = g_sum * mul_sum >> shg_sum;
            pixels[yi + 2] = b_sum * mul_sum >> shg_sum;
            pixels[yi + 3] = a_sum * mul_sum >> shg_sum;
            r_sum -= r_out_sum;
            g_sum -= g_out_sum;
            b_sum -= b_out_sum;
            a_sum -= a_out_sum;
            r_out_sum -= stackIn.r;
            g_out_sum -= stackIn.g;
            b_out_sum -= stackIn.b;
            a_out_sum -= stackIn.a;
            p = yw + ((p = x + radius + 1) < widthMinus1 ? p : widthMinus1) << 2;
            r_in_sum += stackIn.r = pixels[p];
            g_in_sum += stackIn.g = pixels[p + 1];
            b_in_sum += stackIn.b = pixels[p + 2];
            a_in_sum += stackIn.a = pixels[p + 3];
            r_sum += r_in_sum;
            g_sum += g_in_sum;
            b_sum += b_in_sum;
            a_sum += a_in_sum;
            stackIn = stackIn.next;
            r_out_sum += pr = stackOut.r;
            g_out_sum += pg = stackOut.g;
            b_out_sum += pb = stackOut.b;
            a_out_sum += pa = stackOut.a;
            r_in_sum -= pr;
            g_in_sum -= pg;
            b_in_sum -= pb;
            a_in_sum -= pa;
            stackOut = stackOut.next;
            yi += 4;
        }
        yw += width;
    }
    for (x = 0; x < width; x++) {
        g_in_sum = b_in_sum = a_in_sum = r_in_sum = g_sum = b_sum = a_sum = r_sum = 0;
        yi = x << 2;
        r_out_sum = radiusPlus1 * (pr = pixels[yi]);
        g_out_sum = radiusPlus1 * (pg = pixels[yi + 1]);
        b_out_sum = radiusPlus1 * (pb = pixels[yi + 2]);
        a_out_sum = radiusPlus1 * (pa = pixels[yi + 3]);
        r_sum += sumFactor * pr;
        g_sum += sumFactor * pg;
        b_sum += sumFactor * pb;
        a_sum += sumFactor * pa;
        stack = stackStart;
        for (i = 0; i < radiusPlus1; i++) {
            stack.r = pr;
            stack.g = pg;
            stack.b = pb;
            stack.a = pa;
            stack = stack.next;
        }
        yp = width;
        for (i = 1; i <= radius; i++) {
            yi = yp + x << 2;
            r_sum += (stack.r = pr = pixels[yi]) * (rbs = radiusPlus1 - i);
            g_sum += (stack.g = pg = pixels[yi + 1]) * rbs;
            b_sum += (stack.b = pb = pixels[yi + 2]) * rbs;
            a_sum += (stack.a = pa = pixels[yi + 3]) * rbs;
            r_in_sum += pr;
            g_in_sum += pg;
            b_in_sum += pb;
            a_in_sum += pa;
            stack = stack.next;
            if (i < heightMinus1) {
                yp += width;
            }
        }
        yi = x;
        stackIn = stackStart;
        stackOut = stackEnd;
        for (y = 0; y < height; y++) {
            p = yi << 2;
            pixels[p] = r_sum * mul_sum >> shg_sum;
            pixels[p + 1] = g_sum * mul_sum >> shg_sum;
            pixels[p + 2] = b_sum * mul_sum >> shg_sum;
            pixels[p + 3] = a_sum * mul_sum >> shg_sum;
            r_sum -= r_out_sum;
            g_sum -= g_out_sum;
            b_sum -= b_out_sum;
            a_sum -= a_out_sum;
            r_out_sum -= stackIn.r;
            g_out_sum -= stackIn.g;
            b_out_sum -= stackIn.b;
            a_out_sum -= stackIn.a;
            p = x + ((p = y + radiusPlus1) < heightMinus1 ? p : heightMinus1) * width << 2;
            r_sum += r_in_sum += stackIn.r = pixels[p];
            g_sum += g_in_sum += stackIn.g = pixels[p + 1];
            b_sum += b_in_sum += stackIn.b = pixels[p + 2];
            a_sum += a_in_sum += stackIn.a = pixels[p + 3];
            stackIn = stackIn.next;
            r_out_sum += pr = stackOut.r;
            g_out_sum += pg = stackOut.g;
            b_out_sum += pb = stackOut.b;
            a_out_sum += pa = stackOut.a;
            r_in_sum -= pr;
            g_in_sum -= pg;
            b_in_sum -= pb;
            a_in_sum -= pa;
            stackOut = stackOut.next;
            yi += width;
        }
    }
    unpremultiplyAlpha(imageData);
    context.putImageData(imageData, top_x, top_y);
}

function stackBlurCanvasRGB(id, top_x, top_y, width, height, radius) {
    if (isNaN(radius) || radius < 1) return;
    radius |= 0;
    var canvas = document.getElementById(id);
    var context = canvas.getContext("2d");
    var imageData;
    try {
        try {
            imageData = context.getImageData(top_x, top_y, width, height);
        } catch (e) {
            try {
                netscape.security.PrivilegeManager.enablePrivilege("UniversalBrowserRead");
                imageData = context.getImageData(top_x, top_y, width, height);
            } catch (e) {
                alert("Cannot access local image");
                throw new Error("unable to access local image data: " + e);
                return;
            }
        }
    } catch (e) {
        alert("Cannot access image");
        throw new Error("unable to access image data: " + e);
    }
    var pixels = imageData.data;
    var x, y, i, p, yp, yi, yw, r_sum, g_sum, b_sum, r_out_sum, g_out_sum, b_out_sum, r_in_sum, g_in_sum, b_in_sum, pr, pg, pb, rbs;
    var div = radius + radius + 1;
    var w4 = width << 2;
    var widthMinus1 = width - 1;
    var heightMinus1 = height - 1;
    var radiusPlus1 = radius + 1;
    var sumFactor = radiusPlus1 * (radiusPlus1 + 1) / 2;
    var stackStart = new BlurStack();
    var stack = stackStart;
    for (i = 1; i < div; i++) {
        stack = stack.next = new BlurStack();
        if (i == radiusPlus1) var stackEnd = stack;
    }
    stack.next = stackStart;
    var stackIn = null;
    var stackOut = null;
    yw = yi = 0;
    var mul_sum = mul_table[radius];
    var shg_sum = shg_table[radius];
    for (y = 0; y < height; y++) {
        r_in_sum = g_in_sum = b_in_sum = r_sum = g_sum = b_sum = 0;
        r_out_sum = radiusPlus1 * (pr = pixels[yi]);
        g_out_sum = radiusPlus1 * (pg = pixels[yi + 1]);
        b_out_sum = radiusPlus1 * (pb = pixels[yi + 2]);
        r_sum += sumFactor * pr;
        g_sum += sumFactor * pg;
        b_sum += sumFactor * pb;
        stack = stackStart;
        for (i = 0; i < radiusPlus1; i++) {
            stack.r = pr;
            stack.g = pg;
            stack.b = pb;
            stack = stack.next;
        }
        for (i = 1; i < radiusPlus1; i++) {
            p = yi + ((widthMinus1 < i ? widthMinus1 : i) << 2);
            r_sum += (stack.r = pr = pixels[p]) * (rbs = radiusPlus1 - i);
            g_sum += (stack.g = pg = pixels[p + 1]) * rbs;
            b_sum += (stack.b = pb = pixels[p + 2]) * rbs;
            r_in_sum += pr;
            g_in_sum += pg;
            b_in_sum += pb;
            stack = stack.next;
        }
        stackIn = stackStart;
        stackOut = stackEnd;
        for (x = 0; x < width; x++) {
            pixels[yi] = r_sum * mul_sum >> shg_sum;
            pixels[yi + 1] = g_sum * mul_sum >> shg_sum;
            pixels[yi + 2] = b_sum * mul_sum >> shg_sum;
            r_sum -= r_out_sum;
            g_sum -= g_out_sum;
            b_sum -= b_out_sum;
            r_out_sum -= stackIn.r;
            g_out_sum -= stackIn.g;
            b_out_sum -= stackIn.b;
            p = yw + ((p = x + radius + 1) < widthMinus1 ? p : widthMinus1) << 2;
            r_in_sum += stackIn.r = pixels[p];
            g_in_sum += stackIn.g = pixels[p + 1];
            b_in_sum += stackIn.b = pixels[p + 2];
            r_sum += r_in_sum;
            g_sum += g_in_sum;
            b_sum += b_in_sum;
            stackIn = stackIn.next;
            r_out_sum += pr = stackOut.r;
            g_out_sum += pg = stackOut.g;
            b_out_sum += pb = stackOut.b;
            r_in_sum -= pr;
            g_in_sum -= pg;
            b_in_sum -= pb;
            stackOut = stackOut.next;
            yi += 4;
        }
        yw += width;
    }
    for (x = 0; x < width; x++) {
        g_in_sum = b_in_sum = r_in_sum = g_sum = b_sum = r_sum = 0;
        yi = x << 2;
        r_out_sum = radiusPlus1 * (pr = pixels[yi]);
        g_out_sum = radiusPlus1 * (pg = pixels[yi + 1]);
        b_out_sum = radiusPlus1 * (pb = pixels[yi + 2]);
        r_sum += sumFactor * pr;
        g_sum += sumFactor * pg;
        b_sum += sumFactor * pb;
        stack = stackStart;
        for (i = 0; i < radiusPlus1; i++) {
            stack.r = pr;
            stack.g = pg;
            stack.b = pb;
            stack = stack.next;
        }
        yp = width;
        for (i = 1; i <= radius; i++) {
            yi = yp + x << 2;
            r_sum += (stack.r = pr = pixels[yi]) * (rbs = radiusPlus1 - i);
            g_sum += (stack.g = pg = pixels[yi + 1]) * rbs;
            b_sum += (stack.b = pb = pixels[yi + 2]) * rbs;
            r_in_sum += pr;
            g_in_sum += pg;
            b_in_sum += pb;
            stack = stack.next;
            if (i < heightMinus1) {
                yp += width;
            }
        }
        yi = x;
        stackIn = stackStart;
        stackOut = stackEnd;
        for (y = 0; y < height; y++) {
            p = yi << 2;
            pixels[p] = r_sum * mul_sum >> shg_sum;
            pixels[p + 1] = g_sum * mul_sum >> shg_sum;
            pixels[p + 2] = b_sum * mul_sum >> shg_sum;
            r_sum -= r_out_sum;
            g_sum -= g_out_sum;
            b_sum -= b_out_sum;
            r_out_sum -= stackIn.r;
            g_out_sum -= stackIn.g;
            b_out_sum -= stackIn.b;
            p = x + ((p = y + radiusPlus1) < heightMinus1 ? p : heightMinus1) * width << 2;
            r_sum += r_in_sum += stackIn.r = pixels[p];
            g_sum += g_in_sum += stackIn.g = pixels[p + 1];
            b_sum += b_in_sum += stackIn.b = pixels[p + 2];
            stackIn = stackIn.next;
            r_out_sum += pr = stackOut.r;
            g_out_sum += pg = stackOut.g;
            b_out_sum += pb = stackOut.b;
            r_in_sum -= pr;
            g_in_sum -= pg;
            b_in_sum -= pb;
            stackOut = stackOut.next;
            yi += width;
        }
    }
    context.putImageData(imageData, top_x, top_y);
}

function BlurStack() {
    this.r = 0;
    this.g = 0;
    this.b = 0;
    this.a = 0;
    this.next = null;
}

(function(view) {
    "use strict";
    var Uint8Array = view.Uint8Array, HTMLCanvasElement = view.HTMLCanvasElement, canvas_proto = HTMLCanvasElement && HTMLCanvasElement.prototype, is_base64_regex = /\s*;\s*base64\s*(?:;|$)/i, to_data_url = "toDataURL", base64_ranks, decode_base64 = function(base64) {
        var len = base64.length, buffer = new Uint8Array(len / 4 * 3 | 0), i = 0, outptr = 0, last = [ 0, 0 ], state = 0, save = 0, rank, code, undef;
        while (len--) {
            code = base64.charCodeAt(i++);
            rank = base64_ranks[code - 43];
            if (rank !== 255 && rank !== undef) {
                last[1] = last[0];
                last[0] = code;
                save = save << 6 | rank;
                state++;
                if (state === 4) {
                    buffer[outptr++] = save >>> 16;
                    if (last[1] !== 61) {
                        buffer[outptr++] = save >>> 8;
                    }
                    if (last[0] !== 61) {
                        buffer[outptr++] = save;
                    }
                    state = 0;
                }
            }
        }
        return buffer;
    };
    if (Uint8Array) {
        base64_ranks = new Uint8Array([ 62, -1, -1, -1, 63, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, 0, -1, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1, -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51 ]);
    }
    if (HTMLCanvasElement && (!canvas_proto.toBlob || !canvas_proto.toBlobHD)) {
        if (!canvas_proto.toBlob) canvas_proto.toBlob = function(callback, type) {
            if (!type) {
                type = "image/png";
            }
            if (this.mozGetAsFile) {
                callback(this.mozGetAsFile("canvas", type));
                return;
            }
            if (this.msToBlob && /^\s*image\/png\s*(?:$|;)/i.test(type)) {
                callback(this.msToBlob());
                return;
            }
            var args = Array.prototype.slice.call(arguments, 1), dataURI = this[to_data_url].apply(this, args), header_end = dataURI.indexOf(","), data = dataURI.substring(header_end + 1), is_base64 = is_base64_regex.test(dataURI.substring(0, header_end)), blob;
            if (Blob.fake) {
                blob = new Blob();
                if (is_base64) {
                    blob.encoding = "base64";
                } else {
                    blob.encoding = "URI";
                }
                blob.data = data;
                blob.size = data.length;
            } else if (Uint8Array) {
                if (is_base64) {
                    blob = new Blob([ decode_base64(data) ], {
                        type: type
                    });
                } else {
                    blob = new Blob([ decodeURIComponent(data) ], {
                        type: type
                    });
                }
            }
            callback(blob);
        };
        if (!canvas_proto.toBlobHD && canvas_proto.toDataURLHD) {
            canvas_proto.toBlobHD = function() {
                to_data_url = "toDataURLHD";
                var blob = this.toBlob();
                to_data_url = "toDataURL";
                return blob;
            };
        } else {
            canvas_proto.toBlobHD = canvas_proto.toBlob;
        }
    }
})(typeof self !== "undefined" && self || typeof window !== "undefined" && window || this.content || this);

var saveAs = saveAs || function(view) {
    "use strict";
    if (typeof view === "undefined" || typeof navigator !== "undefined" && /MSIE [1-9]\./.test(navigator.userAgent)) {
        return;
    }
    var doc = view.document, get_URL = function() {
        return view.URL || view.webkitURL || view;
    }, save_link = doc.createElementNS("http://www.w3.org/1999/xhtml", "a"), can_use_save_link = "download" in save_link, click = function(node) {
        var event = new MouseEvent("click");
        node.dispatchEvent(event);
    }, is_safari = /constructor/i.test(view.HTMLElement) || view.safari, is_chrome_ios = /CriOS\/[\d]+/.test(navigator.userAgent), throw_outside = function(ex) {
        (view.setImmediate || view.setTimeout)(function() {
            throw ex;
        }, 0);
    }, force_saveable_type = "application/octet-stream", arbitrary_revoke_timeout = 1e3 * 40, revoke = function(file) {
        var revoker = function() {
            if (typeof file === "string") {
                get_URL().revokeObjectURL(file);
            } else {
                file.remove();
            }
        };
        setTimeout(revoker, arbitrary_revoke_timeout);
    }, dispatch = function(filesaver, event_types, event) {
        event_types = [].concat(event_types);
        var i = event_types.length;
        while (i--) {
            var listener = filesaver["on" + event_types[i]];
            if (typeof listener === "function") {
                try {
                    listener.call(filesaver, event || filesaver);
                } catch (ex) {
                    throw_outside(ex);
                }
            }
        }
    }, auto_bom = function(blob) {
        if (/^\s*(?:text\/\S*|application\/xml|\S*\/\S*\+xml)\s*;.*charset\s*=\s*utf-8/i.test(blob.type)) {
            return new Blob([ String.fromCharCode(65279), blob ], {
                type: blob.type
            });
        }
        return blob;
    }, FileSaver = function(blob, name, no_auto_bom) {
        if (!no_auto_bom) {
            blob = auto_bom(blob);
        }
        var filesaver = this, type = blob.type, force = type === force_saveable_type, object_url, dispatch_all = function() {
            dispatch(filesaver, "writestart progress write writeend".split(" "));
        }, fs_error = function() {
            if ((is_chrome_ios || force && is_safari) && view.FileReader) {
                var reader = new FileReader();
                reader.onloadend = function() {
                    var url = is_chrome_ios ? reader.result : reader.result.replace(/^data:[^;]*;/, "data:attachment/file;");
                    var popup = view.open(url, "_blank");
                    if (!popup) view.location.href = url;
                    url = undefined;
                    filesaver.readyState = filesaver.DONE;
                    dispatch_all();
                };
                reader.readAsDataURL(blob);
                filesaver.readyState = filesaver.INIT;
                return;
            }
            if (!object_url) {
                object_url = get_URL().createObjectURL(blob);
            }
            if (force) {
                view.location.href = object_url;
            } else {
                var opened = view.open(object_url, "_blank");
                if (!opened) {
                    view.location.href = object_url;
                }
            }
            filesaver.readyState = filesaver.DONE;
            dispatch_all();
            revoke(object_url);
        };
        filesaver.readyState = filesaver.INIT;
        if (can_use_save_link) {
            object_url = get_URL().createObjectURL(blob);
            setTimeout(function() {
                save_link.href = object_url;
                save_link.download = name;
                click(save_link);
                dispatch_all();
                revoke(object_url);
                filesaver.readyState = filesaver.DONE;
            });
            return;
        }
        fs_error();
    }, FS_proto = FileSaver.prototype, saveAs = function(blob, name, no_auto_bom) {
        return new FileSaver(blob, name || blob.name || "download", no_auto_bom);
    };
    if (typeof navigator !== "undefined" && navigator.msSaveOrOpenBlob) {
        return function(blob, name, no_auto_bom) {
            name = name || blob.name || "download";
            if (!no_auto_bom) {
                blob = auto_bom(blob);
            }
            return navigator.msSaveOrOpenBlob(blob, name);
        };
    }
    FS_proto.abort = function() {};
    FS_proto.readyState = FS_proto.INIT = 0;
    FS_proto.WRITING = 1;
    FS_proto.DONE = 2;
    FS_proto.error = FS_proto.onwritestart = FS_proto.onprogress = FS_proto.onwrite = FS_proto.onabort = FS_proto.onerror = FS_proto.onwriteend = null;
    return saveAs;
}(typeof self !== "undefined" && self || typeof window !== "undefined" && window || this.content);

if (typeof module !== "undefined" && module.exports) {
    module.exports.saveAs = saveAs;
} else if (typeof define !== "undefined" && define !== null && define.amd !== null) {
    define("FileSaver.js", function() {
        return saveAs;
    });
}

function svgToPNG(id, filename) {
    if (filename == null) filename = "structure.png";
    var DOMURL = window.URL || window.webkitURL || window;
    var xml = $("<div>").append($("[id=" + id + "] svg:eq(0)").clone()).html();
    xml = xml.replace('xmlns="http://www.w3.org/2000/svg" xmlns:NS1="" NS1:xmlns:xlink="http://www.w3.org/1999/xlink"', "");
    var blob = new Blob([ xml ], {
        type: "image/svg+xml"
    });
    var url = DOMURL.createObjectURL(blob);
    $("body").append('<canvas display="none" id="tmpcanvas"></canvas>');
    canvg("tmpcanvas", url, {
        renderCallback: function() {
            var c = document.getElementById("tmpcanvas");
            c.toBlob(function(blob) {
                saveAs(blob, filename);
            });
            c.parentNode.removeChild(c);
        }
    });
}

function qmeanLocalColour(score, alpha) {
    if (score > .4) {
        if (alpha == null) return "rgb(" + Math.round(255 * (1 - score) * (1 / .6)) + ",80," + Math.round(255 * (score - .4) * (1 / .6)) + ")"; else return "rgba(" + Math.round(255 * (1 - score) * (1 / .6)) + ",80," + Math.round(255 * (score - .4) * (1 / .6)) + "," + alpha + ")";
    } else if (alpha == null) return "rgb(255,80,0)"; else return "rgba(255,80,0," + alpha + ")";
}

function drawQMEANLocal(struc_id, avgLocalScore, chartNo, titleSmall, resizing, isQMEANBrane, maxThumbHeight = 200, chartIds = undefined, boundingRect) {
    let qm = isQMEANBrane ? "qmb" : "qm";
    let cmap = chainMapping.structures[struc_id];
    chartNo = chartNo ? "-" + chartNo : "";
    let allChainsChart = `lq-chart${chartNo}`;
    if (!chartIds) {
        chartIds = [ `${allChainsChart}-thumb`, `lq-chart${chartNo}` ];
        if (Object.keys(cmap).length > 1) for (let i = 0; i < Object.keys(cmap).length; i++) chartIds.push(`${allChainsChart}-${Object.keys(cmap)[i]}`);
    }
    for (let chartId of chartIds) {
        const elem = document.getElementById(chartId);
        if (elem) {
            if (!resizing && elem.querySelector(`svg`) !== null) return;
            while (elem.firstChild) elem.removeChild(elem.firstChild);
        } else {
            $("body").append(`<div id="${chartId}" style="display:none"></div>`);
            let linkTo = chartId;
            if (chartId.includes("-thumb")) linkTo = linkTo.replace("-thumb", "");
            if (!document.querySelector(`a[href="#${linkTo}"]`)) $("body").append(`<a href="#${linkTo}" rel="lq${chartNo}" class="fancybox"/>`);
        }
    }
    for (let chartId of chartIds) {
        let thumbnail = chartId.includes("-thumb");
        let fontSize;
        let offsetL = 0, offsetT = 0, offsetB = 0;
        let mdThumb = false;
        if (thumbnail) {
            $("#" + chartId).attr("title", "Click to enlarge");
            w = $("#" + chartId).closest(":visible").width();
            h = $("#" + chartId).closest(":visible").height();
            if (w === null || h === null) return;
            if (h === 0) h = Math.min(maxThumbHeight, w * .5);
            if (h >= maxThumbHeight) {
                fontSize = 12;
                offsetL = 45;
                offsetT = 30;
                offsetB = 40;
                mdThumb = true;
            } else {
                if (titleSmall) {
                    offsetL = 30;
                    offsetT = 16;
                }
                fontSize = Math.max(15, h / 10);
            }
        } else {
            fontSize = 20;
            offsetL = 45;
            offsetT = 30;
            offsetB = 40;
            w = boundingRect ? boundingRect.width : $(window).width() * .7 + offsetL;
            h = boundingRect ? boundingRect.height - offsetB - offsetT : $(window).height() * .7 + offsetB + offsetT;
        }
        let qe = Raphael(chartId, w, h);
        let minres = 99999, maxres = 0;
        let tot = 0, count = 0, minScore = 999, maxScore = 0;
        let chains = Object.keys(cmap);
        for (let j = 0; j < chains.length; j++) {
            if (!thumbnail) if (!(chartId === allChainsChart || chartId.endsWith("-" + chains[j]))) continue;
            for (let x = 0; x < cmap[chains[j]].length; x++) {
                if (!cmap[chains[j]][x] || !cmap[chains[j]][x][qm]) continue;
                minScore = Math.min(minScore, cmap[chains[j]][x][qm]);
                maxScore = Math.max(maxScore, cmap[chains[j]][x][qm]);
                minres = Math.min(cmap[chains[j]][x].resno, minres);
                maxres = Math.max(cmap[chains[j]][x].resno, maxres);
                count += 1;
                tot += cmap[chains[j]][x][qm];
            }
        }
        let below1 = maxScore <= 1;
        let ceilY = 1;
        if (below1) {
            minScore = Math.min(.3, minScore);
        } else {
            ceilY = 100;
            minScore = Math.min(70, minScore);
        }
        if (avgLocalScore == null) avgLocalScore = tot / count;
        $("#avgLocalQMEAN,#avgLocalScore_" + chartNo).html(isNaN(avgLocalScore) ? "-" : avgLocalScore.toFixed(2));
        let chartH = h - offsetB - offsetT;
        let resW = (w - offsetL) / (maxres - minres + 1);
        qe.rect(0, 0, w, h).attr({
            fill: "#FFF",
            stroke: "#FFF"
        });
        if (thumbnail && !mdThumb) {
            qe.rect(offsetL, offsetT, w - offsetL, chartH).attr({
                stroke: "#DDD"
            });
            if (titleSmall) {
                qe.text(w / 2, 8, "Local Quality Estimate").attr({
                    "text-anchor": "middle",
                    fill: "#444",
                    "font-size": Math.max(8, h / 24)
                });
                qe.text(6, h / 2 + offsetT / 2, "Predicted Local").attr({
                    "text-anchor": "middle",
                    fill: "#666",
                    "font-size": Math.max(8, h / 26)
                }).rotate(270, 6, h / 2 + offsetT / 2);
                qe.text(18, h / 2 + offsetT / 2, "Similarity to Target").attr({
                    "text-anchor": "middle",
                    fill: "#666",
                    "font-size": Math.max(8, h / 26)
                }).rotate(270, 18, h / 2 + offsetT / 2);
            }
        } else {
            qe.rect(offsetL, offsetT, w - offsetL, chartH).attr({
                stroke: "#DDD"
            });
            let scaleFont = 14;
            let title = "Local Quality Estimate";
            if (isQMEANBrane) title += " (QMEANBrane) ";
            if (!thumbnail && chains.length > 1) {
                if (chartId === allChainsChart) title += " - All Chains"; else title += " - Chain " + chartId.substring(chartId.lastIndexOf("-") + 1);
            }
            qe.text(w / 2, 10, title).attr({
                "text-anchor": "middle",
                fill: "#444",
                "font-size": Math.max(8, h / 26)
            });
            if (thumbnail) {
                qe.text(6, h / 2, "Predicted Local").attr({
                    "text-anchor": "middle",
                    fill: "#666",
                    "font-size": fontSize
                }).rotate(270, 6, h / 2);
                qe.text(18, h / 2, "Similarity to Target").attr({
                    "text-anchor": "middle",
                    fill: "#666",
                    "font-size": fontSize
                }).rotate(270, 18, h / 2);
                scaleFont = fontSize * .8;
            } else qe.text(10, h / 2, "Predicted Local Similarity to Target").attr({
                dy: 10,
                "text-anchor": "middle",
                fill: "#666",
                "font-size": fontSize
            }).rotate(270, 10, h / 2);
            qe.text(offsetL + (w - offsetL) / 2, h - 8, "Residue Number").attr({
                "text-anchor": "middle",
                fill: "#666",
                "font-size": fontSize
            });
            let lastTickY = null;
            for (let j = ceilY; j >= minScore; j -= below1 ? .1 : 10) {
                qe.path("M" + offsetL + " " + (chartH / (ceilY - minScore) * (ceilY - j) + offsetT) + "L" + (offsetL - 5) + " " + (chartH / (ceilY - minScore) * (ceilY - j) + offsetT)).attr({
                    stroke: "#CCC",
                    "stroke-width": 1
                });
                let thisTickY = chartH / (ceilY - minScore) * (ceilY - j) + offsetT;
                if (!thisTickY || lastTickY + scaleFont < thisTickY) {
                    qe.text(offsetL - 12, thisTickY, j.toFixed(below1 ? 1 : 0)).attr({
                        fill: "#444",
                        "font-size": scaleFont
                    });
                    lastTickY = thisTickY;
                }
            }
            let chunksize = Math.floor((maxres - minres) / 8);
            chunksize -= chunksize % 10;
            chunksize = Math.max(10, chunksize);
            let lastX2 = null;
            for (let j = minres; j < maxres; j++) {
                if (j % chunksize == 0) {
                    let x = offsetL + (j - minres) / (maxres - minres) * (w - offsetL);
                    if (lastX2 && lastX2 > x) continue;
                    let t = qe.text(x, h - offsetB + 12, j).attr({
                        fill: "#444",
                        "font-size": scaleFont
                    });
                    lastX2 = x + t.getBBox().width;
                }
            }
        }
        let col = [ "#999999", "#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2", "#D55E00", "#CC79A7" ];
        let colIdx = 0;
        for (let j = 0; j < chains.length; j++) {
            if (!thumbnail) if (!(chartId === allChainsChart || chartId.endsWith("-" + chains[j]))) continue;
            let fullpath = null;
            let lineGraph = thumbnail && chains.length > 1 || !thumbnail && chains.length > 1 && chartId === allChainsChart;
            let baseY = chartH / (ceilY - Math.min(minScore, avgLocalScore)) * (ceilY - avgLocalScore) + offsetT;
            for (let l = 0; l < cmap[chains[j]].length; l++) {
                if (!cmap[chains[j]][l] || cmap[chains[j]][l][qm] == null) continue;
                let score = cmap[chains[j]][l][qm];
                let xPos = offsetL + resW * (cmap[chains[j]][l].resno - minres);
                let yPos = baseY - chartH / (ceilY - minScore) * (score - avgLocalScore);
                if (lineGraph) {
                    yPos = Math.min(yPos, chartH + offsetT);
                    if (fullpath == null) fullpath = `M${xPos} ${yPos}`; else fullpath += `L${xPos} ${yPos}`;
                    fullpath += `L${xPos + resW} ${yPos}`;
                } else {
                    if (score > avgLocalScore) bar = qe.rect(xPos, yPos, resW, chartH / (ceilY - minScore) * (score - avgLocalScore)); else {
                        let h = chartH / (ceilY - minScore) * (avgLocalScore - score);
                        h = Math.min(h, chartH - baseY + offsetT);
                        bar = qe.rect(xPos, baseY, resW, h);
                    }
                    let rgb = qmeanLocalColour(below1 ? score : score / 100);
                    bar.attr({
                        stroke: rgb,
                        "stroke-width": .7,
                        fill: rgb,
                        "fill-opacity": .4
                    });
                }
            }
            if (chains.length > 1 && (thumbnail || chartId === allChainsChart)) {
                fullpath = qe.path(fullpath);
                fullpath.attr({
                    stroke: col[colIdx],
                    "stroke-width": thumbnail ? 1 : 1.5
                });
                if (!thumbnail) $("#" + chartId).append(`${j === 0 ? "<br>Key: " : ""} <span style="margin-right:4px; color:${col[colIdx]}">Chain ${chains[j]};</span>`);
            }
            colIdx = ++colIdx % col.length;
        }
    }
}

function drawQMEANGlobal(qmeanData, residueCount, chartNo, titleSmall, resizing) {
    let $charts = $("#qn-chart" + (chartNo ? "-" + chartNo : "") + "-thumb,#qn-chart" + (chartNo ? "-" + chartNo : ""));
    if (!qmeanData) return;
    let zscoreTblId = "qmeanGlobalTbl" + (chartNo ? chartNo : "");
    if (!resizing && $("#" + zscoreTblId + " tr").length > 0) return;
    $charts.empty();
    $("#" + zscoreTblId + " tr").remove();
    if (qmeanData["QMEAN4"]) {
        qmeanData["qmean4_norm_score"] = qmeanData["qmean4_norm"];
        delete qmeanData["qmean4_norm"];
        qmeanData["qmean4_z_score"] = qmeanData["QMEAN4"];
        delete qmeanData["QMEAN4"];
        qmeanData["cbeta_z_score"] = qmeanData["cbeta"];
        delete qmeanData["cbeta"];
        qmeanData["interaction_z_score"] = qmeanData["all_atom"];
        delete qmeanData["all_atom"];
        qmeanData["packing_z_score"] = qmeanData["solvation"];
        delete qmeanData["solvation"];
        qmeanData["torsion_z_score"] = qmeanData["torsion"];
        delete qmeanData["torsion"];
    }
    if (qmeanData["qmean4"]) {
        qmeanData["qmean4_norm_score"] = parseFloat(qmeanData["qmean4"]["norm"]);
        qmeanData["qmean4_z_score"] = parseFloat(qmeanData["qmean4"]["zscore"]);
        delete qmeanData["qmean4"];
        qmeanData["cbeta_z_score"] = parseFloat(qmeanData["cbeta"]["zscore"]);
        delete qmeanData["cbeta"];
        qmeanData["interaction_z_score"] = parseFloat(qmeanData["all_atom"]["zscore"]);
        delete qmeanData["all_atom"];
        qmeanData["packing_z_score"] = parseFloat(qmeanData["solvation"]["zscore"]);
        delete qmeanData["solvation"];
        qmeanData["torsion_z_score"] = parseFloat(qmeanData["torsion"]["zscore"]);
        delete qmeanData["torsion"];
        delete qmeanData["acc_agreement"];
        delete qmeanData["ss_agreement"];
        delete qmeanData["qmean6"];
    }
    let charts = [ [ "qmean4_z_score", "QMEAN" ], [ "cbeta_z_score", "Cβ" ], [ "interaction_z_score", "All Atom" ], [ "packing_z_score", "solvation" ], [ "torsion_z_score", "torsion" ] ];
    let html = "";
    for (let i = 0; i < charts.length; i++) {
        if (qmeanData[charts[i][0]] != null) html += '<tr><td width="1">' + charts[i][1] + '</td><td width="100%" style="padding:0px 2px 0px 2px;"><div id="' + zscoreTblId + "_" + charts[i][0] + '">&nbsp;</div></td><td width="1" nowrap="nowrap" style="text-align:right">' + qmeanData[charts[i][0]].toFixed(2) + "</td></tr>";
    }
    $("#" + zscoreTblId).append(html);
    let w = $("#" + zscoreTblId + "_" + charts[0][0]).innerWidth();
    let h = $("#" + zscoreTblId + "_" + charts[0][0]).parent().innerHeight();
    if (w < 1 || h < 1) {
        w = 290;
        h = 20;
    }
    for (let i = 0; i < charts.length; i++) $("#" + zscoreTblId + "_" + charts[i][0]).html("");
    let d = w / 8;
    for (let j = 0; j < charts.length; j++) {
        if (qmeanData[charts[j][0]] == null) continue;
        let qm = Raphael(zscoreTblId + "_" + charts[j][0], w, h);
        qm.rect(0, 0, w, h).attr({
            fill: "0-rgb(245,61,61)-#fff:75-rgb(77,77,255)"
        });
        let tickHeight = h / 4;
        for (let i = 1; i < 8; i++) {
            qm.path("M" + d * i + " 0L" + d * i + " " + tickHeight).attr({
                stroke: "#555"
            });
            qm.path("M" + d * i + " " + (h - tickHeight) + "L" + d * i + " " + h).attr({
                stroke: "#555"
            });
        }
        if (qmeanData) {
            let score = qmeanData[charts[j][0]];
            score = Math.min(score, 1.95);
            score = Math.max(score, -5.95);
            qm.path("M" + (d * score + 6 * d) + " 0L" + (d * score + 6 * d) + " " + h).attr({
                stroke: "#222",
                "stroke-width": 3
            });
        }
    }
    let qmean4norm = qmeanData["qmean4_norm_score"];
    let proteinSize = Math.min(600, residueCount);
    let thumbnail = true;
    $charts.each(function(idx, obj) {
        let maxThumbWidth = 340;
        let w, h, fontSize;
        let offsetL = 0, offsetT = 0, offsetB = 0;
        if (thumbnail) {
            $(this).attr("title", "Click to enlarge");
            w = $(this).closest(":visible").width();
            if (w > maxThumbWidth) {
                fontSize = 12;
                offsetL = 45;
                offsetT = 30;
                offsetB = 40;
                w -= offsetL;
                h = 400 / 620 * w - offsetB + offsetT;
                mdThumb = true;
            } else {
                h = w * .5;
                if (titleSmall) {
                    offsetL = 15;
                    offsetT = 30;
                }
                fontSize = Math.max(15, h / 10);
            }
        } else {
            fontSize = 20;
            offsetL = 45;
            offsetT = 30;
            offsetB = 40;
            w = 617 + offsetL;
            h = 398 + offsetB + offsetT;
        }
        let qn = Raphael($(this).attr("id"), w, h);
        let chartH = h - offsetB - offsetT;
        let x = proteinSize / 600 * (w - offsetL) + offsetL;
        let y = chartH - chartH * (qmean4norm / 1.5) + offsetT;
        qn.rect(0, 0, w, h).attr({
            fill: "#FFF",
            stroke: "#FFF"
        });
        if (typeof localQMEANNormImage === "undefined") qn.image("/static/images/qmean_norm.png", offsetL, offsetT, w - offsetL, chartH); else qn.image(localQMEANNormImage, offsetL, offsetT, w - offsetL, chartH);
        qn.rect(offsetL, offsetT, w - offsetL, chartH).attr({
            stroke: "#DDD"
        });
        qn.text(x, y, String.fromCharCode(9733)).attr({
            "text-anchor": "middle",
            fill: "#F44",
            "font-size": fontSize
        });
        let scaleFont = 14;
        if (thumbnail) {
            if (titleSmall) {
                qn.text((w + offsetL) / 2, 8, "Comparison with Non-redundant").attr({
                    "text-anchor": "middle",
                    fill: "#444",
                    "font-size": Math.max(8, w / 32)
                });
                qn.text((w + offsetL) / 2, 22, "Set of PDB Structures").attr({
                    "text-anchor": "middle",
                    fill: "#444",
                    "font-size": Math.max(8, w / 32)
                });
                qn.text(6, h / 2 + offsetT / 2, "Normalized QMEAN4").attr({
                    "text-anchor": "middle",
                    fill: "#666",
                    "font-size": Math.max(8, w / 32)
                }).rotate(270, 6, h / 2 + offsetT / 2);
            }
            qn.rect(offsetL + 1, h - Math.max(16, w / 16), w - offsetL - 2, Math.max(16, w / 16)).attr({
                fill: "#FFF",
                "stroke-width": 0
            });
            qn.text(w / 2, h - 8, "Protein Size (Residues)").attr({
                "text-anchor": "middle",
                fill: "#444",
                "font-size": Math.max(8, w / 32)
            });
        } else {
            qn.text((w + offsetL) / 2, 10, "Comparison with Non-redundant Set of PDB Structures").attr({
                "text-anchor": "middle",
                fill: "#444",
                "font-size": fontSize
            });
            qn.text(10, h / 2, "Normalized QMEAN4 Score").attr({
                "text-anchor": "middle",
                fill: "#666",
                "font-size": fontSize
            }).rotate(270, 10, h / 2);
            qn.text(offsetL + (w - offsetL) / 2, h - 12, "Protein Size (Residues)").attr({
                "text-anchor": "middle",
                fill: "#666",
                "font-size": fontSize
            });
            for (let i = 100; i < 600; i += 100) {
                qn.text(offsetL + i * (w - offsetL) / 600, h - offsetB + 8, i).attr({
                    fill: "#444",
                    "font-size": scaleFont
                });
            }
            for (let i = 1.5; i >= 0; i -= .5) {
                qn.text(offsetL - 12, chartH * (i / 1.5) + offsetT, (1.5 - i).toFixed(1)).attr({
                    fill: "#444",
                    "font-size": scaleFont
                });
            }
        }
        thumbnail = false;
    });
}